#include "prometheus_dispatcher.h"

char read_buffer [max_chunk_size];
fstream log_file;
bool is_debug;
string startup_time;
int connection_requests;

socket_map all_connections;

// jhaley 20111031: For MSVC++ compiler, link with WinSock 2 using linker pragma
#ifdef _MSC_VER
#pragma comment(lib, "ws2_32.lib")
#endif

// jhaley 20111031: Defines for cross-platform portability 
//  (ameliorate differences between BSD sockets and WinSock):
// * my_closesocket wraps closesocket/close
// * my_recv wraps recv/read
// * my_send wraps send/write
// * my_errno wraps WSAGetLastError/errno
// * Define socklen_t as an integer typedef for Windows
#ifdef USE_WINSOCK
#define my_closesocket(i) closesocket((i))
#define my_recv(s, buf, len, flags) recv(s, buf, len, flags)
#define my_send(s, buf, len, flags) send(s, buf, len, flags)
typedef int socklen_t;
#define my_errno (WSAGetLastError())
#else
#define my_closesocket(i) close((i))
#define my_recv(s, buf, len, flags) read(s, buf, len)
#define my_send(s, buf, len, flags) write(s, buf, len)
#define my_errno errno
#endif

//
// socketcheck
//
// jhaley 20111031: Proper error checking for socket failures.
// This was actually incorrect in the original code because it assumed 0 is an
// invalid socket number on BSD sockets, and AFAIK that isn't true. Either way,
// on Windows, INVALID_SOCKET is the only value that indicates a bad return.
//
bool socketcheck(socket_basic_type code)
{
#ifdef USE_WINSOCK
   return (code != INVALID_SOCKET);
#else
   return (code < 0);
#endif
}


void CloseConnections (int & status)
{
   socket_map::iterator i_this_socket, i_next_socket;
   i_this_socket = all_connections.begin();
   while (i_this_socket != all_connections.end())
   {
      i_next_socket = i_this_socket;
      i_next_socket++;

      if (!(i_this_socket->second.is_active))
      {
         output_map (output_status, "Closing connection on handle " + IntToString (i_this_socket->first), i_this_socket->second.properties);
         status = shutdown(i_this_socket->first, 2);
         my_closesocket(i_this_socket->first);
         PostEvent(i_this_socket->first, "client_disconnect", all_connections);
         PostEvent(i_this_socket->first, "unset_property", all_connections);
         all_connections.erase(i_this_socket);
      }
      i_this_socket = i_next_socket;
   }
}

void manual_kill (int z)
{
   cout << "SIGUSR1 caught, exiting semi-gracefully. Not really.\n";
   cerr << "SIGUSR1 caught, exiting semi-gracefully. Not really.\n";
   int i;
   CloseConnections(i);
   exit(z);
}

int main(int argc, char * argv[], char * envp[])
{
#ifndef USE_WINSOCK
   // jhaley: no signals on Windows
   signal (SIGPIPE, SIG_IGN);
   signal (SIGUSR1, manual_kill);
#else
   // jhaley: Init WinSock
   WORD wVersionRequested;
   WSADATA wsaData;
   int err;

   wVersionRequested = MAKEWORD(2, 2);
   if((err = WSAStartup(wVersionRequested, &wsaData)) != 0)
   {
      cout << "WSAStartup failed with error " << IntToString(err) << endl;
      return 1;
   }
#endif
   
   startup_time = right_now();
   connection_requests = 0;

   // stl stuff
   string ip_address;
   string log_file_name;

   // socket address information
   struct sockaddr_in server_socket_address;
   struct sockaddr_in client_socket_address;
   socklen_t size_of_server_socket = sizeof(server_socket_address);
   socklen_t size_of_client_socket = sizeof(client_socket_address);

   // bool
   bool is_done     = false;

   // integers
   int child_pid    = 0;
   int status       = 0;
   int return_value = 0;
   int port_num     = 0;

   // file descriptors
   socket_basic_type my_socket  = 0;
   socket_basic_type highest_fd = 0;
   socket_basic_type this_fd    = 0;
   socket_basic_type this_connection_num = 0;

   // file descriptor bit masks for select()
   fd_set fds_to_read;
   fd_set fds_to_write;

   // file streams

   // time-related
   timeval delay_time;

   // code

   // process command-line params
   is_debug = false;
   if (argc > 2)
   {
      port_num = StringToInt(string(argv[1]));
      log_file_name = string(argv[2]);
      if (argc > 3)
      {
         if (string(argv[3]) == "debug")
         {
            is_debug = true;
            output (output_prefork, "Status: running in debug mode, will not fork, outputting debug to stdout.");
         }
      }
   } 
   else
   {
      cout << "Prometheus' message dispatcher" << endl
           << "Usage: dispatcher port_num log_file_location [debug]" << endl
           << "       You may specify /dev/null as the log_file_location to discard log file information." << endl
           << "       If you specify 'debug', daemon will not fork and some output will be sent to stdout." << endl;
#ifdef USE_WINSOCK
      // jhaley: winsock shutdown
      WSACleanup();
#endif
      exit(0);
   }

   // open log file
   log_file.open(log_file_name.c_str(), ios::out | ios::app);
   if (! log_file)
      output (output_error, "Warning: messaging log file (" + log_file_name + ") could not be opened for writing.");

   try
   {
      output (output_status, "Messaging server started, beginning setup for listen()");

      // create the listener socket
      my_socket = socket(PF_INET, SOCK_STREAM, 0);

      if(!socketcheck(my_socket)) // jhaley: use socketcheck
      {
         output (output_debug, "no socket object available after call to socket()");
         throw 1;
      }

      int want_to_reuse = 1;
      if (setsockopt(my_socket, SOL_SOCKET, SO_REUSEADDR, (const char *)&want_to_reuse, sizeof(want_to_reuse)) < 0)
      {
         output (output_debug, "failed to set SO_REUSEADDR via setsockopt()");
         throw 99;
      }

      // create a structure that describes the socket, accepting connections from anyone on this port_num
      server_socket_address.sin_port = htons(port_num);
      server_socket_address.sin_family = AF_INET;
      server_socket_address.sin_addr.s_addr = htonl(INADDR_ANY);

      // associate the info we just set up with the socket 
      status = bind(my_socket, (sockaddr *)(&server_socket_address), size_of_server_socket);
      if(status)
      {
         output (output_debug, "bind() returned " + IntToString(status));
         throw 2;
      }

      // make the socket into a listening socket
      status = listen(my_socket, 15);
      if(status)
      {
         output (output_debug, "listen() returned " + IntToString(status));
         throw 3;
      }

      // find out where we REALLY are, port-wise
      status = getsockname(my_socket, (sockaddr *)(&server_socket_address), &size_of_server_socket);
      if (status)
      {
         output (output_debug, "getsockname() returned " + IntToString(status));
         throw 7;
      }

      // we're a daemon, remember? fork() returns the pid of the child 
      // jhaley 20111031: there is no fork on Windows; must run as if in debug mode (single process)
#ifndef USE_WINSOCK
      if (!is_debug)
      {
         if (child_pid = fork())
         {
            output (output_prefork, "Message server started on port " + IntToString (ntohs(server_socket_address.sin_port)) + ", process # " + IntToString(child_pid) + ".");
            exit(0);
         }
      } 
      else
#endif
      {
         output (output_prefork, "Daemon will not fork(), as debug mode is enabled. Prepare for an onslaught of messages.");
      }

      // client_socket_address will be filled in with info about the client that 
      // connects (when the call to accept() happens)
      while(! is_done)
      {
         output (output_debug, "Setting up select()");

         highest_fd = 0;
         FD_ZERO(&fds_to_read);
         FD_ZERO(&fds_to_write);
         for(socket_map::iterator i_socket = all_connections.begin(); i_socket != all_connections.end(); i_socket++)
         {
            this_fd = i_socket->first;

            // to speed up the select() call, find highest fd
            highest_fd = max(this_fd, highest_fd);

            // always listen for reading
            FD_SET(this_fd, &fds_to_read);

            // listen for writing if we've got something to send
            if (i_socket->second.out.IsMessageProcessing() || i_socket->second.out.IsMessagePending())
            {
               FD_SET(this_fd, &fds_to_write);
            }
         }

         // now do this for the listening socket as well, for accepting connections
         FD_SET(my_socket, &fds_to_read);

         highest_fd = max(my_socket, highest_fd);

         highest_fd++;
         delay_time.tv_sec  = 180;
         delay_time.tv_usec =   0;

         return_value = select(highest_fd, &fds_to_read, &fds_to_write, NULL, &delay_time);

         output (output_debug, "select(" + IntToString(highest_fd) + ") returned " + IntToString (return_value));

         // if there are connections wanting to read or write, do stuff
         if (return_value)
         {
            // check for newly connecting clients
            if (FD_ISSET(my_socket, &fds_to_read))
            {
               map<string, string> temp_map;
               this_connection_num = accept(my_socket, (sockaddr *)(&client_socket_address), &size_of_client_socket);
               if (!socketcheck(this_connection_num)) // jhaley: use socketcheck
               {
                  output (output_error, "Attempted connection aborted, accept() returned zero.");
                  throw 4;
               }

               // if we got a number, the client connected sucessfully
               // TODO -- put their info in the map (fix this later)
               // ip address of client
               void * addr_byte_array = &(client_socket_address.sin_addr);
               unsigned char * addr_char_array = (unsigned char *)addr_byte_array;
               ip_address = IntToString((int)((unsigned char)(addr_char_array[0])))
                    + "." + IntToString((int)((unsigned char)(addr_char_array[1])))
                    + "." + IntToString((int)((unsigned char)(addr_char_array[2])))
                    + "." + IntToString((int)((unsigned char)(addr_char_array[3])));

               // log ip
               temp_map["ip_address"] = ip_address;

               // log connection date/time
               temp_map["login_timestamp"] = right_now();

               output_map (output_status, "Accepting new connection on handle " + IntToString (this_connection_num), temp_map);
               connection_requests++;
               all_connections[this_connection_num].properties = temp_map;
               all_connections[this_connection_num].is_active  = true;

               PostEvent(this_connection_num, "client_connect", all_connections);
               PostEvent(this_connection_num, "set_property", all_connections);
            } 

            output (output_debug, "Starting reading phase");

            // check for sockets that have received new data
            for(socket_map::iterator i_socket = all_connections.begin(); i_socket != all_connections.end(); i_socket++)
            {
               if (FD_ISSET(i_socket->first, &fds_to_read))
               {
                  if (i_socket->second.in.AttemptRead(i_socket->first))
                  {
                     output (output_debug, "Reading data from stream " + IntToString(i_socket->first));
                  }
                  else
                  {
                     output (output_debug, "Reading from stream " + IntToString(i_socket->first) + " indicated it was closed");
                     i_socket->second.is_active = false;
                  }
               }
            } // end for

            output (output_debug, "Starting routing phase");

            // process ready messages from all sockets
            string this_message;
            for(socket_map::iterator i_socket = all_connections.begin(); (i_socket != all_connections.end()) && (!is_done); i_socket++)
            {
               while ((i_socket->second.in.IsMessagePending()) && (!is_done))
               {
                  this_message = i_socket->second.in.GetMessage();

                  output (output_debug, "Routing from " + IntToString (i_socket->first) + ": " + this_message);

                  if((log_file) && (this_message.length()))
                  {
                     output_map (output_data, this_message, i_socket->second.properties);
                  }         

                  if (!RouteMessage (i_socket->first, this_message, all_connections))
                     is_done = true;
               }
            }

            output (output_debug, "Starting writing phase");

            // check for stuff needing to be sent
            for(socket_map::iterator i_socket = all_connections.begin(); i_socket != all_connections.end(); i_socket++)
            {
               if (FD_ISSET(i_socket->first, &fds_to_write))
               {
                  if (i_socket->second.out.AttemptWrite(i_socket->first))
                  {
                     // sent its fair share
                     output (output_debug, "Writing data to stream " + IntToString(i_socket->first));
                  }
                  else
                  {
                     output (output_debug, "Writing to stream " + IntToString(i_socket->first) + " indicated it was closed with errno " + IntToString(i_socket->second.out.GetCloseReason()));
                     i_socket->second.is_active = false;
                  }
               }
            } // end for

            output (output_debug, "Starting connection dropping phase");

            CloseConnections(status);
         }
         else
         {
            // timed out waiting for anything to happen (no actual requests on any sockets)
            // or maybe some signal did something freaky
            output (output_debug, "Timeout period expired, renewing call to select()");
         }
      } // end while !done

      output (output_debug, "Starting to close all remaining connections");

      // make sure all connections are closed
      for(socket_map::iterator i_socket = all_connections.begin(); i_socket != all_connections.end(); i_socket++)
      {
         if(i_socket->first)
         {
            output_map (output_status, "Closing connection on handle " + IntToString (i_socket->first), i_socket->second.properties);
            status = shutdown(i_socket->first, 2);
            my_closesocket(i_socket->first);
         }
      }

      all_connections.clear();
      status = shutdown(my_socket, 2);
      my_closesocket(my_socket);
      output (output_status, "Closing connection on listening handle " + IntToString (my_socket));
   }
   catch(int the_error)
   {
      if(the_error == 1)
         output (output_error, "Unable to allocate socket memory via socket().");
      else if(the_error == 2)
         output (output_error, "Unable to bind() to server address.");
      else if (the_error == 3)
         output (output_error, "Unable to listen().");
      else if (the_error == 4)
         output (output_error, "Unable to accept().");
      else if (the_error == 5)
         output (output_error, "Unable to shutdown() connection.");
      else if (the_error == 6)
         output (output_error, "Unable to close() connection.");
      else if (the_error == 7)
         output (output_error, "Unable to determine local port number via getsockname().");
      else
         output (output_error, "Unknown integer exception thrown: " + IntToString(the_error));

      if (my_errno)
         output (output_error, "errno has the following value: " + IntToString(my_errno));
   }
   catch (...)
   {
      output (output_error, "catch(...) called, killing daemon.");
      if (my_errno)
         output (output_error, "errno has the following value: " + IntToString(my_errno));
   }

   output (output_status, "Shutting down server. Closing log file.");

   if (log_file)
      log_file.close();

#ifdef USE_WINSOCK
   // jhaley: WinSock shutdown
   WSACleanup();
#endif
   return 0;
}

// logic for routing messages -- joy =)
// needs to have a clue what format stuff is coming in from the client ...

bool RouteMessage (int sending_socket, string message_to_route, socket_map & other_connections)
{
   // formats:
   // see message_text_separator above -- should define which character to use here

   //  SendToAny    (key, value, channel, text)
   //   any~key~value~channel~text

   //  SendToAll    (channel, text)
   //   all~channel~text

   //  SendToServer (text)
   //   server~text

   //  Outgoing messages will only have:
   //   channel~text

   // example of a raw messages being processed:
   //   36:any~user_name~darak~commands~cmd:die
   //   24:all~atlas~people.id|1040
   //   25:server~set~user_name~kims
   // and going out:
   //   20:atlas~people.id|1040
   // the numbers above are NOT seen however -- we only get to see the chunk after the first :

   // cout << "Routing message from " << sending_socket << ": " << message_to_route << endl;

   bool return_value = true;

   message_types message_type;

   vector<string> parsed_message;
   SplitStringToVec(message_to_route, message_text_separator, parsed_message);
   unsigned int parameter_count = parsed_message.size();

   if (parameter_count >= 2)
   {
      message_type = msg_unknown;
      if (parsed_message[0] == "any")
         message_type = msg_to_any;
      if (parsed_message[0] == "all")
         message_type = msg_to_all;
      if (parsed_message[0] == "to")
         message_type = msg_to_connection;
      if (parsed_message[0] == "server")
         message_type = msg_to_server;

      switch (message_type)
      {
      case msg_to_any:
         if (parameter_count >= 5)
         {
            for (socket_map::iterator i_socket = other_connections.begin(); i_socket != other_connections.end(); i_socket++)
            {
               if (i_socket->second.properties[parsed_message[1]] == parsed_message[2])
               {
                  string message_text = rebuild_end(parsed_message, 4);
                  i_socket->second.out.AddMessage(parsed_message[3] + message_text_separator + message_text);
               }
            }
         }
         break;
      case msg_to_all:
         if (parameter_count >= 3)
         {
            for (socket_map::iterator i_socket = other_connections.begin(); i_socket != other_connections.end(); i_socket++)
            {
               string message_text = rebuild_end(parsed_message, 2);
               i_socket->second.out.AddMessage(parsed_message[1] + message_text_separator + message_text);
            }     
         }
         break;
      case msg_to_connection:
         if (parameter_count >= 4)
         {
            socket_map::iterator i_socket = other_connections.find(StringToInt(parsed_message[1]));
            if (i_socket != other_connections.end())
            {
               string message_text = rebuild_end(parsed_message, 3);
               i_socket->second.out.AddMessage(parsed_message[2] + message_text_separator + message_text);
            }
         }
         break;
      case msg_to_server:
         // server~set~key~value
         // server~unset~key
         // server~exit
         // server~list~channel
         // server~notify~channel~event
         // server~unnotify~event
         // server~status~channel
         // server~close~id

         if (parameter_count >= 2)
         {
            if (parsed_message[1] == "set")
            {
               if (parameter_count >= 4)
               {
                  string message_text = rebuild_end(parsed_message, 3);
                  other_connections[sending_socket].properties[parsed_message[2]] = message_text;
                  PostEvent(sending_socket, "set_property", other_connections);
               }
            } 
            else if (parsed_message[1] == "get")
            {
               if (parameter_count >= 4)
               {
                  // TODO -c hack: we're sending the result back via atlas...
                  //string message_text = rebuild_end(parsed_message, 3);
                  //other_connections[sending_socket].properties[parsed_message[2]] = message_text;
                  //PostEvent(sending_socket, "get_property", other_connections); // ?
                  string value = other_connections[sending_socket].properties[parsed_message[2]];
                  string message_text = string("atlas") + message_text_separator + parsed_message[3] + atlas_keyvalue_separator + value;
                  output (output_debug, string("adding message:  ") + message_text);
                  other_connections[sending_socket].out.AddMessage(message_text);
               }
            } 
            else if (parsed_message[1] == "unset")
            {
               if (parameter_count >= 3)
               {
                  other_connections[sending_socket].properties.erase(parsed_message[2]);
                  PostEvent(sending_socket, "unset_property", other_connections);
               }
            } 
            else if (parsed_message[1] == "exit")
            {
               output (output_data, "Command received to shutdown the server. Remaining outgoing messages will attempt to send once. No further messages will be interpreted.");
               return_value = false;
            } 
            else if (parsed_message[1] == "list")
            {
               if (parameter_count >= 3)
               {
                  other_connections[sending_socket].out.AddMessage(parsed_message[2] + message_text_separator + string("starting properties"));
                  for (socket_map::iterator i_socket = other_connections.begin(); i_socket != other_connections.end(); i_socket++)
                  {
                     string connection_id = IntToString(i_socket->first);
                     for (map<string, string>::iterator i_property = i_socket->second.properties.begin(); 
                        i_property != i_socket->second.properties.end(); i_property++)
                        other_connections[sending_socket].out.AddMessage(parsed_message[2] + message_text_separator + connection_id + message_text_separator + i_property->first + message_text_separator + i_property->second);
                  }
                  other_connections[sending_socket].out.AddMessage(parsed_message[2] + message_text_separator + string("ending properties"));
               }
            } 
            else if (parsed_message[1] == "notify")
            {
               if (parameter_count >= 4)
               {
                  string message_text = rebuild_end(parsed_message, 3);
                  other_connections[sending_socket].requests[message_text] = parsed_message[2];
                  output (output_debug, "adding new request: send stuff to '" + parsed_message[2] + "' when event '" + message_text + "' is detected");
               }
            } 
            else if (parsed_message[1] == "unnotify")
            {
               if (parameter_count >= 3)
               {
                  string message_text = rebuild_end(parsed_message, 2);
                  other_connections[sending_socket].requests.erase(message_text);
               }
            } 
            else if (parsed_message[1] == "status")
            {
               if (parameter_count >= 3)
               {
                  string server_info = "Dispatcher: local time is " + right_now() + 
                     ", up since " + startup_time + 
                     ", " + IntToString(connection_requests) + 
                     " connection requests served, " + 
                     IntToString(other_connections.size()) + 
                     " active client connections, your connection id is " + 
                     IntToString(sending_socket) + 
                     ", thanks for asking.";
                  other_connections[sending_socket].out.AddMessage(parsed_message[2] + message_text_separator + server_info);
               }
            } 
            else if (parsed_message[1] == "close")
            {
               if (parameter_count >= 3)
               {
                  socket_map::iterator i_found_socket = other_connections.find(StringToInt(parsed_message[2]));
                  if (i_found_socket != other_connections.end())
                  {
                     i_found_socket->second.is_active = false;
                  }
               }
            }
         }
         break;
      default:
         // who cares?
         break;
      }
   }

   return return_value;
}

void SplitStringToVec(string incoming, string separator, vector<string> &vec)
{
   int lastmatch = 0 - separator.size();
   unsigned int mainmatch = incoming.find(separator, 0);
   while (mainmatch != string::npos)
   {
      vec.push_back(incoming.substr(lastmatch + separator.size(), mainmatch - (lastmatch + separator.size() - 1) - 1));

      lastmatch = mainmatch;
      mainmatch = incoming.find(separator, lastmatch + separator.size());
   }
   vec.push_back(incoming.substr(lastmatch + separator.size()));
}

//convert int to string using C libraries.
string IntToString(int incoming)
{
   string answer = "";
   char floatnumber[30];
   sprintf(floatnumber, "%1i", incoming);
   answer = floatnumber;
   return answer;
}

//convert string to int using C libraries.
int StringToInt(string incoming)
{
   int answer;
   answer = atoi(incoming.c_str());
   return answer;
}

//pad to the left with a zero to make it two digits.
string twodigit(int incoming)
{
   string answer = IntToString(incoming);
   if (answer.length() < 2) answer = "0" + answer;
   return answer;
}

// checks to see if the buffer string is currently in use
bool   inbound_message_queue::IsMessageProcessing ()
{
   return (!current_message.empty());
}

// checks to see if the queue is non-empty
bool   inbound_message_queue::IsMessagePending ()
{
   return (!pending_messages.empty());
}

// returns the next string off the queue
// should be called only if IsMessagePending()
string inbound_message_queue::GetMessage ()
{
   string return_value = pending_messages.front();
   pending_messages.pop_front();
   return return_value;
}

// attempts to read from the file descriptor
// the file descriptor should have been marked as readable prior to call
// returns false if no bytes read (connection closed)
bool   inbound_message_queue::AttemptRead (socket_basic_type file_descriptor)
{
   chunk_size = my_recv(file_descriptor, read_buffer, max_chunk_size - 1, 0); // jhaley: my_recv
   
   read_buffer[chunk_size] = '\0';
   if (!chunk_size)
      return false;
   // cout << ">> " << read_buffer << endl;
   current_message += read_buffer;
   PrepMessage();
   return true;
}

// takes a message, checks for completion, sticks completed message on the queue
void   inbound_message_queue::PrepMessage ()
{
   unsigned int first_colon;
   unsigned int first_char;
   while (IsMessageProcessing())
   {
      first_colon = current_message.find(size_separator);
      first_char  = current_message.find_first_not_of("0123456789");
      if (first_colon != string::npos)
      {
         if (first_char == string::npos || first_char >= first_colon)
         { 
            chunk_size = StringToInt (current_message.substr(0, first_colon));
            if (current_message.length() >= (chunk_size + first_colon + 1))
            {
               pending_messages.push_back(current_message.substr(first_colon + 1, chunk_size));
               if (first_colon + chunk_size + 1 < current_message.length())
                  current_message = current_message.substr(first_colon + chunk_size + 1);
               else
                  current_message = "";
            }
            else
               break;
         }
         else
         {
            // somebody fucked up
            // cout << "dropping packet: " << current_message << endl;
            current_message = "";
         }
      }
      else
      {
         if (first_char != string::npos)
         {
            // cout << "dropping packet: " << current_message << endl;
            current_message = "";
         }
         break;
      }
   }
   return;
}

// checks to see if the buffer string is currently in use
bool outbound_message_queue::IsMessageProcessing ()
{
   return (!current_message.empty());
}

// checks to see if the queue is non-empty
bool outbound_message_queue::IsMessagePending ()
{
   return (!pending_messages.empty());
}

// adds a message to the queue to be sent
// if no message is currently being processed, and some are waiting,
// pulls next available message back off the queue and into the waiting area
void outbound_message_queue::AddMessage (string new_message)
{
   pending_messages.push_back(new_message);
   PrepMessage();
   return;
}

// will attempt to send the next available chunk on the file descriptor
// the file descriptor should have been marked as writable before calling
// returns true so long as nothing went wrong (zero bytes sent, for example)
bool outbound_message_queue::AttemptWrite (socket_basic_type file_descriptor)
{
   bool first_attempt = true;
   while (IsMessageProcessing())
   {
      chunk_size = my_send(file_descriptor, current_message.c_str(), current_message.length(), 0); // jhaley: my_send
      
      // cout << "<< " << current_message.substr(0, chunk_size) << endl;

      if (chunk_size < 0)
      {
         errno_on_close = my_errno;

         if (first_attempt)
            return false;
         else
            break;
      }
      if ((unsigned int)chunk_size < current_message.length())
         current_message = current_message.substr(chunk_size);
      else
         current_message = "";
      PrepMessage();
      first_attempt = false;
   }
   return true;
}

// take next available message for sending and stick it in the sending string, prepared
void outbound_message_queue::PrepMessage ()
{
   if (!IsMessageProcessing() && IsMessagePending())
   {
      current_message = pending_messages.front();
      current_message = IntToString(current_message.length()) + size_separator + current_message;
      pending_messages.pop_front();
   }
   return;
}

int outbound_message_queue::GetCloseReason ()
{
   return errno_on_close;
}

string right_now ()
{
   tm tm_now;
   time_t time_t_now;
   string answer;

   time_t_now = time(&time_t_now);
   tm_now     = *(localtime(&time_t_now));
   answer += IntToString(tm_now.tm_year + 1900) + "/" 
          + twodigit(tm_now.tm_mon + 1)        + "/" 
          + twodigit(tm_now.tm_mday)           + " " 
          + twodigit(tm_now.tm_hour)           + ":" 
          + twodigit(tm_now.tm_min)            + ":" 
          + twodigit(tm_now.tm_sec);
   return answer;
}

void output (output_types output_type, string text)
{
   bool to_cout = false;
   bool to_file = false;
   string answer = "[" + right_now() + " | ";

   switch (output_type)
   {
   case output_debug:
      answer += "Debug";
      if (is_debug) to_cout = true;
      break;
   case output_status: 
      answer += "Status";
      if (is_debug) to_cout = true;
      to_file = true;
      break;
   case output_prefork:
      answer += "Pre-fork";
      to_cout = true;
      break;
   case output_error:
      answer += "Error";
      if (is_debug) to_cout = true;
      to_file = true;
   case output_data:
      answer += "Data";
      to_file = true;
      break;
   default:
      answer += "Unknown";
      to_cout = true;
      to_file = true;
   }

   answer += "] " + text;
   if (to_cout && cout)
      cout << answer << endl << flush;
   if (to_file && log_file)
      log_file << answer << endl << flush;
   return;
}

void output_map (output_types output_type, string text, map<string, string> & params)
{
   string answer = "(" + map_string(params) + ") " + text;
   output (output_type, answer);
   return;
}

string map_string (const map<string, string> & params)
{
   string answer;
   for (map<string, string>::const_iterator i_entry = params.begin(); i_entry != params.end(); i_entry++)
   {
      if (i_entry != params.begin())
         answer += ", ";
      answer += (i_entry->first + ":" + i_entry->second);
   }
   return answer;
}

string rebuild_end (vector<string> & params, int start_at)
{
   string answer;
   for (size_t i = start_at; i < params.size(); i++)
   {
      if (i > (size_t)start_at)
         answer += (message_text_separator);
      answer += params[i];
   }
   return answer;
}

void PostEvent (int cause_socket, string event_name, socket_map & all_connections)
{
   output (output_debug, "Event worth mention from " + IntToString(cause_socket) + ": " + event_name);
   map<string, string>::iterator i_request;
   for (socket_map::iterator i_socket = all_connections.begin(); i_socket != all_connections.end(); i_socket++)
   {
      output (output_debug, "seeing if " + IntToString(i_socket->first) + " cares?");
      i_request = i_socket->second.requests.find(event_name);
      if (i_request != i_socket->second.requests.end())
      {
         i_socket->second.out.AddMessage(i_request->second + message_text_separator + IntToString(cause_socket) + message_text_separator + event_name);
         output (output_debug, "tried to send a notification to " + IntToString(i_socket->first) + " about event '" + event_name + "' coming from " + IntToString(cause_socket));
      }
   }
}
