/*

   HL7Daemon - Record Matcher (People, Tax Entities, Etc.)

   This module is responsible for taking the processed HL7 messages and then
   doing the following:

   * Determining existing records to be updated
   * Determining new records to be inserted
   * Determining records to mark expired or deleted, if any

*/

#include "hl7match.h"
#include "inifile.h"
#include "sqlmanager.h"
#include "utils.h"

#pragma hdrstop
#include "util.h"

//
// Destructor
//
HL7Matcher::~HL7Matcher()
{
}

// Lookup tables to load with standard SQL
static const char *lookups[] =
{
   "current_types",
};

//
// HL7Matcher::getBasicDatabaseInfo
//
// Retrieve the various odds and ends we need to do inserts, updates, etc.
//
bool HL7Matcher::getBasicDatabaseInfo()
{
   string usersql = "select id from users where username = 'hl7daemon'";

   // Get user id for the HL7Daemon Prometheus user account
   db->getOneField(usersql, daemonUserID);
   if(IsBlankOrZero(daemonUserID))
   {
      DEBUGOUT(dbg_matcher, "Could not retrieve hl7daemon user ID");
      return false;
   }

   // Load needed lookups
   PrometheusLookups::LoadLookups(*db, lookups, arrlen(lookups));

   return true;
}

//
// HL7Matcher::optimizeForPrimary
//
// Since the main loop below is iterating in end-to-beginning order and found a
// STATUS_NORMAL message for a given person, this is the newest message for that
// person. So long as it is well-formed, mark it as WILLMATCH and any older
// messages for this primary should then be marked as OUTOFDATE.
//
void HL7Matcher::optimizeForPrimary(reverse_interm_iterator itr)
{
   HL7Intermediate *interm = itr->second;
   string id = interm->primaryPerson.fields["id"];

   if(!IsBlankOrZero(id))
      interm->status = HL7Intermediate::STATUS_WILLMATCH;
   else
   {
      interm->status = HL7Intermediate::STATUS_CANTMATCH; // not matchable.
      return;
   }

   // Start one past the current position passed down from the main loop
   ++itr;

   for(; itr != im->rend(); ++itr)
   {
      HL7Intermediate *curInterm = itr->second;

      // older message for the same person?
      if(curInterm->primaryPerson.fields["id"] == id)
         curInterm->status = HL7Intermediate::STATUS_OUTOFDATE; // it's older.
   }
}

//
// HL7Matcher::optimizeIntermediates
//
// Set the status of HL7Intermediate objects based on temporal analysis of the
// message queue.
//
void HL7Matcher::optimizeIntermediates()
{
   for(reverse_interm_iterator itr = im->rbegin(); itr != im->rend(); ++itr)
   {
      // if not touched yet, process it.
      if(itr->second->status == HL7Intermediate::STATUS_NORMAL)
         optimizeForPrimary(itr); // note iterator is passed by value.
   }
}

//
// HL7Matcher::findBestMatches
//
// Upon successful retrieval, outvm will contain a vector of person IDs and
// the corresponding match score for that person.
//
bool HL7Matcher::findBestMatches(HL7Person &p, pdb::vecmap &outvm)
{
   string matchScript;
   SQLManager::VariableMap vars;
   outvm.clear();

   // At bare minimum, we need an ID and birth date for a successful matching
   // process. Note this is the source db ID, and not necessarily equal to the
   // Prometheus ID. According to the spec, these should always be part of the
   // HL7 message, but I'm not oft one to assume.
   if(IsBlankOrZero(p.fields["id"]) || IsBlankOrZero(p.fields["birth_date"]))
   {
      DEBUGOUT(dbg_matcher, "Person is missing sufficient info for matching");
      return false;
   }

   // prepare the person matching script
   vars["PERSON_ID"]  = p.fields["id"];
   vars["SSN"]        = p.fields["social_security_number"];
   vars["GENDER_ID"]  = p.fields["gender_id"];
   vars["BIRTH_DATE"] = p.fields["birth_date"];
   vars["LAST_NAME"]  = p.names[HL7Person::NAME_MAIN]["last_name"];
   vars["FIRST_NAME"] = p.names[HL7Person::NAME_MAIN]["first_name"];
   vars["MI"]         = p.names[HL7Person::NAME_MAIN]["middle_name"];
   vars["SUFFIX"]     = p.names[HL7Person::NAME_MAIN]["suffix"];

   // The middle name is "like" matched since source cb only stores an initial.
   if(vars["MI"] != "")
      vars["MI"] += "%";

   SQLManager::PrepareScript("person_match.sql", vars, matchScript);

   // do it!
   if(!tr->sqlToVecMap(matchScript, outvm))
      return false;

   // Success
   return true;
}

//
// HL7Matcher::doMatching
//
// Do the matching process for each intermediate marked "will match" by the
// optimization phase.
//
void HL7Matcher::doMatching()
{
   for(reverse_interm_iterator itr = im->rbegin(); itr != im->rend(); ++itr)
   {
      if(itr->second->status != HL7Intermediate::STATUS_WILLMATCH)
         continue;

      // TODO
   }
}

//
// HL7Matcher::matchIntermediates
//
// This is the "main" routine for the matching module. Pass in the output from
// the processor.
//
void HL7Matcher::matchIntermediates(HL7Processor::IntermediateMap &pIM)
{
   PrometheusDB database;

   im = &pIM;
   db = &database;

   // Connect to the database
   if(!db->connect(IniFile::GetIniFile()))
   {
      DEBUGOUT(dbg_matcher, "Could not connect to database, matching aborted");
      return;
   }

   // Get database information
   if(!getBasicDatabaseInfo())
   {
      DEBUGOUT(dbg_matcher, "Basic info not acquired, matching aborted");
      return;
   }

   // TODO: more init?

   // Optimize the queue
   optimizeIntermediates();

   // Run the matching process
   doMatching();
}

// EOF

