/*

  HL7 Parser Prototype

*/

#ifndef HL7PARSE_H__
#define HL7PARSE_H__

#include <string>
#include <vector>
#include <map>
#include <stack>

using namespace std;

// prototype static field info
struct HL7FieldInfo
{
   int level;
   const char *fieldname;
   HL7FieldInfo *subComponents; // expected subcomponents, if any exist
   int numExpectedSubComponents;
};

// runtime field info - each segment constructed has a dynamic copy of its
// static HL7FieldInfo tree in the form of a tree of objects of this class.
class HL7Value
{
protected:
   friend class HL7Segment;
   HL7Value *findInSubComponents0(const char *token);
   
public:
   int level;
   const char *fieldname;
   HL7Value *subComponents;
   int numExpectedSubComponents;
   HL7Value *repeatValues; // root of repeat values, if any, or next value
   int repeatCount;        // repeat generation number
   string value;

   HL7Value()
     : level(0), fieldname(NULL), subComponents(NULL),
       numExpectedSubComponents(0), repeatValues(NULL), repeatCount(0),
       value("")
   {
   }

   HL7Value *findInSubComponents(const char *token);
};

//=============================================================================
//
// HL7Segment
//

class HL7Segment
{
public:
   class Type
   {
   private:
      enum { NUMTYPECHAINS = 31 };

      // addType is invoked by the constructor and places the type into a
      // global hash table for lookups by class name.
      void addType();

      Type *next; // next type on the same hash chain
      
      static Type *segmentTypes[NUMTYPECHAINS]; // hash table

   protected:
      // The constructor is always protected, but the type being proxied is
      // always a friend so that it can construct a singleton instance of the
      // proxy type as a static class member.
      Type(const char *pName, Type *pParent) : parent(pParent), name(pName)
      {
         addType();
      }

      Type       *const parent; // Poitner to the parent class's proxy instance
      const char *const name;   // Name of this class's message

   public:
      typedef HL7Segment Class; // Type of the class being proxied
      friend class HL7Segment;  // The proxied class is always a friend.

      // isAncestorOf: walk the inheritance tree to determine if this type
      // inherits from the indicated type. Supports single inheritance only.
      bool isAncestorOf(const Type *type) const
      {
         while(type)
         {
            if(type == this)
               return true;
            type = type->parent;
         }
         return false;
      }

      static Type *FindType(const char *pName);

      // virtual constructor factory method
      virtual HL7Segment *newSegment() const { return NULL; }
   };

protected:
   HL7Value *valueTree;
   void constructValueTreeRecursive(const HL7FieldInfo *info, int numComponents,
                                    HL7Value **root);

   friend class HL7Parser;

   void destroyValueTree(HL7Value *root);
   
public:
   HL7Segment() : valueTree(NULL) {}

   ~HL7Segment();

   virtual int getNumFields() const = 0;
   virtual const HL7FieldInfo *getFieldInfo() const = 0;

   static HL7Segment *SegmentForSegmentName(const char *);

   // HL7Segment's proxy type instance.
   static Type StaticType;

   // getDynamicType will always return the most-derived (or "actual") type of
   // the object even when invoked through pointers or references to super
   // classes. You are required to override this method.
   virtual const Type *getDynamicType() const { return &StaticType; }

   // getSegmentName will always return the name of the most-derived (or "actual")
   // type of the object even when invoked through pointers or references to
   // super classes.
   const char *getSegmentName() const { return getDynamicType()->name; }

   // isInstanceOf: return true if the segment is of the indicated actual type,
   // using a pointer to the class RTTI object.
   bool isInstanceOf(const Type *type) const
   {
      return (getDynamicType() == type);
   }

   // isInstanceOf: return true if the segment is of the indicated actual type,
   // using the HL7 segment name.
   bool isInstanceOf(const char *segmentName) const
   {
      return !strcmp(getSegmentName(), segmentName);
   }

   // isAncestorOf: return true if the given type is ancestral to this instance
   bool isAncestorOf(const Type *type) const
   {
      return getDynamicType()->isAncestorOf(type);
   }

   // isDescendantOf: return true if the given type is a descendant of this
   // instance's type.
   bool isDescendantOf(const Type *type) const
   {
      return type->isAncestorOf(getDynamicType());
   }

   void constructValueTree();

   HL7Value *getValueTree() { return valueTree; }
   HL7Value *findInValueTree(const char *fieldPath);
   HL7Value *valueForIndex(int i);
};

//
// DECLARE_MSG_TYPE
//
// Use this macro once per HL7Segment descendant, inside the class definition.
// The following public members are declared:
//
// typedef inherited Super;
// * This allows name::Super to be used as a type which implicitly references
//   the immediate parent class.
//
// class Type;
// * This is the class's RTTI proxy type and it automatically inherits from the
//   Super class's proxy. The constructor is protected. The proxy class exposes
//   the type it proxies for (ie., name) as Type::Class, and a virtual 
//   newSegment() factory constructor method. Note using this macro will exact
//   the requirement of a default constructor on the HL7Segment descendant.
//
// static Type StaticType;
// * This is the singleton instance of RTTI proxy type for the HL7Segment
//   descendant. It is instantiated by the IMPLEMENT_MSG_TYPE macro below.
//
// virtual const Type *getDynamicType() const;
// * This method of the HL7Segment descendant will return the StaticType 
//   member, which in the context of each individual class, is the instance
//   representing the actual most-derived type of the object, ie., 
//   name::StaticType (the parent instances of StaticType are progressively
//   hidden in each descendant scope).
//

#define DECLARE_MSG_TYPE(name, inherited) \
public: \
   typedef inherited Super; \
   class Type : public Super::Type \
   { \
   protected: \
      Type(char const *pName, Super::Type *pParent) \
       : Super::Type(pName, pParent) {} \
   public: \
      typedef name Class; \
      friend class name; \
      virtual name *newSegment() const { return new name ; } \
   }; \
   static Type StaticType; \
   virtual const Type *getDynamicType() const { return &StaticType; } \
private:

//
// IMPLEMENT_MSG_TYPE
//
// Use this macro once per HL7Segment descendant, at file scope within a single
// relevant translation module.
//
// Example:
//   IMPLEMENT_MSG_TYPE(MSHSegment, "MSH")
//   This defines MSHSegment::StaticType and constructs it with 
//   "MSH" as its HL7 message type name.
// 

#define IMPLEMENT_MSG_TYPE(className, name) \
   className::Type className::StaticType(name, &Super::StaticType);

// Short-hand macro for use with isInstanceOf, etc.
#define RTTI(cls) (&cls::StaticType)

//=============================================================================
//
// Tokenizer
//

// Possible tokens:
// * Delimiters: | ^ ~ &
// * Anything else: part of a field
// * When read \: Escape character
// * 0x0D: End of Line (EOL)
// * \0: End of String (EOS)

class HL7Tokenizer
{
protected:
   // Tokenizer states
   enum
   {
      TSTATE_SCAN,    // Scanning for start of next token
      TSTATE_INFIELD, // Scanning inside a field
      TSTATE_ESCAPE,  // Found an escape char
      TSTATE_ESCHEX,  // Scanning a Hex number
      TSTATE_DONE     // Done, return token to parser
   };

   // Tokenizer state handlers
   void doStateScan();
   void doStateInField();
   void doStateEscape();
   void doStateEscHex();

   // Tokenizer state table
   static void (HL7Tokenizer::*States[])();

   // Tokenizer state data
   int state;               // Current tokenizer state
   int escapedState;        // State we were in when we escaped a character
   const char *input;       // Input String
   int textPos;             // Position in input string
   int tokenType;           // Type of newest token
   string tokenText;        // Token text
   bool doNotAdvance;       // Don't advance before looping
   string hexNum;           // Building a hexnum for escaped char
   bool disableEscapes;     // Escapes are currently disabled

public:
   // Token Types
   enum
   {
      TOKEN_UNKNOWN,
      TOKEN_FIELDVALUE,
      TOKEN_PIPE,
      TOKEN_HAT,
      TOKEN_TILDE,
      TOKEN_BACKSLASH, // only returned when escapes are disabled.
      TOKEN_AMPERSAND,
      TOKEN_EOL,
      TOKEN_EOS,
      TOKEN_NUMTOKENS
   };

   static const char *const TokenNames[TOKEN_NUMTOKENS];

   // Constructor
   HL7Tokenizer()
      : state(TSTATE_SCAN), escapedState(TSTATE_SCAN), input(NULL), textPos(0), 
        tokenType(TOKEN_UNKNOWN), tokenText(""), doNotAdvance(false), 
        hexNum("")
   {
   }

   // Methods
   void nextToken();

   // Accessors
   int           getTokenType() const { return tokenType; }
   const string &getTokenText() const { return tokenText; }
   
   void setInput(const char *pInput)  { input = pInput; textPos = 0; }
   void setDisableEscapes(bool b)     { disableEscapes = b;          }
};

//==============================================================================
//
// HL7Parser
//

class HL7Parser
{
public:
   typedef vector<HL7Segment *>  SegmentVector;
   typedef stack<HL7Value *> ValueStack;
   
protected:
   // Parser states
   enum
   {
      STATE_EXPECTSEGMENTNAME,
      STATE_MSHSPECIALCHARS,
      STATE_EXPECTDELIMITER,
      STATE_EXPECTVALUEORDELIMITER,
      STATE_SPIN, // throw away unexpected subfields
      STATE_DONE, // terminated normally
      STATE_STOP  // error state
   };

   // State handlers
   void doStateExpectSegmentName();
   void doStateMSHSpecialChars();
   void doStateExpectDelimiter();
   void doStateExpectValueOrDelimiter();
   void doStateSpin();

   // Stuff (tm)
   char determineExpectedDelimiter(bool &needToPop);
   void popStack();
   void popAndStep();
   void popTwoAndStep();
   void stepToNext();
   void badDelim(char expected, char found);
   void nextField(char delimiter);
   void setCurField(HL7Value *newField);
   void setNewSegment(HL7Segment *newSegment);

   // Parser state table
   static void (HL7Parser::*States[])();

   // Parser state data
   int state;               // Current tokenizer state
   const char *input;       // Input String
   HL7Tokenizer tokenizer;  // Tokenizer

   SegmentVector *segments; // Output data structure
   HL7Segment *curSegment;  // Current segment being parsed
   string curValue;         // Current value being built
   ValueStack pstack;       // Parser stack tracking pointers into value tree
   HL7Value *curField;      // Current field at the current level
   bool delimiterError;     // Flag unexpected subfield divisions
   char spinExpectedDelim;  // The delimiter we WANTED to see...
   char spinFoundDelim;     // ...and what we found instead.
   bool errorState;         // True if a parsing error occurred
   int  errorCode;          // Value representing what error occurred

   // repeating fields
   bool       repeating;    // true if doing a subsequent pass over repeating fields
   ValueStack backupStack;  // Saved state of the stack when starting a field
   HL7Value  *backupValue;  // Saved value pointer when starting a field
   int        repeatCount;  // Number of times have repeated consecutively

public:

   // Constructor
   HL7Parser()
      : state(STATE_EXPECTSEGMENTNAME), input(NULL), tokenizer(),
        segments(NULL), curSegment(NULL), curValue(""), pstack(),
        curField(NULL), delimiterError(false), spinExpectedDelim('\0'),
        spinFoundDelim('\0'), errorState(false), errorCode(0), repeating(false),
        backupStack(), backupValue(NULL), repeatCount(0)
   {
   }

   // pass your message back in here when you're done or you will leak a LOT
   // of memory :P
   static void DestroySegments(SegmentVector *sv)
   {
      if(sv)
      {
         for(size_t i = 0; i < sv->size(); i++)
            delete (*sv)[i]; // delete the HL7Segment objects
         delete sv; // delete the vector
      }
   }

   ~HL7Parser()
   {
      if(input)
         delete [] input;

      if(segments)
      {
         DestroySegments(segments);
         segments = NULL;
      }
   }

   bool loadFile(const char *filename);
   void loadString(const char *text); // for testing only! ;)

   SegmentVector *parse();

   bool isBad() const { return errorState; }
   const char *getErrorMessage() const;
};

#endif

// EOF

