/*

  HL7Daemon - HL7 Message Processing Core

  This is the real meat, where we'll organize HL7 messages and translate them
  into an intermediate form as close as possible to Prometheus' database
  schema. Then we'll send the stuff on to the person matcher.

*/

#include <fstream>
using namespace std;

#include "dbmisc.h"
#include "hl7process.h"
#include "inifile.h"
#include "scanner.h"
#include "stringutils.h"
#include "utils.h"

#pragma hdrstop
#include "util.h"

// I hate this warning: "Possibly incorrect assignment"
#pragma warn -8060

//---------------------------------------------------------------------------

//
// HL7Processor::~HL7Processor
//
// HL7 processor destructor
//
HL7Processor::~HL7Processor()
{
   // Destroy all HL7Parser::SegmentVector instances
   size_t numMessages = messages.size();

   for(size_t i = 0; i < numMessages; i++)
      HL7Parser::DestroySegments(messages[i]);

   messages.clear();

   // Destroy all HL7Intermediate instances
   for(IntermediateMap::iterator i = intermediates.begin();
       i != intermediates.end();
       ++i)
   {
      delete i->second;
   }
   intermediates.clear();
}

//
// HL7Processor::scanCallback
//
// Handler method for use with ScanForFiles, called via Borland closure.
//
void HL7Processor::scanCallback(fileinfo &fi)
{
   pair<string, string> item;
   item.first  = (fi.creation_date + " " + fi.creation_time).c_str();
   item.second = fi.GetFullyQualifiedName().c_str();

   timestamp_filename_map.insert(item);
}

//
// HL7Processor::scanForFiles
//
// Use Paul's file scanner module to generate a multimap of timestamps to
// filenames. Then afterward the files can be processed in chronological order.
//
void HL7Processor::scanForFiles()
{
   IniFile::IniMap &ini = IniFile::GetIniOptions();
   queue<AnsiString> paths_to_scan;

   paths_to_scan.push(ini["scanner"]["share"].c_str());
   ScanForFiles(paths_to_scan, this->scanCallback);
}

//
// HL7Processor::parseFiles
//
// Parse all of the files that were added to the timestamp multimap in
// chronological order. The result is a vector of pointers to the SegmentVectors
// that are output by the HL7Parser object, representing each HL7 message.
//
void HL7Processor::parseFiles()
{
   HL7Parser myParser;

   for(TSFNMap::iterator itr = timestamp_filename_map.begin();
       itr != timestamp_filename_map.end();
       ++itr)
   {
      HL7Parser::SegmentVector *segVec;
      AnsiString path_and_file = itr->second.c_str();
      AnsiString fn  = ExtractFileName(path_and_file);
      AnsiString ext = ExtractFileExt(path_and_file);

      DEBUGOUT(dbg_scanner, "Scanner: Examining " << path_and_file.c_str());

      if(!FileExists(path_and_file) || ext.LowerCase() != ".hl7")
      {
         DEBUGOUT(dbg_scanner, "Scanner: not interested in this file.");
         continue;
      }

      // Load the file into the parser
      if(!myParser.loadFile(path_and_file.c_str()))
      {
         DEBUGOUT(dbg_scanner, "Scanner: Parser file load failed.");
         continue;
      }

      // Parse the message and, if successful, save the results
      segVec = myParser.parse();
      if(!myParser.isBad())
         messages.push_back(segVec);
      else
      {
         DEBUGOUT(dbg_scanner, "Scanner: Message parse failed: " << myParser.getErrorMessage());
         HL7Parser::DestroySegments(segVec);
      }
   }
}

//
// HL7Processor::processPrimaryPerson
//
// Extract and translate all data on the primary person addressed by a
// demographics update message.
//
void HL7Processor::processPrimaryPerson(HL7Person &newPerson, PIDSegment *pid)
{
   string countyID;

   // Get references to subordinate records.
   HL7Record &mainName    = newPerson.names[HL7Person::NAME_MAIN];
   HL7Record &homeContact = newPerson.contactInfo[HL7Person::CONTACT_HOME];
   HL7Record &cellContact = newPerson.contactInfo[HL7Person::CONTACT_CELL];
   HL7Record &workContact = newPerson.contactInfo[HL7Person::CONTACT_WORK];

   if(pid)
   {
      Pdate  tempDate;

      // Basic properties

      // ID Number - This will match exactly *if* the person was in our initial
      // demographics import. Otherwise...
      newPerson.fields["id"] = pid->valueForIndex(PIDSegment::PID_INTID)->value;

      // Birth Date
      hl7_DTToPdate(pid->valueForIndex(PIDSegment::PID_DOB)->value, tempDate);
      if(tempDate.IsValidDate())
         newPerson.fields["birth_date"] = tempDate.OutDate(global_mode);

      // Gender
      hl7_GenderToGenderID(pid->valueForIndex(PIDSegment::PID_GENDER)->value,
                           newPerson.fields["gender_id"]);

      // Race
      hl7_RaceToRaceID(pid->valueForIndex(PIDSegment::PID_RACE)->value,
                       newPerson.fields["race_id"]);

      // Marital Status
      hl7_MaritalStatusToMaritalID(pid->valueForIndex(PIDSegment::PID_MARITAL_STATUS)->value,
                                   newPerson.fields["marital_id"]);

      // Social Security Number
      hl7_SSNToNumeric(pid->valueForIndex(PIDSegment::PID_PATSSN)->value,
                       newPerson.fields["social_security_number"]);

      // TODO: Ethnicity? (would need to add to Prometheus schema)

      // Main name for primary person
      mainName["last_name"]   = pid->findInValueTree("patient_name.last_name")->value;
      mainName["first_name"]  = pid->findInValueTree("patient_name.first_name")->value;
      mainName["middle_name"] = pid->findInValueTree("patient_name.mi")->value;
      mainName["suffix"]      = pid->findInValueTree("patient_name.suffix")->value;

      // Contact Info
      //
      // The following information is available from the following PID fields:
      // * Home Address, Home Phone = phone_home
      // * Cell Phone = phone_home repeated value (if any)
      // * Work Phone = phone_business

      // Home address
      homeContact["line1"] = pid->findInValueTree("pat_address.line1")->value;
      homeContact["line2"] = pid->findInValueTree("pat_address.line2")->value;
      homeContact["city"]  = pid->findInValueTree("pat_address.city")->value;

      hl7_StateToStateID(pid->findInValueTree("pat_address.state")->value,
                         homeContact["state_id"]);
      hl7_NormalizeZip(pid->findInValueTree("pat_address.zip")->value,
                       homeContact["zip"]);

      // Lookup the county
      if(CountyByZip(db, homeContact["zip"], countyID))
         homeContact["county_id"] = countyID;

      // Home phone 
      HL7Value *homeXPN   = pid->valueForIndex(PIDSegment::PID_PHONE_HOME);
      HL7Value *homeEquip = homeXPN->findInSubComponents("telecom_equip_type");
      HL7Value *homeEmail = homeXPN->findInSubComponents("email");
      HL7Value *homePhone = homeXPN->findInSubComponents("compound");

      homeContact["email_address"] = homeEmail->value;
      hl7_NormalizePhone(homePhone->value, homeContact["phone_number"]);

      // If telecom_equip_type has a secondary value of "CP", there is a cell
      // phone value provided.
      if(homeEquip->repeatValues && homeEquip->repeatValues->value == "CP")
      {
         HL7Value *cellEmail = homeEmail->repeatValues;
         HL7Value *cellPhone = homePhone->repeatValues;

         if(cellEmail)
            cellContact["email"] = cellEmail->value;

         if(cellPhone)
            hl7_NormalizePhone(cellPhone->value, cellContact["phone_number"]);
      }

      // Work phone
      workContact["email_address"] = pid->findInValueTree("phone_business.email")->value;
      hl7_NormalizePhone(pid->findInValueTree("phone_business.compound")->value,
                         workContact["phone_number"]);
   }
}

//
// HL7Processor::processNextOfKin
//
// Extract and translate all data on the primary person's next-of-kin.
//
void HL7Processor::processNextOfKin(HL7Person &nextOfKin, NK1Segment *nk1)
{
   string countyID;
   HL7Record &mainName    = nextOfKin.names[HL7Person::NAME_MAIN];
   HL7Record &homeContact = nextOfKin.contactInfo[HL7Person::CONTACT_HOME];

   if(nk1)
   {
      // Main name for next-of-kin
      mainName["last_name"]   = nk1->findInValueTree("name.last_name")->value;
      mainName["first_name"]  = nk1->findInValueTree("name.first_name")->value;
      mainName["middle_name"] = nk1->findInValueTree("name.mi")->value;
      mainName["suffix"]      = nk1->findInValueTree("name.suffix")->value;

      // Work address
      homeContact["line1"] = nk1->findInValueTree("address.line1")->value;
      homeContact["line2"] = nk1->findInValueTree("address.line2")->value;
      homeContact["city"]  = nk1->findInValueTree("address.city")->value;

      hl7_StateToStateID(nk1->findInValueTree("address.state")->value,
                         homeContact["state_id"]);
      hl7_NormalizeZip(nk1->findInValueTree("address.zip")->value,
                       homeContact["zip"]);

      // Lookup the county
      if(CountyByZip(db, homeContact["zip"], countyID))
         homeContact["county_id"] = countyID;

      // Work phone
      homeContact["email_address"] = nk1->findInValueTree("phone_number.email")->value;
      hl7_NormalizePhone(nk1->findInValueTree("phone_number.compound")->value,
                         homeContact["phone_number"]);
   }
}

//
// HL7Processor::processGuarantor
//
// Extract and translate all data on the primary person's guarantor.
//
void HL7Processor::processGuarantor(HL7Person &guarantor, GT1Segment *gt1)
{
   string countyID;
   
   // Get references to subordinate records.
   HL7Record &mainName    = guarantor.names[HL7Person::NAME_MAIN];
   HL7Record &homeContact = guarantor.contactInfo[HL7Person::CONTACT_HOME];
   HL7Record &cellContact = guarantor.contactInfo[HL7Person::CONTACT_CELL];
   HL7Record &workContact = guarantor.contactInfo[HL7Person::CONTACT_WORK];

   if(gt1)
   {
      Pdate tempDate;
      
      // Basic properties

      // ID Number - This will match exactly *if* the person was in our initial
      // demographics import. Otherwise...
      guarantor.fields["id"] = gt1->valueForIndex(GT1Segment::GT1_SETID)->value;

      // Birth Date
      hl7_DTToPdate(gt1->valueForIndex(GT1Segment::GT1_BIRTH_TIMESTAMP)->value, tempDate);
      if(tempDate.IsValidDate())
         guarantor.fields["birth_date"] = tempDate.OutDate(global_mode);

      // Gender
      hl7_GenderToGenderID(gt1->valueForIndex(GT1Segment::GT1_GENDER)->value,
                           guarantor.fields["gender_id"]);
      // Marital Status
      hl7_MaritalStatusToMaritalID(gt1->valueForIndex(GT1Segment::GT1_MARITAL_STATUS)->value,
                                   guarantor.fields["marital_id"]);

      // Social Security Number
      hl7_SSNToNumeric(gt1->valueForIndex(GT1Segment::GT1_GUARANTOR_SSN)->value,
                       guarantor.fields["social_security_number"]);

      // Occupation
      guarantor.fields["occupation"] = gt1->valueForIndex(GT1Segment::GT1_EMPLOYMENT_STATUS)->value;

      // Main name for guarantor
      mainName["last_name"]   = gt1->findInValueTree("guarantor_name.last_name")->value;
      mainName["first_name"]  = gt1->findInValueTree("guarantor_name.first_name")->value;
      mainName["middle_name"] = gt1->findInValueTree("guarantor_name.mi")->value;
      mainName["suffix"]      = gt1->findInValueTree("guarantor_name.suffix")->value;

      // Contact Info
      //
      // The following information is available from the following PID fields:
      // * Home Address, Home Phone = address, phone_home
      // * Cell Phone = phone_home repeated value (if any)
      // * Work Phone = phone_business

      // Home address
      homeContact["line1"] = gt1->findInValueTree("address.line1")->value;
      homeContact["line2"] = gt1->findInValueTree("address.line2")->value;
      homeContact["city"]  = gt1->findInValueTree("address.city")->value;

      hl7_StateToStateID(gt1->findInValueTree("address.state")->value,
                         homeContact["state_id"]);
      hl7_NormalizeZip(gt1->findInValueTree("address.zip")->value,
                       homeContact["zip"]);

      // Lookup the county
      if(CountyByZip(db, homeContact["zip"], countyID))
         homeContact["county_id"] = countyID;

      // Home phone
      HL7Value *homeXPN   = gt1->valueForIndex(GT1Segment::GT1_PHONE_HOME);
      HL7Value *homeEquip = homeXPN->findInSubComponents("telecom_equip_type");
      HL7Value *homeEmail = homeXPN->findInSubComponents("email");
      HL7Value *homePhone = homeXPN->findInSubComponents("compound");

      homeContact["email_address"] = homeEmail->value;
      hl7_NormalizePhone(homePhone->value, homeContact["phone_number"]);

      // If telecom_equip_type has a secondary value of "CP", there is a cell
      // phone value provided.
      if(homeEquip->repeatValues && homeEquip->repeatValues->value == "CP")
      {
         HL7Value *cellEmail = homeEmail->repeatValues;
         HL7Value *cellPhone = homePhone->repeatValues;

         if(cellEmail)
            cellContact["email"] = cellEmail->value;

         if(cellPhone)
            hl7_NormalizePhone(cellPhone->value, cellContact["phone_number"]);
      }

      // Work address
      workContact["line1"] = gt1->findInValueTree("employer_address.line1")->value;
      workContact["line2"] = gt1->findInValueTree("employer_address.line2")->value;
      workContact["city"]  = gt1->findInValueTree("employer_address.city")->value;

      hl7_StateToStateID(gt1->findInValueTree("employer_address.state")->value,
                         workContact["state_id"]);
      hl7_NormalizeZip(gt1->findInValueTree("employer_address.zip")->value,
                       workContact["zip"]);

      // Lookup the county
      if(CountyByZip(db, workContact["zip"], countyID))
         workContact["county_id"] = countyID;

      // Work phone
      workContact["email_address"] = gt1->findInValueTree("phone_business.email")->value;
      hl7_NormalizePhone(gt1->findInValueTree("phone_business.compound")->value,
                         workContact["phone_number"]);
   }
}

//
// HL7Processor::processGTEmployer
//
// Extract and translate all data on the guarantor's employer.
//
void HL7Processor::processGTEmployer(HL7Person &gtEmployer, GT1Segment *gt1)
{
   string countyID;
   HL7Record &mainName    = gtEmployer.names[HL7Person::NAME_MAIN];
   HL7Record &workContact = gtEmployer.contactInfo[HL7Person::CONTACT_WORK];

   // Unlikely to be enough information to match against.
   if(gt1)
   {
      // Main name for employer
      mainName["last_name"]   = gt1->findInValueTree("employer_name.last_name")->value;
      mainName["first_name"]  = gt1->findInValueTree("employer_name.first_name")->value;
      mainName["middle_name"] = gt1->findInValueTree("employer_name.mi")->value;
      mainName["suffix"]      = gt1->findInValueTree("employer_name.suffix")->value;

      // Work address
      workContact["line1"] = gt1->findInValueTree("employer_address.line1")->value;
      workContact["line2"] = gt1->findInValueTree("employer_address.line2")->value;
      workContact["city"]  = gt1->findInValueTree("employer_address.city")->value;

      hl7_StateToStateID(gt1->findInValueTree("employer_address.state")->value,
                         workContact["state_id"]);
      hl7_NormalizeZip(gt1->findInValueTree("employer_address.zip")->value,
                       workContact["zip"]);

      // Lookup the county
      if(CountyByZip(db, workContact["zip"], countyID))
         workContact["county_id"] = countyID;

      // Work phone
      workContact["email_address"] = gt1->findInValueTree("employer_phone.email")->value;
      hl7_NormalizePhone(gt1->findInValueTree("employer_phone.compound")->value,
                         workContact["phone_number"]);
   }
}

//
// HL7Processor::processInsCompany
//
// Process the insurance company for one of the primary's insurance policies.
//
void HL7Processor::processInsCompany(HL7TaxEntity &entity, IN1Segment *in1)
{
   if(in1)
   {
      string countyID;

      // TODO: May or may not be the Prometheus ID originally imported, where
      // imported; needs verification.
      entity.fields["id"]   = in1->valueForIndex(IN1Segment::IN1_COMPANYID)->value;
      entity.fields["name"] = in1->findInValueTree("company_name.org_name")->value;

      entity.address["line1"] = in1->findInValueTree("company_address.line1")->value;
      entity.address["line2"] = in1->findInValueTree("company_address.line2")->value;
      entity.address["city"]  = in1->findInValueTree("company_address.city")->value;

      hl7_StateToStateID(in1->findInValueTree("company_address.state")->value,
                         entity.address["state_id"]);

      hl7_NormalizeZip(in1->findInValueTree("company_address.zip")->value,
                       entity.address["zip"]);

      // Lookup the county
      if(CountyByZip(db, entity.address["zip"], countyID))
         entity.address["county_id"] = countyID;

      hl7_NormalizePhone(in1->findInValueTree("company_phone.compound")->value,
                         entity.address["phone_number"]);

      entity.address["email_address"] = in1->findInValueTree("company_phone.email")->value;
   }
}

//
// HL7Processor::processPolicyHolder
//
// Process the policy holder for one of the primary's insurance policies.
//
void HL7Processor::processPolicyHolder(HL7Person &holder,
                                       IN1Segment *in1, IN2Segment *in2)
{
   string countyID;
   
   // Get references to subordinate records.
   HL7Record &mainName    = holder.names[HL7Person::NAME_MAIN];
   HL7Record &homeContact = holder.contactInfo[HL7Person::CONTACT_HOME];
   HL7Record &cellContact = holder.contactInfo[HL7Person::CONTACT_CELL];
   HL7Record &workContact = holder.contactInfo[HL7Person::CONTACT_WORK];

   if(in1)
   {
      Pdate tempDate;

      // Birth Date
      hl7_DTToPdate(in1->valueForIndex(IN1Segment::IN1_INSURED_DOB)->value, tempDate);
      if(tempDate.IsValidDate())
         holder.fields["birth_date"] = tempDate.OutDate(global_mode);

      // Gender
      hl7_GenderToGenderID(in1->valueForIndex(IN1Segment::IN1_INSURED_GENDER)->value,
                           holder.fields["gender_id"]);

      // Main name for policy holder
      mainName["last_name"]   = in1->findInValueTree("insured_name.last_name")->value;
      mainName["first_name"]  = in1->findInValueTree("insured_name.first_name")->value;
      mainName["middle_name"] = in1->findInValueTree("insured_name.mi")->value;
      mainName["suffix"]      = in1->findInValueTree("insured_name.suffix")->value;

      // Contact Info
      //
      // The following information is available from the following PID fields:
      // * Home Address = insured_address

      // Home address
      homeContact["line1"] = in1->findInValueTree("insured_address.line1")->value;
      homeContact["line2"] = in1->findInValueTree("insured_address.line2")->value;
      homeContact["city"]  = in1->findInValueTree("insured_address.city")->value;

      hl7_StateToStateID(in1->findInValueTree("insured_address.state")->value,
                         homeContact["state_id"]);
      hl7_NormalizeZip(in1->findInValueTree("insured_address.zip")->value,
                       homeContact["zip"]);

      // Lookup the county
      if(CountyByZip(db, homeContact["zip"], countyID))
         homeContact["county_id"] = countyID;

      // TODO/FIXME: not sure if we really want to put this here yet.
      // Occupation
      holder.fields["occupation"] = in1->valueForIndex(IN1Segment::IN1_INSURED_EMPLOYMENT_STATUS)->value;
   }

   if(in2)
   {
      // Social Security Number
      hl7_SSNToNumeric(in2->valueForIndex(IN2Segment::IN2_INSURED_SSN)->value,
                       holder.fields["social_security_number"]);

      // Home phone
      HL7Value *homeXPN   = in2->valueForIndex(IN2Segment::IN2_INSURED_HOME_PHONE);
      HL7Value *homeEquip = homeXPN->findInSubComponents("telecom_equip_type");
      HL7Value *homeEmail = homeXPN->findInSubComponents("email");
      HL7Value *homePhone = homeXPN->findInSubComponents("compound");

      homeContact["email_address"] = homeEmail->value;
      hl7_NormalizePhone(homePhone->value, homeContact["phone_number"]);

      // If telecom_equip_type has a secondary value of "CP", there is a cell
      // phone value provided.
      if(homeEquip->repeatValues && homeEquip->repeatValues->value == "CP")
      {
         HL7Value *cellEmail = homeEmail->repeatValues;
         HL7Value *cellPhone = homePhone->repeatValues;

         if(cellEmail)
            cellContact["email"] = cellEmail->value;

         if(cellPhone)
            hl7_NormalizePhone(cellPhone->value, cellContact["phone_number"]);
      }

      // Work phone
      workContact["email_address"] = in2->findInValueTree("insured_work_phone.email")->value;
      hl7_NormalizePhone(in2->findInValueTree("insured_work_phone.compound")->value,
                         workContact["phone_number"]);
   }
}

//
// HL7Processor::processInsurancePolicy
//
// Do all processing (including subordinate records) for one single insurance
// policy defined in the HL7 message. This includes:
//
// * Policy Holder (a person, possibly distinct from the primary update target)
// * Insurance Company (a tax entity)
// * Insurance Policy fields
// * Insurance Coverage fields
//
void HL7Processor::processInsurancePolicy(HL7InsurancePolicy &policy,
                                          IN1Segment *in1, IN2Segment *in2)
{
   // Process policy holder
   processPolicyHolder(policy.holder, in1, in2);

   // Process company
   processInsCompany(policy.company, in1);

   // Process policy fields
   if(in1)
   {
      Pdate tempDate1, tempDate2;

      policy.policy["tax_entity_id"] = in1->valueForIndex(IN1Segment::IN1_COMPANYID)->value;
      policy.policy["group_number"]  = in1->valueForIndex(IN1Segment::IN1_GROUP_NUMBER)->value;

      hl7_DTToPdate(in1->valueForIndex(IN1Segment::IN1_PLAN_EFFECTIVE_DATE)->value, tempDate1);
      if(tempDate1.IsValidDate())
         policy.policy["start_date"] = tempDate1.OutDate(global_mode);

      hl7_DTToPdate(in1->valueForIndex(IN1Segment::IN1_PLAN_EXPIRATION_DATE)->value, tempDate2);
      if(tempDate2.IsValidDate())
         policy.policy["stop_date"] = tempDate2.OutDate(global_mode);

      policy.policy["policy_number"] = in1->valueForIndex(IN1Segment::IN1_POLICY_NUMBER)->value;

      // NB: Policy holder ID must be determined by matching against the policy
      // holder person information in the matcher module...

      // NB: insurance_id (pointer to insurance_types) must be determined in the
      // matcher module by examining the company to which the insurance policy
      // is attached (ie. if Medicare/Medicaid, or private company).
   }

   // Process coverage fields
   policy.coverage["start_date"  ] = policy.policy["start_date"];
   policy.coverage["stop_date"   ] = policy.policy["stop_date"];
   policy.coverage["name_on_card"] =
      CollapseSpaces(policy.holder.names[HL7Person::NAME_MAIN]["first_name"]  + " " +
                     policy.holder.names[HL7Person::NAME_MAIN]["middle_name"] + " " +
                     policy.holder.names[HL7Person::NAME_MAIN]["last_name"]   + " " +
                     policy.holder.names[HL7Person::NAME_MAIN]["suffix"]);
}

//
// HL7Processor::processPolicies
//
// Process each insurance policy consecutively. There are (theoretically at
// least) an unlimited number of policies specified in the HL7 message, in the
// form of back-to-back IN1 and IN2 segments. I expect no more than three will
// be sent, but I'm not going to write the code to assume there are only three
// policies.
//
void HL7Processor::processPolicies(HL7Intermediate &im, HL7Parser::SegmentVector &sv)
{
   for(size_t i = 0; i < sv.size(); i++)
   {
      if(sv[i]->isInstanceOf(RTTI(IN1Segment)))
      {
         IN1Segment *in1 = static_cast<IN1Segment *>(sv[i]);
         IN2Segment *in2 = NULL;
         if(i+1 < sv.size() && sv[i+1]->isInstanceOf(RTTI(IN2Segment)))
            in2 = static_cast<IN2Segment *>(sv[i+1]);

         im.policies.push_back(HL7InsurancePolicy());
         processInsurancePolicy(im.policies.back(), in1, in2);
      }
   }
}

//
// FindSegmentInMsg
//
// Look for the first segment in a message segment vector that is an instance
// of the template parameter type.
//
template<typename T> T *FindSegmentInMsg(HL7Parser::SegmentVector &sv)
{
   for(size_t i = 0; i < sv.size(); i++)
   {
      if(sv[i]->isInstanceOf(RTTI(T)))
         return static_cast<T *>(sv[i]);
   }

   return NULL;
}

//
// HL7Processor::processMessage
//
// Process a single HL7Segment vector.
//
void HL7Processor::processMessage(HL7Parser::SegmentVector &sv)
{
   // First step: find all the segments we need.
   MSHSegment *mshSegment = FindSegmentInMsg<MSHSegment>(sv);
   EVNSegment *evnSegment = FindSegmentInMsg<EVNSegment>(sv);
   PIDSegment *pidSegment = FindSegmentInMsg<PIDSegment>(sv);
   NK1Segment *nk1Segment = FindSegmentInMsg<NK1Segment>(sv);
   GT1Segment *gt1Segment = FindSegmentInMsg<GT1Segment>(sv);

   // MSH header is required to be present, otherwise it's probably malformed
   if(!mshSegment)
   {
      DEBUGOUT(dbg_processor, "No MSH segment in message, ignoring");
      return;
   }

   // Start out examining the EVN segment event type - we are only interested in
   // event A08, demographics update.
   if(!evnSegment)
   {
      DEBUGOUT(dbg_processor, "No EVN segment in message, ignoring");
      return;
   }

   HL7Value *typeCode = evnSegment->valueForIndex(EVNSegment::EVN_TYPECODE);
   if(typeCode->value != "A08")
   {
      DEBUGOUT(dbg_processor, "EVN is not of type A08, ignoring");
      return;
   }

   // Get the message ID and timestamp
   string messageid = mshSegment->valueForIndex(MSHSegment::MSH_MESSAGECTRLID)->value;
   string timestamp = evnSegment->valueForIndex(EVNSegment::EVN_TIMESTAMP)->value;

   // Create a new intermediate
   HL7Intermediate *newIntermediate = new HL7Intermediate();
   intermediates.insert(IMapRow(timestamp, newIntermediate));

   newIntermediate->messageID        = messageid;
   newIntermediate->messageTimeStamp = timestamp;

   // get the data for the primary person
   processPrimaryPerson(newIntermediate->primaryPerson, pidSegment);

   // get the data for the primary person's next-of-kin
   processNextOfKin(newIntermediate->nextOfKin, nk1Segment);

   // get the data for the primary person's guarantor
   processGuarantor(newIntermediate->guarantor, gt1Segment);

   // get the data for the guarantor's employer
   processGTEmployer(newIntermediate->gtEmployer, gt1Segment);

   // process all insurance policies
   processPolicies(*newIntermediate, sv);
}

// Lookup tables needed by the message processing code
static const char *lookups[] =
{
   "gender_types",
   "marital_types",
   "race_types",
};

//
// HL7Processor::processMessages
//
// Transform the raw HL7 messages into an intermediate form against which the
// person matcher module will be able to do comparisons with data within the
// Prometheus database schema.
//
void HL7Processor::processMessages()
{
   PrometheusDB database; // Connection will close when this method returns.
   db = &database;

   if(!db->connect(IniFile::GetIniFile()))
   {
      DEBUGOUT(dbg_processor, "Could not connect to database, processing failed");
      return;
   }

   // Cache required lookup tables
   PrometheusLookups::LoadLookups(database, lookups, arrlen(lookups));

   // Load some lookups using custom SQL too.
   PrometheusLookups::LoadCustom(database, "state_types",
      "select id, abbreviation item from state_types where valid = 1");

   // Process all messages
   for(size_t i = 0; i < messages.size(); i++)
      processMessage(*messages[i]);

   // Dump the lookups we cached, as the matcher module may want to load them
   // differently (state_types in particular...)
   PrometheusLookups::PurgeLookups();

   db = NULL;
}

//
// HL7Processor::doProcessing
//
// This is the "main" procedure for this module. Instantiate an instance of
// HL7Processor, call this method, and then operate on the resulting reference.
// The HL7Processor will free all resources when it is destroyed.
//
HL7Processor::IntermediateMap &HL7Processor::doProcessing()
{
   scanForFiles();       // signed
   parseFiles();         // sealed
   processMessages();    // delivered

   return intermediates; // it's yours
}

// EOF



