/*

  HL7Daemon - HL7 Message Processing Core

  This is the real meat, where we'll organize HL7 messages and translate them
  into an intermediate form as close as possible to Prometheus' database
  schema. Then we'll send the stuff on to the person matcher.
  
*/

#ifndef HL7PROCESS_H__
#define HL7PROCESS_H__

#include "hl7segments.h"
#include "prometheusdb.h"

class fileinfo;

typedef pdb::stringmap HL7Record;
typedef pdb::vecmap    HL7RecordVec;

//
// HL7Person
//
// This class represents a single person (people table row in Prometheus) and
// all of the immediately subordinate records.
//
class HL7Person
{
public:
   // names enumeration
   enum
   {
      NAME_MAIN,       // main person name is at index 0

      NAME_NUMEXPECTED
   };

   // contact info enumeration
   enum
   {
      CONTACT_HOME, // home address and phone
      CONTACT_CELL, // cell phone
      CONTACT_WORK, // work address and phone

      CONTACT_NUMEXPECTED
   };

   HL7Record    fields;
   HL7RecordVec names;
   HL7RecordVec contactInfo;

   HL7Person() : fields(), names(), contactInfo()
   {
      names.resize(NAME_NUMEXPECTED);
      contactInfo.resize(CONTACT_NUMEXPECTED);
   }
};

//
// HL7TaxEntity
//
// This class represents an insurance company and all immediately subordinate
// records.
//
class HL7TaxEntity
{
public:
   HL7Record fields;
   HL7Record address;

   HL7TaxEntity() : fields(), address() {}
};

//
// HL7InsurancePolicy
//
// This class represents everything that is part of an insurance policy.
//
class HL7InsurancePolicy
{
public:
   HL7Person    holder;   // policy holder
   HL7TaxEntity company;  // insurance company entity and its address
   HL7Record    policy;   // insurance_policies fields
   HL7Record    coverage; // insurance_coverage fields; insert for both primary
                          // person and the policy holder (if distinct)

   HL7InsurancePolicy() : holder(), company(), policy(), coverage() {}
};

//
// HL7Intermediate
//
// A multimap of these structures with the message timestamps as the keys will
// be served up as the ultimate return value of the HL7Processor::processMessages
// method.
//
// This class will contain all of the extracted usable data from the HL7
// message, in a format that is suited to matching and insertion/update via
// PrometheusDB.
//
// Presence of information in this structure does not indicate an *obligation*
// to push it into the database. I have extracted as much info as possible;
// whether we want to use it all or not remains to be decided - jhaley
//
class HL7Intermediate
{
public:
   typedef std::vector<HL7InsurancePolicy> PolicyVec;

   // status enumeration (used by HL7Matcher)
   enum
   {
      STATUS_NORMAL,      // not examined yet by the matcher
      STATUS_OUTOFDATE,   // a newer message has outdated this one
      STATUS_ALREADYDONE, // determined to have already been applied
      STATUS_WILLMATCH,   // this one is ready for matching
      STATUS_CANTMATCH,   // something is wrong/missing, so this can't be used

      STATUS_NUMSTATUSES
   };

   string    messageTimeStamp; // message timestamp, to the millisecond
   string    messageID;        // unique message ID
   HL7Person primaryPerson;    // primary person who is the update target
   HL7Person nextOfKin;        // primary person's next-of-kin
   HL7Person guarantor;        // primary person's guarantor
   HL7Person gtEmployer;       // guarantor's employer
   PolicyVec policies;         // insurance policies
   int       status;           // intermediate status, during matching.

   HL7Intermediate()
      : messageTimeStamp(), messageID(), primaryPerson(), nextOfKin(),
        guarantor(), gtEmployer(), policies(), status(STATUS_NORMAL)
   {
   }
};

//
// HL7Processor
//
// This class does three major things:
// * Scans for all HL7 message files to parse, via the scanner module.
// * Parses the files, via the hl7parse module.
// * Processes the resulting HL7Segment vectors and returns a vector of
//   HL7Intermediate instances.
//
class HL7Processor
{
public:
   typedef std::multimap<std::string, std::string> TSFNMap;
   typedef std::vector<HL7Parser::SegmentVector *> MessageVec;
   typedef std::multimap<std::string, HL7Intermediate *>  IntermediateMap;
   typedef std::pair<std::string, HL7Intermediate *> IMapRow;

protected:
   TSFNMap         timestamp_filename_map; // map of timestamps to filenames
   MessageVec      messages;               // HL7 messages as read from file
   IntermediateMap intermediates;          // intermediate results

   PrometheusDB *db;                       // database connection

   void scanCallback(fileinfo &fi);

   void scanForFiles();     // scan for the files to parse
   void parseFiles();       // do the actual parsing

   void processPrimaryPerson(HL7Person &newPerson, PIDSegment *pid);
   void processNextOfKin(HL7Person &nextOfKin, NK1Segment *nk1);
   void processGuarantor(HL7Person &guarantor, GT1Segment *gt1);
   void processGTEmployer(HL7Person &gtEmployer, GT1Segment *gt1);
   void processInsCompany(HL7TaxEntity &entity, IN1Segment *in1);
   void processPolicyHolder(HL7Person &holder, IN1Segment *in1, IN2Segment *in2);
   void processInsurancePolicy(HL7InsurancePolicy &policy,
                               IN1Segment *in1, IN2Segment *in2);
   void processPolicies(HL7Intermediate &im, HL7Parser::SegmentVector &sv);

   void processMessage(HL7Parser::SegmentVector &sv);
   void processMessages();  // transform parsed output into intermediates

public:
   HL7Processor()
     : timestamp_filename_map(), messages(), intermediates(), db(NULL)
   {
   }
   ~HL7Processor();

   IntermediateMap &doProcessing();
};

#endif

// EOF

