/*

   HL7Segment Descendant Classes

   Here live the static fields for the HL7Segment descendants.
   These include instantiation of the RTTI proxies for each segment type, and
   the field array trees for expected fields in each segment.

*/

#include <cstdlib>
#include <cstring>
#include "hl7segments.h"
#include "utils.h"

//=============================================================================
//
// HL7 Standard Datatypes and Recurring Subsegment Layouts
// These are shared between segment types.
//

// CE - Coded Element
// Used for:
// PID::language (use field "text")
static HL7FieldInfo StdCodedElement[] =
{
   { 2, "identifier"     },
   { 2, "text"           },
   { 2, "name"           },
   { 2, "alternate_id"   },
   { 2, "alternate_text" },
   { 2, "alt_code_name"  },
   { 1, NULL             }
};

// XAD - Extended Address
static HL7FieldInfo StdFullAddress[] =
{
   { 2, "line1" },
   { 2, "line2" },
   { 2, "city"  },
   { 2, "state" },
   { 2, "zip"   },
   { 1, NULL    }
};

// XCN - Extended Composite ID/Name
static HL7FieldInfo StdExtCompPName[] =
{
   // PM doesn't seem to ever provide these for our site, despite appearing in
   // the spec. Too bad. We'll expect them anyway in case they ever show up.
   { 2, "id_abbreviation" },
   { 2, "last_name"       },
   { 2, "first_name"      },
   { 2, "mi"              },
   { 2, "suffix"          },
   { 2, "prefix"          },
   { 2, "degree"          },
   { 2, "source_table"    },
   { 2, "assign_auth"     },
   { 2, "name_type"       },
   { 2, "check_digit"     },
   { 2, "check_scheme"    },
   { 2, "id_type"         },
   { 2, "assign_fac_id"   },
   { 1, NULL              }
};

// XON - Extended Organization Composite ID/Name
static HL7FieldInfo StdExtOrgCompName[] =
{
   { 2, "org_name"      },
   { 2, "name_type"     },
   { 2, "id_number"     },
   { 2, "check_digit"   },
   { 2, "check_scheme"  },
   { 2, "assign_auth"   },
   { 2, "id_type"       },
   { 2, "assign_fac_id" },
   { 1, NULL            }
};

// XPN - Extended Person Name
static HL7FieldInfo StdFullName[] =
{
   { 2, "last_name"  },
   { 2, "first_name" },
   { 2, "mi"         },
   { 2, "suffix"     },
   { 1, NULL         }
};

// XTN - Extended Communications Number
static HL7FieldInfo StdExtCommsNumber[] =
{
   { 2, "compound"           }, // (???)???-????X????? - easier to use below.
   { 2, "telecom_use_code"   },
   { 2, "telecom_equip_type" }, // PH for PHONE, CP for CELL
   { 2, "email"              }, // WILL be valid if recorded in PM!
   { 2, "country_code"       },
   { 2, "area_code"          },
   { 2, "phone_number"       },
   { 2, "extension"          },
   { 2, "any_text"           },
   { 1, NULL                 }
};

//=============================================================================
//
// MSH - Message Segment Header
//

IMPLEMENT_MSG_TYPE(MSHSegment, "MSH")

// MSH::Sending_Application
static HL7FieldInfo MSHSendingAppFields[] =
{
   { 2, "appname"    },
   { 2, "company"    },
   { 2, "identifier" },
   { 1, NULL         }
};

// MSH::Sending_Facility
static HL7FieldInfo MSHSendingFacFields[] =
{
   { 2, "facility_name" },
   { 2, "app_name"      },
   { 1, NULL            }
};

// MSH::Receiving_Application
static HL7FieldInfo MSHRecvAppFields[] =
{
   { 2, "appname1" },
   { 2, "appname2" },
   { 1, NULL       }
};

// MSH::Message_Type
static HL7FieldInfo MSHMsgTypeFields[] =
{
   { 2, "message_type" },
   { 2, "event_type"   },
   { 1, NULL           }
};

HL7FieldInfo MSHSegment::ExpectedFields[MSH_NUMFIELDS] =
{
   { 1, "encoding_chars"     },
   { 1, "sending_app",       MSHSendingAppFields, arrlen(MSHSendingAppFields) },
   { 1, "sending_facility",  MSHSendingFacFields, arrlen(MSHSendingFacFields) },
   { 1, "receiving_app",     MSHRecvAppFields,    arrlen(MSHRecvAppFields)    },
   { 1, "receiving_facility" },
   { 1, "timestamp"          },
   { 1, "security"           },
   { 1, "message_type",      MSHMsgTypeFields,    arrlen(MSHMsgTypeFields)    },
   { 1, "msg_ctrl_id"        },
   { 1, "msg_processing_id"  },
   { 1, "version_id"         },
   { 1, "sequence_num"       },
   { 1, "continuation_ptr"   },
   { 1, "accept_ack_type"    },
   { 1, "app_ack_type"       },
   { 1, "country_code"       },
   { 1, "charset"            },
   { 1, "language"           },
   { 0, NULL                 }
};

//=============================================================================
//
// EVN - Event Type Segment
//

IMPLEMENT_MSG_TYPE(EVNSegment, "EVN")


HL7FieldInfo EVNSegment::ExpectedFields[EVN_NUMFIELDS] =
{
   { 1, "type_code"      },
   { 1, "timestamp"      },
   { 1, "planned_stamp"  },
   { 1, "reason_code"    },
   { 1, "operator_id",   StdExtCompPName, arrlen(StdExtCompPName) },
   { 1, "event_occurred" },
   { 0, NULL             }
};

//=============================================================================
//
// PID - Patient ID Segment
//

IMPLEMENT_MSG_TYPE(PIDSegment, "PID")

HL7FieldInfo PIDSegment::ExpectedFields[PID_NUMFIELDS] =
{
   { 1, "set_id"          },
   { 1, "external_id"     },
   { 1, "internal_id"     }, // this will match the Prometheus ID if they were imported initially!
   { 1, "alternate_id"    },
   { 1, "patient_name",   StdFullName,       arrlen(StdFullName)       },
   { 1, "mother_mname",   StdFullName,       arrlen(StdFullName)       },
   { 1, "date_of_birth"   },
   { 1, "gender"          },
   { 1, "patient_alias",  StdFullName,       arrlen(StdFullName)       },
   { 1, "race"            },
   { 1, "pat_address",    StdFullAddress,    arrlen(StdFullAddress)    },
   { 1, "county_code"     },
   { 1, "phone_home",     StdExtCommsNumber, arrlen(StdExtCommsNumber) },
   { 1, "phone_business", StdExtCommsNumber, arrlen(StdExtCommsNumber) },
   { 1, "language",       StdCodedElement,   arrlen(StdCodedElement)   },
   { 1, "marital_status"  }, // same as format that was used for export.
   { 1, "region"          },
   { 1, "account_number"  }, // this will match the Prometheus ID if they were imported initially!
   { 1, "patient_ssn"     }, // xxx-xx-xxxx - extremely useful to have.
   { 1, "drivers_license" },
   { 1, "mother_id"       },
   { 1, "ethnicity"       }, // spec says NOT USED, but this will be valid! CCHIT requirement.
   { 1, "birth_place"     },
   { 1, "multbirth_indic" },
   { 1, "birth_order"     },
   { 1, "citizenship"     },
   { 1, "vet_mil_status"  },
   { 1, "nationality"     },
   { 1, "death_timestamp" },
   { 1, "death_indicator" }, // spec says NOT USED, but seems to be sent!
   { 0, NULL              }
};

//=============================================================================
//
// NK1 - Next-of-Kin Segment
//

IMPLEMENT_MSG_TYPE(NK1Segment, "NK1")

HL7FieldInfo NK1Segment::ExpectedFields[NK1_NUMFIELDS] =
{
   { 1, "set_id"            },
   { 1, "name",             StdFullName,       arrlen(StdFullName)       },
   { 1, "relationship"      }, // spec incorrectly says NOT USED
   { 1, "address",          StdFullAddress,    arrlen(StdFullAddress)    },
   { 1, "phone_number",     StdExtCommsNumber, arrlen(StdExtCommsNumber) },
   { 1, "phone_business",   StdExtCommsNumber, arrlen(StdExtCommsNumber) },
   { 1, "contact_role"      },
   { 1, "start_date"        },
   { 1, "end_date"          },
   { 1, "job_title"         },
   { 1, "job_code"          },
   { 1, "job_employeenum"   },
   { 1, "org_name",         StdExtOrgCompName, arrlen(StdExtOrgCompName) },
   { 1, "marital_status"    },
   { 1, "admin_gender"      },
   { 1, "birth_timestamp"   },
   { 1, "living_dependency" },
   { 1, "ambulatory_status" },
   { 1, "citizenship"       },
   { 1, "language"          },
   { 1, "livingarrangement" },
   { 1, "publicity_code"    },
   { 1, "protection_indic"  },
   { 1, "student_indic"     },
   { 1, "religion"          },
   { 1, "mother_mname",     StdFullName,       arrlen(StdFullName)       },
   { 1, "nationality"       },
   { 1, "ethnicity"         },
   { 1, "contact_reason"    },
   { 1, "contact_name",     StdFullName,       arrlen(StdFullName)       },
   { 1, "contact_phone",    StdExtCommsNumber, arrlen(StdExtCommsNumber) },
   { 1, "contact_address",  StdFullAddress,    arrlen(StdFullAddress)    },
   { 1, "nk_identifier"     },
   { 1, "job_status"        },
   { 1, "race"              },
   { 1, "handicap"          },
   { 1, "contact_ssn"       },
   { 0, NULL                }
};

//=============================================================================
//
// PV1 - Patient Visit Segment
//

IMPLEMENT_MSG_TYPE(PV1Segment, "PV1")

HL7FieldInfo PV1Segment::ExpectedFields[PV1_NUMFIELDS] =
{
   { 1, "set_id"               },
   { 1, "patient_class"        },
   { 1, "assigned_location"    },
   { 1, "admission_type"       },
   { 1, "preadmit_num"         },
   { 1, "prior_location"       },
   { 1, "attending_doctor",    StdExtCompPName, arrlen(StdExtCompPName) }, // NB: different from IHS!
   { 1, "referring_doctor",    StdExtCompPName, arrlen(StdExtCompPName) },
   { 1, "consulting_doctor",   StdExtCompPName, arrlen(StdExtCompPName) },
   { 1, "hospital_service"     },
   { 1, "temp_location"        },
   { 1, "preadmit_test_indic"  },
   { 1, "readmission_indic"    },
   { 1, "admit_source"         },
   { 1, "ambulatory_status"    },
   { 1, "vip_indicator"        },
   { 1, "admitting_doctor",    StdExtCompPName, arrlen(StdExtCompPName) },
   { 1, "patient_type"         }, // spec says NOT USED, but contains 'A' !
   { 1, "visit_number"         },
   { 1, "financial_class"      },
   { 1, "charge_price_indic"   },
   { 1, "courtesy_code"        },
   { 1, "credit_rating"        },
   { 1, "contract_code"        },
   { 1, "contract_eff_date"    },
   { 1, "contract_amount"      },
   { 1, "contract_period"      },
   { 1, "interest_code"        },
   { 1, "xfer_bad_debt_code"   },
   { 1, "xfer_bad_debt_date"   },
   { 1, "bad_debt_agency_code" },
   { 1, "bad_debt_xfer_amount" },
   { 1, "bad_debt_recov_amt"   },
   { 1, "delete_acct_indic"    },
   { 1, "delete_acct_date"     },
   { 1, "discharge_dispstn"    },
   { 1, "discharged_to_loc"    },
   { 1, "diet_type"            },
   { 1, "servicing_facility"   },
   { 1, "bed_status"           },
   { 1, "account_status"       },
   { 1, "pending_location"     },
   { 1, "prior_temp_location"  },
   { 1, "admit_timestamp"      }, // spec says NOT USED, but contains '0' !
   { 1, "discharge_timestamp"  },
   { 1, "cur_patient_balance"  },
   { 1, "total_charge"         },
   { 1, "total_adjustment"     },
   { 1, "total_payments"       },
   { 1, "alternate_visit_id"   }, // appointment id for schedule messages
   { 1, "visit_indicator"      },
   { 1, "other_hc_provider",   StdExtCompPName, arrlen(StdExtCompPName) },
   { 0, NULL                   }
};

//=============================================================================
//
// GT1 - Guarantor Information Segment
//
// This would translate into the policy holder for the insurance policy in
// Prometheus, *if* we're capable of doing reliable matching with the
// information at hand - this remains to be seen.
//

IMPLEMENT_MSG_TYPE(GT1Segment, "GT1")

HL7FieldInfo GT1Segment::ExpectedFields[GT1_NUMFIELDS] =
{
   { 1, "set_id"             },
   { 1, "guarantor_number"   },
   { 1, "guarantor_name",    StdFullName,       arrlen(StdFullName)       },
   { 1, "spouse_name",       StdFullName,       arrlen(StdFullName)       },
   { 1, "address",           StdFullAddress,    arrlen(StdFullAddress)    },
   { 1, "phone_home",        StdExtCommsNumber, arrlen(StdExtCommsNumber) },
   { 1, "phone_business",    StdExtCommsNumber, arrlen(StdExtCommsNumber) },
   { 1, "birth_timestamp"    },
   { 1, "gender"             },
   { 1, "guarantor_type"     },
   { 1, "relationship"       },
   { 1, "guarantor_ssn"      },
   { 1, "begin_date"         },
   { 1, "end_date"           },
   { 1, "priority"           },
   { 1, "employer_name",     StdFullName,       arrlen(StdFullName)       },
   { 1, "employer_address",  StdFullAddress,    arrlen(StdFullAddress)    },
   { 1, "employer_phone",    StdExtCommsNumber, arrlen(StdExtCommsNumber) },
   { 1, "employee_id_num"    },
   { 1, "employment_status"  },
   { 1, "organization",      StdExtOrgCompName, arrlen(StdExtOrgCompName) },
   { 1, "billing_hold_flag"  },
   { 1, "credit_rating"      },
   { 1, "death_timestamp"    },
   { 1, "death_flag"         },
   { 1, "charge_adjust_code" },
   { 1, "hh_annual_income"   },
   { 1, "household_size"     },
   { 1, "employer_id_num"    },
   { 1, "marital_status"     },
   { 1, "hire_eff_date"      },
   { 1, "employ_stop_date"   },
   { 1, "living_dependency"  },
   { 1, "ambulatory_status"  },
   { 1, "citizenship"        },
   { 1, "primary_language"   },
   { 1, "living_arrangement" },
   { 1, "publicity_indic"    },
   { 1, "protection_indic"   },
   { 1, "student_indic"      },
   { 1, "religion"           },
   { 1, "mother_mname",      StdFullName,       arrlen(StdFullName)       },
   { 1, "nationality"        },
   { 1, "ethnicity"          },
   { 1, "contact_name",      StdFullName,       arrlen(StdFullName)       },
   { 1, "contact_phone",     StdExtCommsNumber, arrlen(StdExtCommsNumber) },
   { 1, "contact_reason"     },
   { 1, "contact_relation"   },
   { 1, "job_title"          }, // NB: spec unclear if self or contact...
   { 1, "job_code"           }, // NB: spec unclear if self or contact...
   { 1, "employer_org_name", StdExtOrgCompName, arrlen(StdExtOrgCompName) },
   { 1, "handicap"           },
   { 1, "job_status"         }, // NB: spec unclear if self or contact...
   { 1, "financial_class"    },
   { 1, "race"               },
   { 0, NULL                 }
};

//=============================================================================
//
// IN1 - Insurance Information Segment 1
//

IMPLEMENT_MSG_TYPE(IN1Segment, "IN1")

HL7FieldInfo IN1Segment::ExpectedFields[IN1_NUMFIELDS] =
{
   { 1, "set_id"                 },
   { 1, "plan_id"                },
   { 1, "company_id"             },
   { 1, "company_name",          StdExtOrgCompName, arrlen(StdExtOrgCompName) },
   { 1, "company_address",       StdFullAddress,    arrlen(StdFullAddress)    },
   { 1, "company_contact",       StdFullName,       arrlen(StdFullName)       },
   { 1, "company_phone",         StdExtCommsNumber, arrlen(StdExtCommsNumber) },
   { 1, "group_number"           },
   { 1, "group_name",            StdExtOrgCompName, arrlen(StdExtOrgCompName) },
   { 1, "group_employer_id"      },
   { 1, "group_employer_name",   StdExtOrgCompName, arrlen(StdExtOrgCompName) },
   { 1, "plan_effective_date"    },
   { 1, "plan_expiration_date"   },
   { 1, "authorization_info"     },
   { 1, "plan_type"              },
   { 1, "insured_name",          StdFullName,       arrlen(StdFullName)       },
   { 1, "insured_relation"       },
   { 1, "insured_dob"            },
   { 1, "insured_address",       StdFullAddress,    arrlen(StdFullAddress)    },
   { 1, "assignment_of_benefits" },
   { 1, "coord_of_benefits"      },
   { 1, "cob_priority"           },
   { 1, "notice_of_admit_flag"   },
   { 1, "notice_of_admit_date"   },
   { 1, "rpt_eligibility_flag"   },
   { 1, "rpt_eligibility_date"   },
   { 1, "release_info_code"      },
   { 1, "preadmit_certification" },
   { 1, "verification_timestamp" },
   { 1, "verification_by",       StdFullName,       arrlen(StdFullName)       },
   { 1, "type_of_agreement_code" },
   { 1, "billing_status"         },
   { 1, "lifetime_reserve_days"  },
   { 1, "delay_before_lr_day"    },
   { 1, "company_plan_code"      },
   { 1, "policy_number"          },
   { 1, "policy_deductable"      },
   { 1, "policy_limit_amount"    },
   { 1, "policy_limit_days"      },
   { 1, "room_rate_semiprivate"  },
   { 1, "room_rate_private"      },
   { 1, "insured_employ_status"  },
   { 1, "insured_gender"         },
   { 1, "insured_employer_addr", StdFullAddress,    arrlen(StdFullAddress)    },
   { 1, "verification_status"    },
   { 1, "prior_plan_id"          },
   { 1, "coverage_type"          },
   { 1, "handicap"               },
   { 1, "insured_id_number"      },
   { 0, NULL                     }
};

//=============================================================================
//
// IN2 - Insurance Information Segment 2
//

IMPLEMENT_MSG_TYPE(IN2Segment, "IN2")

HL7FieldInfo IN2Segment::ExpectedFields[IN2_NUMFIELDS] =
{
   { 1, "insured_employee_id"    },
   { 1, "insured_ssn"            },
   { 1, "insured_employer_name", StdExtCompPName,   arrlen(StdExtCompPName)   },
   { 1, "employer_info_data"     },
   { 1, "mail_claim_party"       },
   { 1, "medicare_card_number"   },
   { 1, "medicaid_case_name",    StdFullName,       arrlen(StdFullName)       },
   { 1, "medicaid_case_number"   },
   { 1, "champus_sponsor_name",  StdFullName,       arrlen(StdFullName)       },
   { 1, "champus_id_number"      },
   { 1, "dep_champus_recipient"  },
   { 1, "champus_organization"   },
   { 1, "champus_station"        },
   { 1, "champus_service"        },
   { 1, "champus_rank_grade"     },
   { 1, "champus_status"         },
   { 1, "champus_retire_date"    },
   { 1, "champus_na_certonfile"  },
   { 1, "baby_coverage"          },
   { 1, "combine_baby_bill"      },
   { 1, "blood_deductible"       },
   { 1, "sp_coverage_aprv_name", StdFullName,       arrlen(StdFullName)       },
   { 1, "sp_coverage_aprv_title" },
   { 1, "noncovered_ins_code"    },
   { 1, "payer_id"               },
   { 1, "payer_subscriber_id"    },
   { 1, "eligibility_source"     },
   { 1, "room_coverage_type_amt" },
   { 1, "policy_type_amt"        },
   { 1, "daily_deductible"       },
   { 1, "living_dependency"      },
   { 1, "ambulatory_status"      },
   { 1, "citizenship"            },
   { 1, "primary_language"       },
   { 1, "living_arrangement"     },
   { 1, "publicity_indicator"    },
   { 1, "protection_indicator"   },
   { 1, "student_indicator"      },
   { 1, "religion"               },
   { 1, "mother_maiden_name",    StdFullName,       arrlen(StdFullName)       },
   { 1, "nationality"            },
   { 1, "ethnicity"              },
   { 1, "marital_status"         },
   { 1, "insured_emp_start_date" },
   { 1, "insured_emp_stop_date"  },
   { 1, "job_title"              },
   { 1, "job_code_class"         },
   { 1, "job_status"             },
   { 1, "emp_contact_name",      StdFullName,       arrlen(StdFullName)       },
   { 1, "emp_contact_phone",     StdExtCommsNumber, arrlen(StdExtCommsNumber) },
   { 1, "emp_contact_reason"     },
   { 1, "ins_contact_pers_name", StdFullName,       arrlen(StdFullName)       },
   { 1, "ins_contact_per_phone", StdExtCommsNumber, arrlen(StdExtCommsNumber) },
   { 1, "ins_contact_p_reason"   },
   { 1, "rel_to_patient_start"   },
   { 1, "rel_to_patient_stop"    },
   { 1, "inscomp_contact_reason" },
   { 1, "inscomp_contact_phone", StdExtCommsNumber, arrlen(StdExtCommsNumber) },
   { 1, "policy_scope"           },
   { 1, "policy_source"          },
   { 1, "patient_member_number"  },
   { 1, "guarantor_relationship" },
   { 1, "insured_home_phone",    StdExtCommsNumber, arrlen(StdExtCommsNumber) },
   { 1, "insured_work_phone",    StdExtCommsNumber, arrlen(StdExtCommsNumber) },
   { 1, "mil_handicap_program"   },
   { 1, "suspend_flag"           },
   { 1, "copay_limit_flag"       },
   { 1, "stoploss_limit_flag"    },
   { 1, "ins_org_name",          StdExtOrgCompName, arrlen(StdExtOrgCompName) },
   { 1, "ins_employer_org_name", StdExtOrgCompName, arrlen(StdExtOrgCompName) },
   { 1, "race"                   },
   { 1, "hcfa_pat_rel_2_insured" },
   { 0, NULL                     }
};

//=============================================================================
//
// ZAL - Unknown custom PM/EHR segment. This is undocumented.
//

IMPLEMENT_MSG_TYPE(ZALSegment, "ZAL")

HL7FieldInfo ZALSegment::ExpectedFields[ZAL_NUMFIELDS] =
{
   { 1, "unknown_date" }, // always seems to be ~6-12 months in the future
   { 0, NULL           }
};

// EOF


