//---------------------------------------------------------------------------

#include <cstdlib>
#include <iostream>
#include "windows.h"
using namespace std;

#include "argvs.h"
#include "globaldefines.h"
#include "inifile.h"
#include "sqlmanager.h"
#include "utils.h"

#pragma hdrstop
#include "util.h"

//=============================================================================
//
// Initialization Routines
//

static HANDLE stopAppEvent;
static BOOL WINAPI ConsoleHandler(DWORD CEvent);

//
// DoCommandLineParams
//
// Process the command line parameters for the program.
//
static void DoCommandLineParams()
{
   const char *temp;

   temp = globalArgMgr->getParamTo("-debugflags");
   if(temp)
      hl7_SetDebugFlags(temp);
}

//
// ReadIniFile
//
// Read the initialization/configuration file
//
static bool ReadIniFile()
{
   IniFile &iniFile = IniFile::GetIniFile();

   return iniFile.loadOptionsFromFile(INI_FILE_NAME);
}

//
// InitProgramShared
//
// Shared initialization code.
//
static bool InitProgramShared()
{
   // Process command-line parameters
   DoCommandLineParams();

   // Initialization file
   if(!ReadIniFile())
   {
      DEBUGOUT(dbg_error, "INI file load failed!");
      return false;
   }

   // SQL Manager
   if(!SQLManager::LoadScripts())
      return false;

   return true;
}

//
// InitProgramApp
//
// Run initialization processes specific to application mode, and then run
// shared initialization.
//
static bool InitProgramApp()
{
   bool all_is_well = true;

   // Setup console ctrl event handler
   if(!SetConsoleCtrlHandler(ConsoleHandler, TRUE))
      DEBUGOUT(dbg_error, "WARNING: Could not set console handler");

   // Do shared setup
   all_is_well = InitProgramShared();

   return all_is_well;
}

//
// InitProgramSvc
//
// Initialize up as a service.
//
static bool InitProgramSvc()
{
   // Send debug output to file
   hl7_DebugInitLog(true);

   return InitProgramShared();
}

//=============================================================================
//
// Application Main Loop
//

//
// ConsoleHandler
//
// Win32 console exception handler which is setup during initialization only
// when running as an application. This will get called whenever an important
// event is caught by the application's console window.
//
static BOOL WINAPI ConsoleHandler(DWORD CEvent)
{
   switch(CEvent)
   {
   case CTRL_CLOSE_EVENT:
   case CTRL_BREAK_EVENT:
   case CTRL_C_EVENT:
      SetEvent(stopAppEvent); // signal the stop event
      break;
   }

   return TRUE;
}

static void ExecuteMainLoopApp()
{
   IniFile::IniMap &ini = IniFile::GetIniOptions();
   unsigned long sleepPeriod;

   sleepPeriod = static_cast<unsigned long>(StringToInt(ini["scanner"]["frequency"]));

   stopAppEvent = CreateEvent(0, FALSE, FALSE, 0);

   do
   {
      try
      {
         // TODO: scan files and push demographics
         DEBUGOUT(dbg_main, "peekaboo!");
      }
      catch(...)
      {
         // Some exception occurred, but we'll keep trying to run.
         DEBUGOUT(dbg_error, "Unknown exception caught in ExecuteMainLoopApp");
      }
   }
   while(WaitForSingleObject(stopAppEvent, sleepPeriod) == WAIT_TIMEOUT);
}

//
// AppMain
//
// Main program logic when running as an application.
//
static void AppMain()
{
   // Do initialization tasks.
   if(InitProgramApp())
   {
      // Run the daemon. Ideally, this never returns.
      ExecuteMainLoopApp();
   }

   // TODO: Application shutdown here.
   Beep(500, 300);
   ExitProcess(0);
}

//=============================================================================
//
// Service Facilities and Main Loop
//

static SERVICE_STATUS ServiceStatus;
static SERVICE_STATUS_HANDLE hStatus;
static HANDLE stopServiceEvent;

//
// ExecuteMainLoopSvc
//
static void ExecuteMainLoopSvc()
{
   IniFile::IniMap &ini = IniFile::GetIniOptions();
   unsigned long sleepPeriod;

   sleepPeriod = static_cast<unsigned long>(StringToInt(ini["scanner"]["frequency"]));

   do
   {
      try
      {
         // TODO: scan files and push demographics here
      }
      catch(...)
      {
         // Some exception occurred, but we'll keep trying to run.
         DEBUGOUT(dbg_error, "Unknown exception caught in ExecuteMainLoopSvc");
      }
   }
   while(WaitForSingleObject(stopServiceEvent, sleepPeriod) == WAIT_TIMEOUT);
}

//
// SvcControlHandler
//
// Service control message handler.
// Note that this routine runs asynchronously; don't call into the rest of the
// application's code from here unless you want to create race conditions!
//
static void WINAPI SvcControlHandler(DWORD request)
{
   switch(request)
   {
   case SERVICE_CONTROL_STOP:
   case SERVICE_CONTROL_SHUTDOWN:
      ServiceStatus.dwCurrentState = SERVICE_STOP_PENDING;
      SetServiceStatus(hStatus, &ServiceStatus);
      SetEvent(stopServiceEvent); // signal the stop event
      break;
   default:
      SetServiceStatus(hStatus, &ServiceStatus); // report current status
      break;
   }
}

//
// ServiceMain
//
// Service handler method for ASTHL7Daemon service
//
static void WINAPI ServiceMain(unsigned long argc, char **argv)
{
   ArgumentManager svcArgMgr(argc, const_cast<const char *const *>(argv));
   bool error;

   globalArgMgr = &svcArgMgr;

   ServiceStatus.dwServiceType      = SERVICE_WIN32;
   ServiceStatus.dwCurrentState     = SERVICE_STOPPED;
   ServiceStatus.dwControlsAccepted = 0;
   ServiceStatus.dwWin32ExitCode    = NO_ERROR;
   ServiceStatus.dwCheckPoint       = 0;
   ServiceStatus.dwWaitHint         = 0;

   ServiceStatus.dwServiceSpecificExitCode = NO_ERROR;

   hStatus = RegisterServiceCtrlHandler("ASTHL7Daemon", SvcControlHandler);

   if(hStatus == (SERVICE_STATUS_HANDLE)0)
      return; // registering of the control handler failed

   error = InitProgramSvc();
   if(error)
   {
      // initialization failed
      ServiceStatus.dwCurrentState  = SERVICE_STOPPED;
      ServiceStatus.dwWin32ExitCode = -1;
      SetServiceStatus(hStatus, &ServiceStatus);
   }
   else
   {
      ServiceStatus.dwCurrentState = SERVICE_START_PENDING;
      SetServiceStatus(hStatus, &ServiceStatus);

      // create waitable event for synchronization
      stopServiceEvent = CreateEvent(0, FALSE, FALSE, 0);

      // now running
      ServiceStatus.dwControlsAccepted = SERVICE_ACCEPT_STOP | SERVICE_ACCEPT_SHUTDOWN;
      ServiceStatus.dwCurrentState = SERVICE_RUNNING;
      SetServiceStatus(hStatus, &ServiceStatus); 

      ExecuteMainLoopSvc();

      // service wants to stop.
      ServiceStatus.dwCurrentState = SERVICE_STOP_PENDING;
      SetServiceStatus(hStatus, &ServiceStatus);

      // do cleanup
      CloseHandle(stopServiceEvent);
      stopServiceEvent = NULL;

      // service is now stopped
      ServiceStatus.dwControlsAccepted = 0;
      ServiceStatus.dwCurrentState = SERVICE_STOPPED;
      SetServiceStatus(hStatus, &ServiceStatus);
   }
}

// Service Dispatch Table
static SERVICE_TABLE_ENTRY ServiceTable[2] =
{
   { "ASTHL7Daemon", ServiceMain },
   { NULL,           NULL        }
};

//=============================================================================
//
// MAIN PROGRAM
//
// Either start up the Win32 service, or run as an application for testing and
// debugging purposes.
//

int main(int argc, const char *argv[])
{
   ArgumentManager argMgr(argc, argv);

   // Enable global access to command line arguments
   globalArgMgr = &argMgr;

   if(globalArgMgr->checkParam("-app"))
      AppMain();
   else
      StartServiceCtrlDispatcher(ServiceTable);

   return 0;
}
//---------------------------------------------------------------------------
