/*

   HL7Daemon - Prometheus database connection manager

   I'm hiding as much of the IBDatabase/VCL nastiness as possible in here by
   wrapping sqlLib with some simple-to-use objects that employ the common
   pImpl idiom for concealing implementation details.
   
   Should make the rest of the code a lot cleaner and more stable:
   * All methods are exception safe (no need for logic-clouding try/catch).
   * Memory management is automated with destructors for stack-allocated
     objects, whereas the VCL components hidden here can only be heap allocated.
   * Common errors are checked for and worked around.
   * Methods take reference params and return object results into references
     as much as is practical, to improve efficiency of user code (sqlLib is
     still subject to the inefficient passing of arguments and results,
     however).

   If I had the time, importing this into Prometheus itself would probably be
   worth it for these benefits.
*/

#include <iostream>

#include "inifile.h"
#include "prometheusdb.h"
#include "utils.h"

// ideally, this is the only place this will be included.
#pragma hdrstop
#include "sqlLib.h"

//=============================================================================
//
// PrometheusDB
//

//
// Private implementation details for PrometheusDB
//
class PrometheusDBPimpl
{
public:
   TIBDatabase *db;

   PrometheusDBPimpl()
   {
      try
      {
         db = new TIBDatabase(NULL);
      }
      catch(...)
      {
         DEBUGOUT(dbg_database, "Instantiation of the TIBDatabase instance failed!");
      }
   }

   ~PrometheusDBPimpl()
   {
      if(db)
      {
         try
         {
            if(db->TestConnected()) // If still connected, close it now.
              db->Close();
         }
         catch(...)
         {
            DEBUGOUT(dbg_database, "Closing the TIBDatabase instance failed!");
         }

         try
         {
            delete db; // Even if that failed, destroy the database component.
         }
         catch(...)
         {
            DEBUGOUT(dbg_database, "Destruction of the TIBDatabase instance failed!");
         }

         db = NULL; // No matter what, we're only trying this stuff once.
      }
   }
};

//
// PrometheusDB Constructor
//
PrometheusDB::PrometheusDB()
{
   pImpl = new PrometheusDBPimpl;
}

//
// PrometheusDB Destructor
//
PrometheusDB::~PrometheusDB()
{
   if(pImpl)
   {
      delete pImpl;
      pImpl = NULL;
   }
}

//
// PrometheusDB::connect
//
// Connect to the database, if not already connected.
//
bool PrometheusDB::connect(IniFile &iniFile)
{
   bool result = false;
   TIBDatabase *db = pImpl->db;
   
   IniFile::IniMap &ini = iniFile.getIniOptions();

   try
   {
      if(db)
      {
         if(db->TestConnected())
            result = true;       // Already connected.
         else
         {
            // Establish a new connection
            result = ConnectToDatabase(db,
                                       ini["database"]["db"],
                                       ini["database"]["user"],
                                       ini["database"]["password"]);
         }
      }
   }
   catch(...)
   {
      DEBUGOUT(dbg_database, "Database connection failed due to an unknown exception");
   }

   return result;
}

//
// PrometheusDB::disconnect
//
// Disconnect from the database, if it's connected. It can be reconnected if
// necessary by calling connect() again. This invalidates any open datasets or
// transactions.
//
void PrometheusDB::disconnect()
{
   TIBDatabase *db = pImpl->db;

   try
   {
      if(db && db->TestConnected())
         db->Close();
   }
   catch(...)
   {
      DEBUGOUT(dbg_database, "Failed trying to close database connection");
   }
}

//
// PrometheusDB::isConnected
//
// Returns true or false regarding the database's connection state.
//
bool PrometheusDB::isConnected()
{
   bool result = false;

   try
   {
      if(pImpl->db)
         result = pImpl->db->TestConnected();
   }
   catch(...)
   {
      DEBUGOUT(dbg_database, "Exception while testing database connection state");
   }

   return result;
}

//
// PrometheusDB::executeUpdateStatement
//
// Send in a table name, map of fields, and optionally, a map of options for
// those fields, to update records in the database.
//
bool PrometheusDB::executeUpdateStatement
(
   const pdb::string      &tableName,
   const pdb::stringmap   &fieldMap,
   const pdb::strtointmap *fieldOptions
)
{
   bool result = false;
   const pdb::strtointmap &options =
      fieldOptions ? *fieldOptions : pdb::strtointmap();

   try
   {
      if(pImpl->db)
         result = ExecuteUpdateStatement(pImpl->db, tableName, fieldMap, options);
   }
   catch(...)
   {
      DEBUGOUT(dbg_database, "Unknown exception during ExecuteUpdateStatement");
   }

   return result;
}

//
// PrometheusDB::executeInsertStatement
//
// Send in a table name, map of fields, and optionally, a map of options for
// those fields, to insert new records into the database.
//
bool PrometheusDB::executeInsertStatement
(
   const pdb::string      &tableName,
   const pdb::stringmap   &fieldMap,
   const pdb::strtointmap *fieldOptions
)
{
   bool result = false;
   const pdb::strtointmap &options =
      fieldOptions ? *fieldOptions : pdb::strtointmap();

   try
   {
      if(pImpl->db)
         result = ExecuteInsertStatement(pImpl->db, tableName, fieldMap, options);
   }
   catch(...)
   {
      DEBUGOUT(dbg_database, "Unknown exception during ExecuteInsertStatement");
   }

   return result;
}

//
// PrometheusDB::executeStatement
//
// Execute a generic SQL statement against the database.
//
bool PrometheusDB::executeStatement(const pdb::string &sql)
{
   bool result = false;

   try
   {
      if(pImpl->db)
         result = ExecuteStatement(pImpl->db, sql);
   }
   catch(...)
   {
      DEBUGOUT(dbg_database, "Unknown exception during ExecuteStatement");
   }

   return result;
}

//
// PrometheusDB::getNextId
//
// For properly setup tables (those with a uniformly-named id generator), this
// will retrieve a suitable id for the next record, in a manner that can be
// rolled back if a subsequent insert op should fail.
//
void PrometheusDB::getNextId(const pdb::string &tableName, pdb::string &result)
{
   try
   {
      if(pImpl->db)
         result = GetNextId(pImpl->db, tableName);
   }
   catch(...)
   {
      DEBUGOUT(dbg_database, "Unknown exception during GetNextId");
   }
}

//
// PrometheusDB::getOneField
//
// Returns the value of a single field in "result" given a SQL statement that
// generates a single column and row.
//
void PrometheusDB::getOneField(const pdb::string &sql, pdb::string &result)
{
   try
   {
      if(pImpl->db)
         result = GetOneField(pImpl->db, sql);
   }
   catch(...)
   {
      DEBUGOUT(dbg_database, "Unknown exception during GetOneField");
   }
}

//
// PrometheusDB::sqlToMap
//
// Return the single-row results of a SQL query in a map<string, string>
//
bool PrometheusDB::sqlToMap(const pdb::string &sql, pdb::stringmap &fieldMap)
{
   bool result = false;

   try
   {
      if(pImpl->db)
         result = SqlToMap(pImpl->db, sql, fieldMap);
   }
   catch(...)
   {
      DEBUGOUT(dbg_database, "Unknown exception during SqlToMap");
   }

   return result;
}

//
// PrometheusDB::sqlToVecMap
//
// Returns the multiple-row result of a SQL query as a vector<map<string, string> >,
// which while not of maximum efficiency due to the duplication of field names
// in every map row of the vector, is MUCH easier to use than a more efficient
// data structure would be.
//
bool PrometheusDB::sqlToVecMap(const pdb::string &sql, pdb::vecmap &fieldVecMap)
{
   bool result = false;

   try
   {
      if(pImpl->db)
         result = SqlToVecMap(pImpl->db, sql, fieldVecMap);
   }
   catch(...)
   {
      DEBUGOUT(dbg_database, "Unknown exception during SqlToVecMap");
   }

   return result;
}

//
// PrometheusDB::sqlToSet
//
// Returns the single-row results of a SQL query as a set of strings.
//
bool PrometheusDB::sqlToSet(const pdb::string &sql, pdb::stringset &fieldSet)
{
   bool result = false;

   try
   {
      if(pImpl->db)
         result = SqlToSet(pImpl->db, sql, fieldSet);
   }
   catch(...)
   {
      DEBUGOUT(dbg_database, "Unknown exception during SqlToSet");
   }

   return result;
}

//
// PrometheusDB::sqlToVec
//
// Returns the single-row results of a SQL query as a vector of strings.
//
bool PrometheusDB::sqlToVec(const pdb::string &sql, pdb::stringvec &fieldVec)
{
   bool result = false;

   try
   {
      if(pImpl->db)
         result = SqlToVec(pImpl->db, sql, fieldVec);
   }
   catch(...)
   {
      DEBUGOUT(dbg_database, "Unknown exception during SqlToVec");
   }

   return result;
}

//=============================================================================
//
// PrometheusTransaction
//

//
// Private implementation details for PrometheusTransaction
//
class PrometheusTransactionPimpl
{
public:
   TIBTransaction *transaction;

   PrometheusTransactionPimpl()
   {
      try
      {
         transaction = new TIBTransaction(NULL);
      }
      catch(...)
      {
         DEBUGOUT(dbg_database, "Instantiation of a TIBTransaction instance failed!");
      }
   }

   ~PrometheusTransactionPimpl()
   {
      try
      {
         // If a transaction has been leaked while active, roll it back.
         if(transaction && transaction->Active)
            transaction->Rollback();
      }
      catch(...)
      {
         DEBUGOUT(dbg_database, "Rollback of an active leaked transaction failed!");
      }

      try
      {
         delete transaction;
      }
      catch(...)
      {
         DEBUGOUT(dbg_database, "Destruction of a TIBTransaction failed!");
      }
      
      transaction = NULL;
   }
};

//
// PrometheusTransaction Constructor
//
PrometheusTransaction::PrometheusTransaction()
{
   pImpl = new PrometheusTransactionPimpl;
}

//
// PrometheusTransaction Destructor
//
// This will help with memory management enormously. By declaring this object
// on the stack, the enclosed TIBTransaction will be closed and freed
// automatically when the calling function returns. You must still commit
// any written changes first though, or the transaction will rollback by
// default for safety.
//
PrometheusTransaction::~PrometheusTransaction()
{
   if(pImpl)
   {
      delete pImpl;
      pImpl = NULL;
   }
}

//
// PrometheusTransaction::isActive
//
// Returns true or false regarding the transaction's active state.
//
bool PrometheusTransaction::isActive()
{
   bool result = false;

   try
   {
      if(pImpl->transaction)
         result = pImpl->transaction->Active;
   }
   catch(...)
   {
      DEBUGOUT(dbg_database, "Exception while testing transaction active state");
   }

   return result;
}

//
// PrometheusTransaction::commit
//
// Commit a transaction. It should be in the active state or the resulting VCL
// exception will cause a false return value. The transaction is closed after
// this call.
//
bool PrometheusTransaction::commit()
{
   bool result = false;

   try
   {
      if(pImpl->transaction)
      {
         pImpl->transaction->Commit();
         result = true;
      }
   }
   catch(...)
   {
      DEBUGOUT(dbg_error, "WARNING: transaction commit failure! Data may be lost!");
   }

   return result;
}

//
// PrometheusTransaction::rollback
//
// Rollback a transaction. It should be in the active state or the resulting VCL
// exception will cause a false return value. The transaction is closed after
// this call.
//
bool PrometheusTransaction::rollback()
{
   bool result = false;

   try
   {
      if(pImpl->transaction)
      {
         pImpl->transaction->Rollback();
         result = true;
      }
   }
   catch(...)
   {
      DEBUGOUT(dbg_database, "Transaction rollback failure");
   }

   return result;
}

//
// PrometheusTransaction::stdTransaction
//
// Attach a transaction to the database with the standard parameters.
//
bool PrometheusTransaction::stdTransaction(PrometheusDB &db)
{
   bool result = false;
   TIBDatabase    *ibdb = db.pImpl->db;
   TIBTransaction *ibta = pImpl->transaction;

   if(ibdb && ibta)
      result = StdTransaction(ibta, ibdb);

   return result;
}

//
// PrometheusTransaction::executeUpdateStatement
//
// Send in a table name, map of fields, and optionally, a map of options for
// those fields, to update records in the database.
//
bool PrometheusTransaction::executeUpdateStatement
(
   const pdb::string      &tableName,
   const pdb::stringmap   &fieldMap,
   const pdb::strtointmap *fieldOptions
)
{
   bool result = false;
   const pdb::strtointmap &options =
      fieldOptions ? *fieldOptions : pdb::strtointmap();

   try
   {
      if(pImpl->transaction)
         result = ExecuteUpdateStatement(pImpl->transaction, tableName, fieldMap,
                                         options);
   }
   catch(...)
   {
      DEBUGOUT(dbg_database, "Unknown exception during ExecuteUpdateStatement");
   }

   return result;
}

//
// PrometheusTransaction::executeInsertStatement
//
// Send in a table name, map of fields, and optionally, a map of options for
// those fields, to insert new records into the database.
//
bool PrometheusTransaction::executeInsertStatement
(
   const pdb::string      &tableName,
   const pdb::stringmap   &fieldMap,
   const pdb::strtointmap *fieldOptions
)
{
   bool result = false;
   const pdb::strtointmap &options =
      fieldOptions ? *fieldOptions : pdb::strtointmap();

   try
   {
      if(pImpl->transaction)
         result = ExecuteInsertStatement(pImpl->transaction, tableName, fieldMap,
                                         options);
   }
   catch(...)
   {
      DEBUGOUT(dbg_database, "Unknown exception during ExecuteInsertStatement");
   }

   return result;
}

//
// PrometheusTransaction::executeStatement
//
// Execute a generic SQL statement against the database.
//
bool PrometheusTransaction::executeStatement(const pdb::string &sql)
{
   bool result = false;

   try
   {
      if(pImpl->transaction)
         result = ExecuteStatement(pImpl->transaction, sql);
   }
   catch(...)
   {
      DEBUGOUT(dbg_database, "Unknown exception during ExecuteStatement");
   }

   return result;
}

//
// PrometheusTransaction::getNextId
//
// For properly setup tables (those with a uniformly-named id generator), this
// will retrieve a suitable id for the next record, in a manner that can be
// rolled back if a subsequent insert op should fail.
//
void PrometheusTransaction::getNextId(const pdb::string &tableName, pdb::string &result)
{
   try
   {
      if(pImpl->transaction)
         result = GetNextId(pImpl->transaction, tableName);
   }
   catch(...)
   {
      DEBUGOUT(dbg_database, "Unknown exception during GetNextId");
   }
}

//
// PrometheusTransaction::getOneField
//
// Returns the value of a single field in "result" given a SQL statement that
// generates a single column and row.
//
void PrometheusTransaction::getOneField(const pdb::string &sql, pdb::string &result)
{
   try
   {
      if(pImpl->transaction)
         result = GetOneField(pImpl->transaction, sql);
   }
   catch(...)
   {
      DEBUGOUT(dbg_database, "Unknown exception during GetOneField");
   }
}

//
// PrometheusTransaction::sqlToMap
//
// Return the single-row results of a SQL query in a map<string, string>
//
bool PrometheusTransaction::sqlToMap(const pdb::string &sql, pdb::stringmap &fieldMap)
{
   bool result = false;

   try
   {
      if(pImpl->transaction)
         result = SqlToMap(pImpl->transaction, sql, fieldMap);
   }
   catch(...)
   {
      DEBUGOUT(dbg_database, "Unknown exception during SqlToMap");
   }

   return result;
}

//
// PrometheusTransaction::sqlToVecMap
//
// Returns the multiple-row result of a SQL query as a vector<map<string, string> >,
// which while not of maximum efficiency due to the duplication of field names
// in every map row of the vector, is MUCH easier to use than a more efficient
// data structure would be.
//
bool PrometheusTransaction::sqlToVecMap(const pdb::string &sql, pdb::vecmap &fieldVecMap)
{
   bool result = false;

   try
   {
      if(pImpl->transaction)
         result = SqlToVecMap(pImpl->transaction, sql, fieldVecMap);
   }
   catch(...)
   {
      DEBUGOUT(dbg_database, "Unknown exception during SqlToVecMap");
   }

   return result;
}

//
// PrometheusTransaction::sqlToSet
//
// Returns the single-row results of a SQL query as a set of strings.
//
bool PrometheusTransaction::sqlToSet(const pdb::string &sql, pdb::stringset &fieldSet)
{
   bool result = false;

   try
   {
      if(pImpl->transaction)
         result = SqlToSet(pImpl->transaction, sql, fieldSet);
   }
   catch(...)
   {
      DEBUGOUT(dbg_database, "Unknown exception during SqlToSet");
   }

   return result;
}

//
// PrometheusTransaction::sqlToVec
//
// Returns the single-row results of a SQL query as a vector of strings.
//
bool PrometheusTransaction::sqlToVec(const pdb::string &sql, pdb::stringvec &fieldVec)
{
   bool result = false;

   try
   {
      if(pImpl->transaction)
         result = SqlToVec(pImpl->transaction, sql, fieldVec);
   }
   catch(...)
   {
      DEBUGOUT(dbg_database, "Unknown exception during SqlToVec");
   }

   return result;
}

//=============================================================================
//
// PrometheusLookup
//

//
// PromtheusLookup - Parameterized Constructor
//
PrometheusLookup::PrometheusLookup(PrometheusDB &db, const string &pTableName)
   : tableName(), valueToID(), idToValue()
{
   load(db, pTableName);
}

//
// PrometheusLookup - Copy Constructor
//
PrometheusLookup::PrometheusLookup(const PrometheusLookup &other)
   : tableName(other.tableName), valueToID(other.valueToID), idToValue(other.idToValue)
{
}

//
// PrometheusLookup::load
//
// Load the contents of a lookup table from the passed-in database.
// Lookup tables need to have an (id, item, valid) column layout.
//
void PrometheusLookup::load(PrometheusDB &db, const pdb::string &pTableName)
{
   pdb::vecmap lookup;
   pdb::string sql;

   tableName = pTableName;
   sql = "select id, item from " + tableName + " where valid = 1";

   if(!db.sqlToVecMap(sql, lookup))
   {
      DEBUGOUT(dbg_database, "Failed to load lookup " << tableName.c_str());
      return;
   }

   valueToID.clear();
   idToValue.clear();

   // build fast bi-directional lookup maps
   for(pdb::vecmap::iterator i = lookup.begin(); i != lookup.end(); ++i)
   {
      string id   = (*i)["id"];
      string item = LowercaseString((*i)["item"]);

      valueToID[item] = id;
      idToValue[ id ] = item;
   }
}

//
// PrometheusLookup::loadCustom
//
// Load the contents of a lookup table from the passed-in database.
// Lookup tables need to have an (id, item, valid) column layout.
// This overload accepts a custom SQL query. It must return columns (id, item).
//
void PrometheusLookup::loadCustom(PrometheusDB &db, const pdb::string &pTableName,
                                  const pdb::string &sql)
{
   pdb::vecmap lookup;

   tableName = pTableName;

   if(!db.sqlToVecMap(sql, lookup))
   {
      DEBUGOUT(dbg_database, "Failed to load lookup " << tableName.c_str());
      return;
   }

   valueToID.clear();
   idToValue.clear();

   // build fast bi-directional lookup maps
   for(pdb::vecmap::iterator i = lookup.begin(); i != lookup.end(); ++i)
   {
      string id   = (*i)["id"];
      string item = LowercaseString((*i)["item"]);

      valueToID[item] = id;
      idToValue[ id ] = item;
   }
}

//
// PrometheusLookup::idOf
//
// Find the value in this lookup table; if it exists, the corresponding id
// is returned as a string. Otherwise, an empty string is returned.
//
const pdb::string &PrometheusLookup::idOf(const pdb::string &value) const
{
   static const string noresult = "";

   pdb::stringmap::const_iterator i = valueToID.find(LowercaseString(value));

   if(i != valueToID.end())
      return i->second;
   else
      return noresult;
}

//
// PrometheusLookup::valueOf
//
// Find the id in this lookup table; if it exists, the corresponding value is
// returned as a string. Otherwise, an empty string is returned.
//
const pdb::string &PrometheusLookup::valueOf(const pdb::string &id) const
{
   static const string noresult = "";

   pdb::stringmap::const_iterator i = idToValue.find(id);

   if(i != idToValue.end())
      return i->second;
   else
      return noresult;
}

//
// PrometheusLookup::isValidID
//
// Returns true if the id actually exists in the lookup, and false otherwise.
//
bool PrometheusLookup::isValidID(const pdb::string &id) const
{
   return (idToValue.find(id) != idToValue.end());
}

//
// PrometheusLookup::isValidValue
//
// Returns true if the value actually exists in the lookup, and false otherwise.
//
bool PrometheusLookup::isValidValue(const pdb::string &value) const
{
   return (valueToID.find(LowercaseString(value)) != valueToID.end());
}

//=============================================================================
//
// PrometheusLookups
//

//
// PrometheusLookups::haveLookup
//
// Test to see if the named table has been loaded as a lookup already.
//
bool PrometheusLookups::haveLookup(const pdb::string &tableName) const
{
   return (lookupTables.find(tableName) != lookupTables.end());
}

//
// PrometheusLookups::loadLookup
//
// Load the indicated lookup table into a PrometheusLookup structure, provided
// we have not already loaded that table.
//
void PrometheusLookups::loadLookup(PrometheusDB &db, const pdb::string &tableName)
{
   // one time only.
   if(!haveLookup(tableName))
   {
      PrometheusLookup &newLookup = lookupTables[tableName];
      newLookup.load(db, tableName);
   }
}

//
// PrometheusLookups::loadCustom
//
// Load the indicated lookup table into a PrometheusLookup structure, provided
// we have not already loaded that table. Use a custom SQL query to do it.
//
void PrometheusLookups::loadCustom(PrometheusDB &db, const pdb::string &tableName,
                                   const pdb::string &sql)
{
   // one time only.
   if(!haveLookup(tableName))
   {
      PrometheusLookup &newLookup = lookupTables[tableName];
      newLookup.loadCustom(db, tableName, sql);
   }
}

//
// PrometheusLookups::purgeLookups
//
// Drops all cached lookup tables.
//
void PrometheusLookups::purgeLookups()
{
   lookupTables.clear();
}

//
// PrometheusLookups::idOf
//
// Find the value for the given id in the indicated lookup table.
//
const pdb::string &PrometheusLookups::idOf(const pdb::string &tableName,
                                           const pdb::string &value) const
{
   static const string noresult = "";

   LookupMap::const_iterator lookup = lookupTables.find(tableName);

   if(lookup != lookupTables.end())
      return lookup->second.idOf(value);
   else
   {
      DEBUGOUT(dbg_database, "No such lookup " << tableName.c_str());
      return noresult;
   }
}

//
// PrometheusLookups::valueOf
//
// Find the id for the given value in the indicated lookup table.
//
const pdb::string &PrometheusLookups::valueOf(const pdb::string &tableName,
                                              const pdb::string &id) const
{
   static const string noresult = "";

   LookupMap::const_iterator lookup = lookupTables.find(tableName);

   if(lookup != lookupTables.end())
      return lookup->second.valueOf(id);
   else
   {
      DEBUGOUT(dbg_database, "No such lookup " << tableName.c_str());
      return noresult;
   }
}

//
// PrometheusLookups::isValidID
//
// Returns true if the lookup is valid and the id actually exists in the lookup,
// and false otherwise.
//
bool PrometheusLookups::isValidID(const pdb::string &tableName,
                                  const pdb::string &id) const
{
   bool result = false;

   LookupMap::const_iterator lookup = lookupTables.find(tableName);

   if(lookup != lookupTables.end())
      result = lookup->second.isValidID(id);
   else
      DEBUGOUT(dbg_database, "No such lookup " << tableName.c_str());

   return result;
}

//
// PrometheusLookups::isValidValue
//
// Returns true if the lookup is valid and the value actually exists in the
// lookup, and false otherwise.
//
bool PrometheusLookups::isValidValue(const pdb::string &tableName,
                                     const pdb::string &value) const
{
   bool result = false;

   LookupMap::const_iterator lookup = lookupTables.find(tableName);

   if(lookup != lookupTables.end())
      result = lookup->second.isValidValue(value);
   else
      DEBUGOUT(dbg_database, "No such lookup " << tableName.c_str());

   return result;
}

//
// PrometheusLookups::GetLookups
//
// Static method. The singleton lookup object will be created if it does not
// already exist, and then a reference to it will be returned.
//
PrometheusLookups &PrometheusLookups::GetLookups()
{
   static PrometheusLookups *theLookups = NULL;

   if(!theLookups)
      theLookups = new PrometheusLookups();

   return *theLookups;
}

//
// PrometheusLookups::LoadLookup
//
// Convenience static utility method to load a lookup table into the global
// singleton object.
//
void PrometheusLookups::LoadLookup(PrometheusDB &db, const pdb::string &tableName)
{
   GetLookups().loadLookup(db, tableName);
}

//
// PrometheusLookups::LoadCustom
//
// Convenience static utility method to load a lookup table into the global
// singleton object using a custom SQL query.
//
void PrometheusLookups::LoadCustom(PrometheusDB &db, const pdb::string &tableName,
                                   const pdb::string &sql)
{
   GetLookups().loadCustom(db, tableName, sql);
}

//
// PrometheusLookups::LoadLookups
//
// Load multiple lookups as named in an array of string constants.
//
void PrometheusLookups::LoadLookups(PrometheusDB &db, const char **tableNames,
                                    size_t numTables)
{
   for(size_t i = 0; i < numTables; i++)
      LoadLookup(db, tableNames[i]);
}

//
// PrometheusLookups::PurgeLookups
//
// Convenience static utility method to purge all loaded lookup tables from
// the global singleton object.
//
void PrometheusLookups::PurgeLookups()
{
   GetLookups().purgeLookups();
}

//
// PrometheusLookups::IdOf
//
// Convenience static utility method forwarding to PrometheusLookups::idOf
// for the global singleton object.
//
const pdb::string &PrometheusLookups::IdOf(const pdb::string &tableName,
                                           const pdb::string &value)
{
   return GetLookups().idOf(tableName, value);
}

//
// PrometheusLookups::ValueOf
//
// Convenience static utility method forwarding to PrometheusLookups::valueOf
// for the global singleton object.
//
const pdb::string &PrometheusLookups::ValueOf(const pdb::string &tableName,
                                              const pdb::string &id)
{
   return GetLookups().valueOf(tableName, id);
}

//
// PrometheusLookups::IsValidID
//
// Convenience static utility method forwarding to PrometheusLookups::isValidID
// for the global singleton object.
//
bool PrometheusLookups::IsValidID(const pdb::string &tableName,
                                  const pdb::string &id)
{
   return GetLookups().isValidID(tableName, id);
}

//
// PrometheusLookups::IsValidValue
//
// Convenience static utility method forwarding to
// PrometheusLookups::isValidValue for the global singleton object.
//
bool PrometheusLookups::IsValidValue(const pdb::string &tableName,
                                     const pdb::string &value)
{
   return GetLookups().isValidValue(tableName, value);
}

// EOF

