/*

   HL7Daemon - Prometheus database connection manager

*/

#ifndef PROMETHEUSDB_H__
#define PROMETHEUSDB_H__

#include <map>
#include <set>
#include <string>
#include <vector>

class PrometheusTransactionPimpl;
class PrometheusDB;

// Convenience typedefs for annoying verbose STL composite types
namespace pdb
{
   typedef std::string string;
   typedef std::map<std::string, std::string> stringmap;
   typedef std::set<std::string> stringset;
   typedef std::vector<std::string> stringvec;
   typedef std::map<std::string, int> strtointmap;
   typedef std::vector<std::map<std::string, std::string> > vecmap;
}

//
// PrometheusTransaction
//
// Represents a transaction against the Prometheus database.
//
class PrometheusTransaction
{
protected:
   PrometheusTransactionPimpl *pImpl;
   friend class PrometheusDB;

public:
   PrometheusTransaction();
   ~PrometheusTransaction();

   bool isActive();

   bool commit();
   bool rollback();

   bool stdTransaction(PrometheusDB &db);

   bool executeUpdateStatement(const pdb::string      &tableName,
                               const pdb::stringmap   &fieldMap,
                               const pdb::strtointmap *fieldOptions = NULL);
   bool executeInsertStatement(const pdb::string      &tableName,
                               const pdb::stringmap   &fieldMap,
                               const pdb::strtointmap *fieldOptions = NULL);
   bool executeStatement(const pdb::string &sql);

   void getNextId(const pdb::string &tableName, pdb::string &result);
   void getOneField(const pdb::string &sql, pdb::string &result);
   bool sqlToMap(const pdb::string &sql, pdb::stringmap &fieldMap);
   bool sqlToVecMap(const pdb::string &sql, pdb::vecmap &fieldVecMap);
   bool sqlToSet(const pdb::string &sql, pdb::stringset &fieldSet);
   bool sqlToVec(const pdb::string &sql, pdb::stringvec &fieldVec);
};

class IniFile;
class PrometheusDBPimpl;

//
// PrometheusDB
//
// Represents a connection to the Prometheus Firebird database.
//
class PrometheusDB
{
protected:
   PrometheusDBPimpl *pImpl;
   friend class PrometheusTransaction;

public:
   PrometheusDB();
   ~PrometheusDB();

   bool connect(IniFile &);
   void disconnect();
   bool isConnected();

   bool executeUpdateStatement(const pdb::string      &tableName,
                               const pdb::stringmap   &fieldMap,
                               const pdb::strtointmap *fieldOptions = NULL);
   bool executeInsertStatement(const pdb::string      &tableName,
                               const pdb::stringmap   &fieldMap,
                               const pdb::strtointmap *fieldOptions = NULL);
   bool executeStatement(const pdb::string &sql);

   void getNextId(const pdb::string &tableName, pdb::string &result);
   void getOneField(const pdb::string &sql, pdb::string &result);
   bool sqlToMap(const pdb::string &sql, pdb::stringmap &fieldMap);
   bool sqlToVecMap(const pdb::string &sql, pdb::vecmap &fieldVecMap);
   bool sqlToSet(const pdb::string &sql, pdb::stringset &fieldSet);
   bool sqlToVec(const pdb::string &sql, pdb::stringvec &fieldVec);
};

//
// PrometheusLookup
//
// Represents a lookup table.
//
class PrometheusLookup
{
protected:
   pdb::string tableName;
   pdb::stringmap valueToID;
   pdb::stringmap idToValue;

public:
   PrometheusLookup() : tableName(), valueToID(), idToValue() {}
   PrometheusLookup(PrometheusDB &db, const pdb::string &pTableName);
   PrometheusLookup(const PrometheusLookup &other);

   void load(PrometheusDB &db, const pdb::string &pTableName);
   void loadCustom(PrometheusDB &db, const pdb::string &pTableName, const pdb::string &sql);

   const pdb::string &getTableName() const { return tableName; }
   const pdb::string &idOf(const pdb::string &value) const;
   const pdb::string &valueOf(const pdb::string &id) const;

   bool isValidID(const pdb::string &id) const;
   bool isValidValue(const pdb::string &value) const;
};

//
// PrometheusLookups
//
// Global singleton lookup table manager
//
class PrometheusLookups
{
protected:
   typedef std::map<pdb::string, PrometheusLookup> LookupMap;

   LookupMap lookupTables;

   // Protected constructor for singleton object
   PrometheusLookups() : lookupTables() {}

public:
   bool haveLookup(const pdb::string &tableName) const;
   void loadLookup(PrometheusDB &db, const pdb::string &tableName);
   void loadCustom(PrometheusDB &db, const pdb::string &tableName,
                   const pdb::string &sql);
   void purgeLookups();

   const pdb::string &idOf(const pdb::string &tableName, const pdb::string &value) const;
   const pdb::string &valueOf(const pdb::string &tableName, const pdb::string &id) const;

   bool isValidID(const pdb::string &tableName, const pdb::string &id) const;
   bool isValidValue(const pdb::string &tableName, const pdb::string &value) const;

   static PrometheusLookups &GetLookups();
   static void LoadLookup(PrometheusDB &db, const pdb::string &tableName);
   static void LoadCustom(PrometheusDB &db, const pdb::string &tableName,
                          const pdb::string &sql);
   static void LoadLookups(PrometheusDB &db, const char **tableNames, size_t numTables);
   static void PurgeLookups();
   static const pdb::string &IdOf(const pdb::string &tableName, const pdb::string &value);
   static const pdb::string &ValueOf(const pdb::string &tableName, const pdb::string &id);
   static bool IsValidID(const pdb::string &tableName, const pdb::string &id);
   static bool IsValidValue(const pdb::string &tableName, const pdb::string &value);
};

#endif

// EOF

