/*

   HL7Daemon - String processing utilities

*/

#include <string>
using namespace std;

#include "prometheusdb.h"
#include "stringutils.h"

#pragma hdrstop
#include "util.h"

// Split a string similar to A=B into A and B
bool SplitKeyValue(const string &incoming, string &key, string &value)
{
   if(incoming.find_first_of("=") != string::npos)
   {
      key = incoming.substr(0, incoming.find_first_of("="));
      value = StripFirst(incoming.substr(incoming.find_first_of("=")));
   }
   else
      return false;

   return true;
}

// Strip off the first character of a string and return a new copy
string StripFirst(const string &incoming)
{
   if(incoming.length() > 1)
      return incoming.substr(1, incoming.length() - 1);

   return "";
}

// Strip off the last character of a string and return a new copy
string StripLast(const string &incoming)
{
   return incoming.substr(0, incoming.length() - 1);
}

// Returns true if the string is surrounded with [], false otherwise.
bool IsBracketed(const string &incoming)
{
   if(incoming.length())
   {
      if(incoming[0] == '[' && incoming[incoming.length() - 1] == ']')
         return true;
   }

   return false;
}

//
// cpp_strdup
//
// jhaley - a strdup substitute that uses new [] to do allocation.
//
char *cpp_strdup(const char *input)
{
   size_t len = strlen(input) + 1;

   char *buffer = new char [len];

   if(buffer)
   {
      memset(buffer, 0, len);
      strcpy(buffer, input);
   }

   return buffer;
}

//=============================================================================
//
// HL7-related utilities (adapted from Prometheus hl7.cpp)
//

//
// hl7_BLToBoolean
//   IN:  HL7 two-state value, and the value that means "true"
//   OUT: C++ boolean, true if value matches true_value.
//
bool hl7_BLToBoolean(const string &value, const string &true_value)
{
   return (value == true_value);
}

//
// hl7_DTToPdate
//   IN:  An HL7 timestamp or date string of 8 or more characters
//   OUT: Prometheus Pdate instance, if the input is a valid date
//
void hl7_DTToPdate(const string &value, Pdate &answer)
{
   if(value.length() >= 8) // CCYYMMDD
   {
      Pdate temp;
      temp.year  = StringToInt(value.substr(0, 4));
      temp.month = StringToInt(value.substr(4, 2));
      temp.day   = StringToInt(value.substr(6, 2));
      if(temp.IsValidDate())
         answer = temp;
   }
}

//
// hl7_TMToPtime
//   IN:  An HL7 time string of 4 or more characters
//   OUT: Prometheus Ptime instance, if the input is a valid time
//
void hl7_TMToPtime(const string &value, Ptime &answer)
{
   if(value.length() >= 4) // HHMM
   {
      Ptime temp;
      temp.hour   = StringToInt(value.substr(0, 2));
      temp.minute = StringToInt(value.substr(2, 2));
      temp.second = temp.millisecond = 0;

      // jhaley 20120614: support for seconds, milliseconds
      if(value.length() >= 6) // HHMMSS
      {
         temp.second = StringToInt(value.substr(4, 2));

         if(value.length() >= 9) // HHMMSSnnn
           temp.millisecond = StringToInt(value.substr(6, 3));
      }

      if(temp.IsValidTime())
         answer = temp;
   }
}

//
// hl7_TSToPstamp
//   IN:  An HL7 timestamp string of 12 or more characters
//   OUT: Prometheus Pstamp instance, if both the date and time are valid.
//
void hl7_TSToPstamp(const string &value, Pstamp &answer)
{
   if(value.length() >= 12) // CCYYMMDDhhmmssnnn
   {
      Ptime  temp_time;
      Pdate  temp_date;
      Pstamp temp_stamp;

      hl7_TMToPtime(value.substr(8),    temp_time);
      hl7_DTToPdate(value.substr(0, 8), temp_date);

      temp_stamp.year        = temp_date.year;
      temp_stamp.month       = temp_date.month;
      temp_stamp.day         = temp_date.day;
      temp_stamp.hour        = temp_time.hour;
      temp_stamp.minute      = temp_time.minute;
      temp_stamp.second      = temp_time.second; // jhaley: seconds, milliseconds
      temp_stamp.millisecond = temp_time.millisecond;

      if(temp_stamp.IsValidStamp())
         answer = temp_stamp;
   }
}

//
// hl7_GenderToGenderID
//   IN:  source cb gender representation - 'M', 'F', or null
//   OUT: ID referencing Prometheus gender_types lookup table
//
void hl7_GenderToGenderID(const string &hl7gender, string &promID)
{
   if(hl7gender.empty())
   {
      promID = "0";
      return;
   }
   
   switch(toupper(hl7gender[0]))
   {
   case 'M':
      promID = PrometheusLookups::IdOf("gender_types", "Male");
      break;
   case 'F':
      promID = PrometheusLookups::IdOf("gender_types", "Female");
      break;
   default:
      promID = "0";
      break;
   }
}

//
// hl7_RaceToRaceID
//   IN:  source db race representation (various strings)
//   OUT: ID referencing Prometheus race_types lookup table
//
void hl7_RaceToRaceID(const string &hl7race, string &promID)
{
   string lcrace = LowercaseString(hl7race);

   // Direct correspondences:
   if(lcrace == "caucasian"        ||
      lcrace == "hispanic"         ||
      lcrace == "african american" ||
      lcrace == "asian"            ||
      lcrace == "other")
   {
      promID = PrometheusLookups::IdOf("race_types", lcrace);
   }
   else if(LikeString("american indian%", lcrace)) // spec is vague/incorrect...
   {
      promID = PrometheusLookups::IdOf("race_types", "Native American");
   }
   else if(lcrace == "" || lcrace == "undefined") // explicitly undefined
   {
      promID = "0";
   }
   else // Any other explicit type we don't support, change to "Other"
   {
      promID = PrometheusLookups::IdOf("race_types", "Other");
   }
}

//
// hl7_MaritalStatusToMaritalID
//   IN:  source db relationship status representation (S, M, D, W, X, null)
//   OUT: ID referencing Prometheus marital_types lookup table
//
void hl7_MaritalStatusToMaritalID(const string &hl7ms, string &promID)
{
   if(hl7ms.empty())
   {
      promID = "0";
      return;
   }

   // Perfect correspondence!
   switch(toupper(hl7ms[0]))
   {
   case 'S':
      promID = PrometheusLookups::IdOf("marital_types", "Single");
      break;
   case 'M':
      promID = PrometheusLookups::IdOf("marital_types", "Married");
      break;
   case 'D':
      promID = PrometheusLookups::IdOf("marital_types", "Divorced");
      break;
   case 'W':
      promID = PrometheusLookups::IdOf("marital_types", "Widowed");
      break;
   case 'X':
      promID = PrometheusLookups::IdOf("marital_types", "Seperated"); // (sic)
      break;
   default:
      promID = "0";
      break;
   }
}

//
// hl7_SSNToNumeric
//   IN:  Optionally dash-delimited social security number: xxx[-]?yy[-]?zzzz
//   OUT: SSN represented by numerals only: xxxyyzzzz
//
void hl7_SSNToNumeric(const string &hl7SSN, string &promSSN)
{
   if(!hl7SSN.empty())
      promSSN = StripAllExcept(hl7SSN, uints);
}

//
// hl7_StateToStateID
//   IN:  source db state representation (two-letter postal abbreviation)
//   OUT: ID referencing Prometheus state_types lookup table
//
// NB: Assumes custom SQL used to build state_types in hl7process.cpp
//
void hl7_StateToStateID(const string &hl7state, string &promID)
{
   promID = PrometheusLookups::IdOf("state_types", hl7state);
}

//
// hl7_NormalizeZip
//   IN:  Standard zip5 or zip5+4
//   OUT: A string of no more than 9 digits, numerals only
//
void hl7_NormalizeZip(const string &hl7zip, string &outZip)
{
   if(!hl7zip.empty())
      outZip = StripAllExcept(hl7zip, uints).substr(0, 9);
}

//
// hl7_NormalizePhone
//   IN:  Standard HL7 compound phone number (aaa)eee-llll[Xnnnn]? OR aaaeeellll
//   OUT: aaaeeellll (ex: (405)598-5460 -> 4055985460)
//
void hl7_NormalizePhone(const string &hl7phone, string &outPhone)
{
   size_t x;

   if(hl7phone.empty())
      return;

   // TODO: Prometheus does not (yet?) support extensions in phone numbers.
   // There's room in the database schema (varchar(255)), but none of the
   // formatting routines etc. tolerate them properly. Until then, if then is
   // ever, we'll strip them out here.

   if((x = hl7phone.find_first_of("Xx")) != string::npos)
      outPhone = StripAllExcept(hl7phone.substr(0, x), uints);
   else
      outPhone = StripAllExcept(hl7phone, uints);
}

// EOF

