//
// jhaley 20110224: Cached Files Module 
//
// Maintain a local cache of files that are stored in the database, so that they don't have to be downloaded at every
// run. The routines are smart and still enable use of the resources even if the save-to-disk portion of the loading
// process fails.
//

#include <sys/stat.h>
#include <jpeg.hpp>
#include "cached_files.h"
#include "sqlLib.h"

#ifdef use_prometheus_util
 #include "derivedheaders.h"
#endif

//---------------------------------------------------------------------------------------------------------------------
// *** File Cache Management *****************************************************************************************
// The Prometheus file cache is a directory tree underneath the Windows user's local 'Application Data' directory.
//---------------------------------------------------------------------------------------------------------------------
// FileCache::GetBasePath
// Return the base path under which we create the main cache directory
string FileCache::GetBasePath(void)
{
  string ret_path;
  char path_buffer[MAX_PATH+1];
  HRESULT ret_code;

  memset(path_buffer, 0, sizeof(path_buffer));
  ret_code = SHGetFolderPath(NULL, CSIDL_LOCAL_APPDATA, NULL, SHGFP_TYPE_CURRENT, path_buffer);
  if(SUCCEEDED(ret_code))
    ret_path = path_buffer;
  else
    ret_path = "."; // default to trying to use the current directory, though it may not work.

  return ret_path;
}
//---------------------------------------------------------------------------------------------------------------------
// FileCache::RemoveTrailingSlash
// Removes a trailing slash from a path component if it ends in one.
string FileCache::RemoveTrailingSlash(const string &source_path)
{
  string ret_str = source_path;

  if(source_path.length() > 0 && source_path.at(source_path.length() - 1) == '\\')
    ret_str = source_path.substr(0, source_path.length() - 1);

  return ret_str;
}
//---------------------------------------------------------------------------------------------------------------------
// FileCache::RemoveLeadingSlash
// Removes a leading slash from a path component, if it begins with one.
string FileCache::RemoveLeadingSlash(const string &source_path)
{
  string ret_str = source_path;

  if(source_path.length() > 0 && source_path.at(0) == '\\')
    ret_str = source_path.substr(1);

  return ret_str;
}
//---------------------------------------------------------------------------------------------------------------------
// FileCache::DirectoryExists
// Tests an absolute path for existence as a directory. If test_if_full_perms is true, then the directory's existence
// will be ignored unless the user has full permissions (read/write/execute) to that directory.
bool FileCache::DirectoryExists(const string &path, bool test_if_full_perms)
{
  bool dir_exists = false;
  struct stat s;

  string tmp_path = RemoveTrailingSlash(path); // stat doesn't find paths ending with a slash

  if(!stat(tmp_path.c_str(), &s) && S_ISDIR(s.st_mode))
  {
    if(test_if_full_perms)
      dir_exists = ((s.st_mode & S_IRWXU) == S_IRWXU);
    else
      dir_exists = true;
  }

  return dir_exists;
}
//---------------------------------------------------------------------------------------------------------------------
// FileCache::FileExists
// Similar to the above but for a readable disk file rather than a directory. Send in an absolute path.
bool FileCache::FileExists(const string &path, const string &file)
{
  bool file_exists = false;
  struct stat s;

  string tmp_path = PathConcatenate(path, file);

  if(!stat(tmp_path.c_str(), &s) && !S_ISDIR(s.st_mode))
    file_exists = true;

  return file_exists;
}
//---------------------------------------------------------------------------------------------------------------------
// FileCache::FileExistsInCache
// Calls the above routine after concatenating the local file cache basepath to the passed-in path; ie., that path
// should be relative to the cache basepath - ex: 'Prometheus\images...', NOT 'C:\Users\FooBar\AppData...'
bool FileCache::FileExistsInCache(const string &path, const string &file)
{
  return FileExists(PathConcatenate(GetBasePath(), path), file);
}
//---------------------------------------------------------------------------------------------------------------------
// FileCache::PathConcatenate
// Concatenate two portions of a file path safely with respect to whether either part might end or begin with a slash.
string FileCache::PathConcatenate(const string &base, const string &newpart)
{
  string base_trunc    = RemoveTrailingSlash(base);
  string newpart_trunc = RemoveTrailingSlash(RemoveLeadingSlash(newpart));

  return (base_trunc + "\\" + newpart_trunc);
}
//---------------------------------------------------------------------------------------------------------------------
// FileCache::GetFileSpec
// Extract the filename from a path and return the path and filename as a pair.
pair<string, string> FileCache::GetFileSpec(const string &full_path)
{
  pair<string, string> parts_to_return;
  size_t last_slash_pos = full_path.find_last_of('\\'); // find final slash character

  // If found a slash and...            not at beginning, and not at very end then:
  if(last_slash_pos != string::npos && last_slash_pos != 0 && last_slash_pos != full_path.size() - 1)
  {
    parts_to_return.first  = full_path.substr(0, last_slash_pos);  // file path in .first
    parts_to_return.second = full_path.substr(last_slash_pos + 1); // file name in .second
  }
  else
  {
    parts_to_return.first = "";         // no valid path component to return
    parts_to_return.second = full_path; // return full parameter in .second
  }

  return parts_to_return;
}
//---------------------------------------------------------------------------------------------------------------------
// FileCache::CreateDirectory
// Create a directory under the cache base path, non-recursive, provided it doesn't exist already.
bool FileCache::CreateDirectory(const string &dir_name)
{
  string concat_path = PathConcatenate(GetBasePath(), dir_name);

  if(DirectoryExists(concat_path, false))
    return true; // already exists

  // Create the directory
  return !mkdir(concat_path.c_str());
}
//---------------------------------------------------------------------------------------------------------------------
// FileCache::CreateDirectoryRecursive
// Creates all components of a file path that might not already exist under the cache base path.
bool FileCache::CreateDirectoryRecursive(const string &path)
{
  string trimmed_path = RemoveTrailingSlash(RemoveLeadingSlash(path));
  size_t last_slash   = trimmed_path.find_last_of('\\');

  if(last_slash == string::npos && trimmed_path.length() > 0)
  {
    // Final directory component to create.
    return CreateDirectory(trimmed_path);
  }
  else
  {
    // Recurse to create parent directory first.
    if(!CreateDirectoryRecursive(trimmed_path.substr(0, last_slash)))
      return false;

    // Create this directory.
    return CreateDirectory(trimmed_path);
  }
}
//---------------------------------------------------------------------------------------------------------------------
// FileCache::CopyIntoCache
// Copies an external (non-versioned) file into the file cache's base directory.
// "file" should be an absolute path (or at least relative to the working directory); "relativedir" is the destination
// under the file cache base directory where the file should be copied.
bool FileCache::CopyIntoCache(const string &file, const string &relativedir)
{
  if(CreateDirectoryRecursive(relativedir))
  {
    pair<string, string> file_parts = GetFileSpec(file);
    string dest_path = PathConcatenate(PathConcatenate(GetBasePath(), relativedir), file_parts.second);

    return (CopyFile((LPCTSTR)file.c_str(), (LPCTSTR)dest_path.c_str(), false) != 0);
  }
  else
    return false; // failed to create that path, can't copy the file
}
//---------------------------------------------------------------------------------------------------------------------
// FileCache::DropFromCache
// Attempt to delete the local cached copy of a file. Returns true if a file was removed, and false if it didn't exist
// or couldn't be removed for whatever reason.
bool FileCache::DropFromCache(const string &filepath, const string &filename)
{
  bool removed_file = false;

  if(FileExistsInCache(filepath, filename))
  {
    string full_path = PathConcatenate(GetBasePath(), PathConcatenate(filepath, filename));
    removed_file = !remove(full_path.c_str());
  }

  return removed_file;
}
//---------------------------------------------------------------------------------------------------------------------
// FileCache::IncrementVersionNum
// Returns the next version number in string form.
string FileCache::IncrementVersionNum(string version1, string version2, string version3, string version4,
                                      int version1_increment, int version2_increment, int version3_increment,
                                      int version4_increment)
{
  int version1_temp = StringToInt(version1);
  int version2_temp = StringToInt(version2);
  int version3_temp = StringToInt(version3);
  int version4_temp = StringToInt(version4);

  version1_temp += version1_increment;
  version2_temp += version2_increment;
  version3_temp += version3_increment;
  version4_temp += version4_increment;

  version1 = IntToString(version1_temp);
  version2 = IntToString(version2_temp);
  version3 = IntToString(version3_temp);
  version4 = IntToString(version4_temp);

  return (version1 + "." + version2 + "." + version3 + "." + version4);
}
//---------------------------------------------------------------------------------------------------------------------
// FileCache::PushFileError
// Gets a descriptive error message for a push_file_status return code.
string FileCache::PushFileError(FileCache::push_file_status pfs)
{
  string error_message;

  switch(pfs)
  {
  case pf_ok:
    error_message = "No error";
    break;
  case pf_source_file_failure:
    error_message = "Could not open the source input file";
    break;
  case pf_version_exists:
    error_message = "This version already exists in the database";
    break;
  case pf_interbase_error:
    error_message = "An InterBase exception occurred";
    break;
  case pf_hash_calculation_error:
    error_message = "File is too large for hash computation";
    break;
  case pf_file_hash_mismatch:
    error_message = "Uploaded file did not match local file SHA-1 hash";
    break;
  case pf_unknown_error:
    error_message = "An unknown error has occurred";
    break;
  }
  return error_message;
}
/*---------------------------------------------------------------------------------------------------------------------
   FileCache::PushFileToDatabase
    dbUploadDatabase : required; should be already connected
    p_transaction    : optional; NULL or an existing transaction (will be activated if not NULL and not already Active)
    user_id          : required; ID of user inserting the file
    filename         : required; fully qualified filename (with path) to upload
    version          : required; version number to associate with this file
    download_path    : required; where the file will be located on the user's system, relative to the cache root
    comments         : optional; any comments to associate with the file blob in the DB

    The idea was that this could be used for supporting files, like:
    - DLL
    - GDS
    - MSI/EXE (push out a new version of Firebird, install in-place?)
    - images!  Right now the AST logo is embedded in the COB report, and can't be used on other screens, without
      embedding there too.  And it adds 4 MB (!!!) to the EXE size (before EXE-packing), when the image never even
      changes!  This is dumb!
    - scanned images - technically also images as above, but turning Prometheus into a Document Management System is a
      whole 'nother ballgame.
*/
FileCache::push_file_status
FileCache::PushFileToDatabase(TIBDatabase *dbUploadDatabase, TIBTransaction *p_transaction, string user_id,
                              string filename, string version, string download_path, string download_filename,
                              string comments, string &ib_error)
{
  push_file_status result = pf_unknown_error;
  string sql;
  ifstream inputFile;
  int first  = version.find_first_of(".");
  int second = version.find_first_of(".", first + 1);
  int third  = version.find_first_of(".", second + 1);
  bool free_transaction = false;

  TIBTransaction *dbUploadTransaction = p_transaction;
  TIBSQL         *dbUploadSQL;
  TIBDataSet     *dbUploadDataSet;
  if(!p_transaction)
  {
    dbUploadTransaction = new TIBTransaction(NULL);
    free_transaction    = true;
  }
  dbUploadSQL     = new TIBSQL(NULL);
  dbUploadDataSet = new TIBDataSet(NULL);

  try
  {
    dbUploadTransaction->DefaultDatabase = dbUploadDatabase;
    dbUploadSQL->Database                = dbUploadDatabase;
    dbUploadDataSet->Database            = dbUploadDatabase;
    StdTransaction(dbUploadTransaction, dbUploadDatabase);
    dbUploadSQL->Transaction             = dbUploadTransaction;
    dbUploadDataSet->Transaction         = dbUploadTransaction;

    sql = "select version_number from file_blobs where version_number = '" + version + "' "
          "and file_path = '" + download_path + "' and file_name = '" + download_filename + "'";
    dbUploadDataSet->SelectSQL->Add(sql.c_str());
    dbUploadDataSet->Open();

    if(dbUploadDataSet->Eof)
    {
      dbUploadDataSet->Close();

      inputFile.open(filename.c_str(), ios::in || ios::nocreate || ios::binary);
      if(!inputFile)
        throw pf_source_file_failure;

      // jhaley: compute SHA-1 hash of the file
      SHA1Hash fileHash;
      fileHash.HashFile(inputFile);
      if(!fileHash.IsValid())
        throw pf_hash_calculation_error; // This means the file is too large (> 512 MB, a VERY generous limit IMHO)

      // jhaley - GetNextId added as requested by numerous TODO's that were here before.
      string file_id = GetNextId(dbUploadTransaction, "file_blobs");
      sql = "insert into file_blobs (id, last_update_stamp, last_updated_by, valid, version_number, version_number1, "
            "version_number2, version_number3, version_number4, file_path, file_name, sha1_hash, comments) values (" +
             file_id + ", '" + RightNow() + "', " + user_id + ", 1, '" + version + "', " +
             version.substr(0, first) + ", " +
             version.substr(first + 1, (second - first) - 1) + ", " +
             version.substr(second + 1, (third - second) - 1) + ", " +
             version.substr(third + 1) + ", '" + download_path + "', '" + download_filename + "', '" +
             fileHash.ToString() + "', '" + comments + "')";
      dbUploadSQL->SQL->Text = sql.c_str();
      dbUploadSQL->ExecQuery();

      // needed for the bit inside the if {} below
      sql = "select * from file_blobs where id = " + file_id;
      dbUploadDataSet->SelectSQL->Clear();
      dbUploadDataSet->SelectSQL->Add(sql.c_str());

      sql = "update file_blobs set file_blob = :file_blob where id = " + file_id;
      dbUploadDataSet->ModifySQL->Add(sql.c_str());

      dbUploadDataSet->Open();
      dbUploadDataSet->Edit();

      inputFile.close();
      dbUploadDataSet->FieldByName("file_blob")->SetFieldType(ftBlob);
      ((TBlobField *)dbUploadDataSet->FieldByName("file_blob"))->LoadFromFile(filename.c_str());
      dbUploadDataSet->Post();
      dbUploadDataSet->Close();

      // jhaley 20120229: Removed SHA-1 verification of upload as doesn't always work >_>
      result = pf_ok; // All is well.
      dbUploadTransaction->Commit(); // jhaley 20111220: probably best to commit the transaction @_@
    }
    else
    {
      result = pf_version_exists;
      dbUploadDataSet->Close();
      dbUploadTransaction->Rollback();
    }
  }
  catch(EIBError &error) // Database problem
  {
    if(dbUploadTransaction->Active)
      dbUploadTransaction->Rollback();
    result = pf_interbase_error;
    ib_error = error.Message.c_str();
  }
  catch(push_file_status pfs) // Other problem (user error most likely)
  {
    if(dbUploadTransaction->Active)
      dbUploadTransaction->Rollback();
    result = pfs;
  }
  catch(...) // Unknown problem (the worst kind of problem)
  {
    if (dbUploadTransaction->Active)
      dbUploadTransaction->Rollback();
    result = pf_unknown_error;
  }

  delete dbUploadDataSet;
  delete dbUploadSQL;

  // if we didn't allocate it, we don't free it
  if (free_transaction)
    delete dbUploadTransaction;

  return result;
}
//---------------------------------------------------------------------------------------------------------------------
// FileCache::GetFileFromDatabase
// Retrieves the newest version of the file available matching both the path and the name loaded into a TMemoryStream
// object.
TMemoryStream *FileCache::GetFileFromDatabase(TIBDatabase *dbDownloadDatabase, TIBTransaction *p_transaction,
                                              string filepath, string filename, string &ib_error)
{
  string sql;
  TMemoryStream  *reader = NULL;
  TIBTransaction *dbDownloadTransaction = p_transaction;
  TIBSQL         *dbDownloadSQL;
  TIBDataSet     *dbDownloadDataSet;
  bool            free_transaction = false;

  if(!p_transaction)
  {
    dbDownloadTransaction = new TIBTransaction(NULL);
    free_transaction = true;
  }
  dbDownloadSQL     = new TIBSQL(NULL);
  dbDownloadDataSet = new TIBDataSet(NULL);

  try
  {
    dbDownloadSQL->Database     = dbDownloadDatabase;
    dbDownloadDataSet->Database = dbDownloadDatabase;

    if(!dbDownloadTransaction->Active)
    {
      dbDownloadTransaction->DefaultDatabase = dbDownloadDatabase;
      StdTransaction(dbDownloadTransaction, dbDownloadDatabase);
    }

    dbDownloadDataSet->Transaction = dbDownloadTransaction;
    dbDownloadSQL->Transaction     = dbDownloadTransaction;

    sql = "select first 1 file_blob from file_blobs "
          "where file_path = '" + filepath + "' and file_name = '" + filename + "' "
          "order by version_number1 desc, version_number2 desc, version_number3 desc, version_number4 desc ";
    dbDownloadDataSet->SelectSQL->Clear();
    dbDownloadDataSet->SelectSQL->Add(sql.c_str());
    dbDownloadDataSet->Open();
    dbDownloadDataSet->First();

    if(!dbDownloadDataSet->Eof)
    {
      TBlobField *blob = (TBlobField *)(dbDownloadDataSet->FieldByName("file_blob"));
      reader = new TMemoryStream();
      blob->SaveToStream(reader);
    }

    dbDownloadDataSet->Close();
  }
  catch(EIBError &error)
  {
    if(free_transaction && dbDownloadTransaction->Active)
      dbDownloadTransaction->Rollback();
    ib_error = error.Message.c_str();

    if(reader)
    {
      delete reader;
      reader = NULL;
    }
  }
  catch(...)
  {
    if(free_transaction && dbDownloadTransaction->Active)
      dbDownloadTransaction->Rollback();

    if(reader)
    {
      delete reader;
      reader = NULL;
    }
  }

  delete dbDownloadSQL;
  delete dbDownloadDataSet;
  if(free_transaction)
    delete dbDownloadTransaction;

  return reader;
}
//---------------------------------------------------------------------------------------------------------------------
// FileCache::CheckIsNewest
// Simply checks if a local file matches the newest version of the file in the database, by SHA-1 hash.
bool FileCache::CheckIsNewest(TIBDatabase *db, string filepath, string filename)
{
  string full_file_path = PathConcatenate(GetBasePath(), PathConcatenate(filepath, filename));
  bool is_newest = false;

  // Check if it even exists first, for speed.
  if(!FileExistsInCache(filepath, filename))
    return false;

  // It does exist. Get the SHA-1 hash of the newest version from the indicated database.
  try
  {
    string newest_hash = GetOneField(db,
      "select first 1 sha1_hash from file_blobs "
      "where file_path = '" + filepath + "' and file_name = '" + filename + "' "
      "order by version_number1 desc, version_number2 desc, version_number3 desc, version_number4 desc ");

    if(newest_hash == "") // no such file in db??
      return false;

    ifstream inputFile;
    inputFile.open(full_file_path.c_str(), ios::in | ios::nocreate | ios::binary);

    if(inputFile) // Successfully opened...
    {
      // ...But is it the newest version?
      SHA1Hash dbHash, localHash;

      dbHash.FromString(newest_hash); // recreate serialized SHA-1 hash from database
      localHash.HashFile(inputFile);  // hash the local version of the file

      inputFile.close();

      is_newest = (localHash == dbHash); // Hash match?
    }
    else
      is_newest = false; // Really shouldn't happen.
  }
  catch(...)
  {
    is_newest = false; // Cannot safely say it is the newest version, since something blew up.
  }

  return is_newest;
}
//---------------------------------------------------------------------------------------------------------------------
// FileCache::GetCacheableFile
// This routine will either get a local, cached copy of a file, or it will retrieve the newest version of the file from
// the database and attempt to save it for later reuse. If the saving portion of the process fails, the in-memory
// version loaded from the database will be kept. As such, the object returned is either a TMemoryStream or TFileStream.
TStream *FileCache::GetCacheableFile(TIBDatabase *db, TIBTransaction *transact, string filepath, string filename,
                                     get_file_flags &flags)
{
  TStream *reader = NULL;

  // jhaley - construct the full file path for writing, distinct from the components passed in.
  // The Prometheus file cache's base path is concatenated automatically and should neither be passed in, nor stored
  // in the database. 
  // Ex: Given the following:
  // * filepath = 'Prometheus\images'
  // * filename = 'astlogo.bmp'
  // * GetBasePath() = 'C:\Users\FooBar\AppData\Local'
  // The result will be C:\Users\FooBar\AppData\Local\Prometheus\images\astlogo.bmp
  string full_file_path = PathConcatenate(GetBasePath(), PathConcatenate(filepath, filename));

  try
  {
    // Get info on the newest version of this file from the database
    string db_file_hash = GetOneField(db,
      "select first 1 sha1_hash from file_blobs "
      "where file_path = '" + filepath + "' and file_name = '" + filename + "' "
      "order by version_number1 desc, version_number2 desc, version_number3 desc, version_number4 desc ");

    if(db_file_hash == "") // no such file??
      throw 1;

    ifstream inputFile;
    inputFile.open(full_file_path.c_str(), ios::in | ios::nocreate | ios::binary);

    if(inputFile) // File exists...
    {
      // ...But is it the newest version?
      SHA1Hash dbHash, localHash;

      dbHash.FromString(db_file_hash); // recreate serialized SHA-1 hash from database
      localHash.HashFile(inputFile);   // hash the local version of the file

      inputFile.close();

      if(localHash == dbHash) // Hash match? Return the disk file!
        reader = new TFileStream(full_file_path.c_str(), fmOpenRead|fmShareDenyWrite);

      flags = gf_sourcedisk;
    }

    // If reader wasn't set, the file either doesn't exist or isn't the newest version
    if(!reader)
    {
      string ib_error;
      TMemoryStream *mem_reader = GetFileFromDatabase(db, transact, filepath, filename, ib_error);
      // Try writing it to disk for next time
      try
      {
        if(mem_reader)
        {
          flags = gf_sourcememory;

          // Create the required directory structure under the cache base folder first. If not successful, there is no
          // point in even trying the file write.
          if (CreateDirectoryRecursive(filepath))
          {
            mem_reader->SaveToFile(full_file_path.c_str()); // Note throws an exception if the write fails
            flags |= gf_wrotetodisk;
          }

          mem_reader->Seek(soFromBeginning, 0);       // Seek back to the start.
        }
        else
        {
#ifdef use_prometheus_util
           string error_msg = string("Failure in GetCacheableFile for '") + full_file_path + "' on " + computer_name;
           if(ib_error != "")
             error_msg += " (ib_error = '" + ib_error + "')";
           if(errno)
             error_msg += string(" (stdlib error = '") + strerror(errno) + "')";
           LogIt("cached_files.log", error_msg);
           ::PPretendToSay(error_msg, m_debug, "FileCache::GetCacheableFile");
#else
           Application->MessageBox(error_msg, "File Download Error", MB_OK);
#endif
        }
      }
      catch(...)
      {
        // Do not care if it fails here. We have our memory-cached copy to work with, and we will return it as-is.
        // We DO, however, want notification of the fact when such is available, because this could be a serious
        // problem for end-users in Little Axe.
#ifdef use_prometheus_util
        string error_msg = string("Failed to save downloaded file '") + full_file_path + "' on " + computer_name;
        if(errno)
          error_msg += string(" (stdlib error = '") + strerror(errno) + "')";
        LogIt("cached_files.log", error_msg);
        ::PPretendToSay(error_msg, m_debug, "FileCache::GetCacheableFile");
#else
        Application->MessageBox(error_msg, "File Write Error", MB_OK);
#endif
      }
      reader = mem_reader;
      // TODO: Do something with ib_error?...
    }
  }
  catch(...)
  {
#ifdef use_prometheus_util
    string error_msg = string("Failed to download file '") + full_file_path + "' on " + computer_name;
    if(errno)
      error_msg += string(" (stdlib error = '") + strerror(errno) + "')";
    LogIt("cached_files.log", error_msg);
    ::PPretendToSay(error_msg, m_debug, "FileCache::GetCacheableFile");
#else
    Application->MessageBox(error_msg, "File Download Error", MB_OK);
#endif
    if(reader) // Something went awry, don't return whatever corrupt data we might have happened to read
    {
      delete reader;
      reader = NULL;
    }
  }

  return reader;
}
//---------------------------------------------------------------------------------------------------------------------
// FileCache::GetBitmapFromCache
// Retrieve a cacheable database file into an existing TBitmap - very handy for TImage / TQRImage controls. Always check
// the return value, as it will return NULL if any type of error occurred.
bool FileCache::GetBitmapFromCache(Graphics::TBitmap *gBitmap, TIBDatabase *db, TIBTransaction *transact,
                                   string filepath, string filename)
{
  bool success = false;
  TStream *reader = NULL;
  get_file_flags flags = gf_noresult;

  try
  {
    // Get the cacheable file. This will either come from disk, or it will come from the database. We don't care which.
    // Whenever we download it, it will attempt to be written to disk, and that's taken care of above.
    reader = GetCacheableFile(db, transact, filepath, filename, flags);
    if(reader)
    {
      gBitmap->LoadFromStream(reader);
      success = true;
      delete reader;
    }
  }
  catch(...)
  {
    delete reader;
  }

  return success;
}
//---------------------------------------------------------------------------------------------------------------------
// FileCache::GetBitmapFromCache
// This overload of the above will create and return a new TBitmap, loaded from a cacheable database file. You are
// responsible for freeing the TBitmap when you're done with it.
Graphics::TBitmap *FileCache::GetBitmapFromCache(TIBDatabase *db, TIBTransaction *transact, string filepath,
                                                 string filename)
{
  Graphics::TBitmap *bitmap = new Graphics::TBitmap();

  if(!GetBitmapFromCache(bitmap, db, transact, filepath, filename))
  {
    // Failed; delete and return NULL
    delete bitmap;
    bitmap = NULL;
  }

  return bitmap;
}
//---------------------------------------------------------------------------------------------------------------------
// FileCache::GetJPEGFromCache
// Retrieve a cacheable database file into an existing TPicture - very handy for TImage / TQRImage controls.
bool FileCache::GetJPEGFromCache(Graphics::TPicture *gPicture, TIBDatabase *db, TIBTransaction *transact,
                                 string filepath, string filename)
{
  bool success = false;
  TStream *reader = NULL;
  get_file_flags flags = gf_noresult;

  try
  {
    // Get the cacheable file. This will either come from disk, or it will come from the database. We don't care which.
    // Whenever we download it, it will attempt to be written to disk, and that's taken care of above.
    reader = GetCacheableFile(db, transact, filepath, filename, flags);
    if(reader)
    {
      Jpeg::TJPEGImage *Jpg = new Jpeg::TJPEGImage();
      try
      {
        Jpg->LoadFromStream(reader);
        gPicture->Graphic = Jpg;
        success = true;
      }
      catch(...)
      {
        success = false;
        delete Jpg;
      }
      delete reader;
    }
  }
  catch(...)
  {
    delete reader;
  }

  return success;
}
//---------------------------------------------------------------------------------------------------------------------
// *** SHA1Hash Routines **********************************************************************************************
// For data verification and identification purposes, NOT requiring any external information source, such as hacking
// the heck out of the Prometheus config file. This is derived from my pre-existing personal implementation, which for
// purposes of inclusion here, I grant unlimited license for any use whatsoever. Known to be "right", and very fast.
//---------------------------------------------------------------------------------------------------------------------
SHA1Hash::SHA1Hash()
{
  memset(message, 0, SHA1BLOCKSIZE);

  // Start with the standard initialization vector for digest
  digest[0] = 0x67452301;
  digest[1] = 0xEFCDAB89;
  digest[2] = 0x98BADCFE;
  digest[3] = 0x10325476;
  digest[4] = 0xC3D2E1F0;

  messagelen = 0;
  messageidx = 0;
  gonebad    = false;
}
//---------------------------------------------------------------------------------------------------------------------
SHA1Hash::SHA1Hash(const SHA1Hash &other)
{
  Assign(other);
}
//---------------------------------------------------------------------------------------------------------------------
// SHA1Hash::ProcessBlock
// Called when the SHA-1 hash has accumulated 512 bits of input. The main "block cipher" part of the algorithm.
#define leftrotate(w, b) (((w) << (b)) | ((w) >> (32 - (b))))
void SHA1Hash::ProcessBlock()
{
  int i;
  sha1dword temp;
  sha1dword work[80];       // 80-word extension workspace
  sha1dword a, b, c, d, e;  // current hash vector for chunk
  sha1dword f;              // the "f" function, which varies on every 20 subsequent words
  sha1byte *msg = message;

  // Step 1: turn the 64 bytes into 16 DWORDs with SHA1 big-endian byte ordering
  for(i = 0; i < 16; i++)
  {
    work[i]  = ((size_t)msg[i * 4    ]) << 24;
    work[i] |= ((size_t)msg[i * 4 + 1]) << 16;
    work[i] |= ((size_t)msg[i * 4 + 2]) <<  8;
    work[i] |= ((size_t)msg[i * 4 + 3]);
  }

  // Step 2: extend the message with XOR shuffling
  for(; i < 80; i++)
  {
    temp = work[i - 3] ^ work[i - 8] ^ work[i - 14] ^ work[i - 16];

    work[i] = leftrotate(temp, 1);
  }

  // hash for this chunk starts out as current message hash
  a = digest[0];
  b = digest[1];
  c = digest[2];
  d = digest[3];
  e = digest[4];

  // Step 3: Compute hash
  for(i = 0; i < 20; i++)
  {
    f = (b & c) | (~b & d);

    temp = leftrotate(a, 5) + f + e + 0x5A827999 + work[i];
    e = d;
    d = c;
    c = leftrotate(b, 30);
    b = a;
    a = temp;
  }
  for(; i < 40; i++)
  {
    f  = b ^ c ^ d;

    temp = leftrotate(a, 5) + f + e + 0x6ED9EBA1 + work[i];
    e = d;
    d = c;
    c = leftrotate(b, 30);
    b = a;
    a = temp;
  }
  for(; i < 60; i++)
  {
    f = (b & c) | (b & d) | (c & d);

    temp = leftrotate(a, 5) + f + e + 0x8F1BBCDC + work[i];
    e = d;
    d = c;
    c = leftrotate(b, 30);
    b = a;
    a = temp;
  }
  for(; i < 80; i++)
  {
    f = b ^ c ^ d;

    temp = leftrotate(a, 5) + f + e + 0xCA62C1D6 + work[i];
    e = d;
    d = c;
    c = leftrotate(b, 30);
    b = a;
    a = temp;
  }

  // Step 4: add chunk hash to message hash
  digest[0] += a;
  digest[1] += b;
  digest[2] += c;
  digest[3] += d;
  digest[4] += e;

  // current buffered block has been digested, so clear the index
  messageidx = 0;
}
//---------------------------------------------------------------------------------------------------------------------
// SHA1Hash::HashData
// Given data, this function will buffer it in the SHA1Hash object until 512 bytes of input have been accumulated, and
// then it will compute the hash for that block. If data is > 512 bits, it will be processed 512 bits at a time. Call
// this routine repeatedly until all input is exhausted, or send it all input at once.
void SHA1Hash::HashData(const SHA1Hash::sha1byte *data, size_t len)
{
  if(!data || !len || gonebad)
    return;

  while(len--)
  {
    message[messageidx++] = *data;

    // added 8 bits
    messagelen += 8;

    // If messagelen wraps to 0, the message is too long - supports a maximum message length of 512 MB
    if(!messagelen)
    {
      gonebad = true;
      break;
    }

    // when 512 bits (64 bytes) of message has been accumulated, process it.
    if(messageidx == SHA1BLOCKSIZE)
      ProcessBlock();

    ++data;
  }
}
//---------------------------------------------------------------------------------------------------------------------
// SHA1Hash::HashFile
// Reads a file into a temporary buffer and computes the SHA-1 hash.
void SHA1Hash::HashFile(ifstream &inputFile)
{
  sha1byte *buffer;
  size_t filelen;

  // Get file length and allocate a buffer
  inputFile.seekg(0, ios::end);
  filelen = inputFile.tellg();
  inputFile.seekg(0, ios::beg);
  buffer = new sha1byte [filelen];

  // Read in the full file
  inputFile.read(buffer, filelen);

  // Hash it.
  HashData(buffer, filelen);
  WrapUp();

  // rewind the file stream and dump the buffer
  inputFile.seekg(0, ios::beg);
  delete [] buffer;
}
//---------------------------------------------------------------------------------------------------------------------
// SHA1Hash::WrapUp
// Finishes the SHA-1 hash algorithm by padding the message as needed and completing the hash computation on any data
// left in the message buffer.
void SHA1Hash::WrapUp()
{
  // The message must be padded out to an even multiple of 512 bits, and the padding must include an initial "1" bit
  // followed by zeroes, and then wrapped up with 8 bytes representing the length of the message. Given the 32-bit
  // length limitation, the first 4 bytes of the message length will also always be zero. This means we need 9 bytes
  // available in the buffer. If there aren't 9 bytes available, pad out the current block and continue padding into
  // a new block.

  if(messageidx > SHA1BLOCKSIZE - 9)
  {
    message[messageidx++] = 0x80;
    while(messageidx < SHA1BLOCKSIZE)
      message[messageidx++] = 0;

    // we filled the current block, so process it
    ProcessBlock();

    // finish padding in the new block with zeroes
    while(messageidx <= SHA1BLOCKSIZE - 5)
      message[messageidx++] = 0;
  }
  else
  {
    // we have enough room in the current block
    message[messageidx++] = 0x80;
    while(messageidx <= SHA1BLOCKSIZE - 5)
      message[messageidx++] = 0;
  }

  // fill the remainder of the final block with the encoded length
  message[60] = (sha1byte)((messagelen >> 24) & 0xFF);
  message[61] = (sha1byte)((messagelen >> 16) & 0xFF);
  message[62] = (sha1byte)((messagelen >>  8) & 0xFF);
  message[63] = (sha1byte)((messagelen      ) & 0xFF);

  // process the final padded block.
  ProcessBlock();
}
//---------------------------------------------------------------------------------------------------------------------
// SHA1Hash::Assign
// Implemention for copy constructor and operator =
void SHA1Hash::Assign(const SHA1Hash &other)
{
  memcpy(digest,  other.digest,  SHA1NUMDIGEST * sizeof(size_t));
  memcpy(message, other.message, SHA1BLOCKSIZE);

  messagelen = other.messagelen;
  messageidx = other.messageidx;
  gonebad    = other.gonebad;
}
//---------------------------------------------------------------------------------------------------------------------
// SHA1Hash::Compare
// Compares two SHA1Hash objects for equivalency. If either are "bad" hashes, the comparison evaluates as false.
// Also serves as the implementation for operator == and operator !=
bool SHA1Hash::Compare(const SHA1Hash &other) const
{
  int nummatch = 0;

  // We can't compare corrupted hashes
  if(gonebad || other.gonebad)
    return false;

  for(int i = 0; i < SHA1NUMDIGEST; i++)
    if(digest[i] == other.digest[i])
      ++nummatch;

  return (nummatch == SHA1NUMDIGEST);
}
//---------------------------------------------------------------------------------------------------------------------
// SHA1Hash::ToString
// We do not rely on standard library or external Prometheus utilities to do conversion to/from hex strings, because
// SHA-1 hashes have special semantics with respect to length, case and leading zeroes that must be carefully respected.
string SHA1Hash::ToString() const
{
  string output;

  for(int i = 0; i < SHA1NUMDIGEST; i++)
  {
    for(int shift = 28; shift >= 0; shift -= 4)
      output.push_back("0123456789abcdef"[(digest[i] >> shift) & 0x0F]);
  }

  return output;
}
//---------------------------------------------------------------------------------------------------------------------
// SHA1Hash::FromString
// We do not rely on standard library or external Prometheus utilities to do conversion to/from hex strings, because
// SHA-1 hashes have special semantics with respect to length, case and leading zeroes that must be carefully respected.
// Also serves as the implementation for operator = (string &)
#define SHA1HexCharToNum(c) \
   (sha1dword)(((c) >= '0' && (c) <= '9') ?  (c) - '0' : \
               ((c) >= 'a' && (c) <= 'f') ? ((c) - 'a') + 10 : \
               ((c) >= 'A' && (c) <= 'F') ? ((c) - 'A') + 10 : 0)
void SHA1Hash::FromString(const string &str)
{
  int digestidx = 0;
  int digestshift = 28;
  size_t stringlen = str.length();

  memset(digest, 0, sizeof(digest));

  for(size_t i = 0; i < stringlen; i++)
  {
    if(digestidx >= SHA1NUMDIGEST)
      return;

    digest[digestidx] |= (SHA1HexCharToNum(str[i]) << digestshift);
    digestshift -= 4;

    if(digestshift < 0) // step to next part of digest
    {
      ++digestidx;
      digestshift = 28;
    }
  }
}
//---------------------------------------------------------------------------------------------------------------------
SHA1Hash &SHA1Hash::operator = (const SHA1Hash &other)
{
  Assign(other);
  return *this;
}
//---------------------------------------------------------------------------------------------------------------------
SHA1Hash &SHA1Hash::operator = (const string &str)
{
  FromString(str);
  return *this;
}
//---------------------------------------------------------------------------------------------------------------------
bool SHA1Hash::operator == (const SHA1Hash &other) const
{
  return Compare(other);
}
//---------------------------------------------------------------------------------------------------------------------
bool SHA1Hash::operator != (const SHA1Hash &other) const
{
  return !Compare(other);
}
//---------------------------------------------------------------------------------------------------------------------
