#ifndef cached_files_h
#define cached_files_h

#include <string>
#include <Classes.hpp>
#include <Controls.hpp>
#include <StdCtrls.hpp>
#include <Db.hpp>
#include <IBCustomDataSet.hpp>
#include <IBDatabase.hpp>
#include <IBSQL.hpp>
#include <map>
#include <vector>
#include <set>
#include <list>
#ifdef use_tcpip_messaging
#include "tcpip_messaging.h"
#endif
#ifdef use_prometheus_util
#include "prometheus_util.h"
#endif
#include "util.h"
#include "allsymbols.h"

using namespace std;

namespace FileCache
{
  enum push_file_status
  {
    pf_ok,
    pf_source_file_failure,
    pf_version_exists,
    pf_interbase_error,
    pf_hash_calculation_error,
    pf_file_hash_mismatch,
    pf_unknown_error
  };

  enum get_file_flags
  {
    gf_noresult     = 0,
    gf_sourcedisk   = 0x0001, // read the file from disk
    gf_sourcememory = 0x0002, // read the file from the database
    gf_wrotetodisk  = 0x0004  // asserted if the file got written to the disk
  };

  // Utilities
  string GetBasePath(void);
  string RemoveTrailingSlash(const string &source_path);
  string RemoveLeadingSlash(const string &source_path);
  string PathConcatenate(const string &base, const string &newpart);
  bool   DirectoryExists(const string &path, bool test_if_full_perms);
  bool   FileExists(const string &path, const string &file);
  bool   FileExistsInCache(const string &path, const string &file);
  bool   CreateDirectory(const string &dir_name);
  bool   CreateDirectoryRecursive(const string &path);
  bool   CopyIntoCache(const string &file, const string &relativedir);
  bool   DropFromCache(const string &path, const string &file);
  bool   CheckIsNewest(TIBDatabase *db, string filepath, string filename);
  string IncrementVersionNum(string version1, string version2, string version3, string version4, int version1_increment,
                             int version2_increment, int version3_increment, int version4_increment);
  string PushFileError(FileCache::push_file_status pfs);
  pair<string, string> GetFileSpec(const string &full_path);

  // Upload a file that can be retrieved and cached later
  push_file_status PushFileToDatabase(TIBDatabase *dbUploadDatabase, TIBTransaction *p_transaction, string user_id,
    string filename, string version, string download_path, string download_filename, string comments, string &ib_error);

  // Get a file from the database
  TMemoryStream *GetFileFromDatabase(TIBDatabase *dbDownloadDatabase, TIBTransaction *p_transaction,
                                     string filepath, string filename, string &ib_error);

  // Get a cacheable file. This will either come from the database or from the local cache on disk.
  TStream *GetCacheableFile(TIBDatabase *db, TIBTransaction *transact, string filepath, string filename,
                            get_file_flags &flags);

  // Get a bitmap from the file cache, and load it into an existing TBitmap.
  bool GetBitmapFromCache(Graphics::TBitmap *gBitmap, TIBDatabase *db, TIBTransaction *transact, string filepath,
                          string filename);

  // Get a bitmap from the file cache, and load it into a newly allocated TBitmap. The caller must free it.
  Graphics::TBitmap *GetBitmapFromCache(TIBDatabase *db, TIBTransaction *transact, string filepath, string filename);

  // Get a JPEG image from the file cache, and load it into an existing TPicture.
  bool GetJPEGFromCache(Graphics::TPicture *gPicture, TIBDatabase *db, TIBTransaction *transact,
                        string filepath, string filename);
};

#define SHA1NUMDIGEST 5
#define SHA1BLOCKSIZE 64

// SHA-1 hashes are used to do two things by the FileCache routines:
// * Verifying that the upload was fully successful.
// * Verifying that the user's cached version of the file matches the latest one in the database. This is MUCH more
//   clean and foolproof than trying to use the configuration file to store versions, which was my first idea.
// This code is general enough to use anywhere else in Prometheus, too.
class SHA1Hash
{
public:
  typedef unsigned char sha1byte;
  typedef unsigned int  sha1dword; // Note: subject to 64-bit portability concerns (really needs C99 uint32_t)

protected:
  sha1dword digest[SHA1NUMDIGEST];
  sha1dword messagelen;
  sha1byte  message[SHA1BLOCKSIZE];
  int       messageidx;
  bool      gonebad;

  void ProcessBlock();

public:
  SHA1Hash();
  SHA1Hash(const SHA1Hash &other);

  // Public Methods
  void   HashData(const sha1byte *data, size_t len);
  void   HashFile(ifstream &inputFile);
  void   WrapUp();
  void   Assign(const SHA1Hash &other);
  bool   Compare(const SHA1Hash &other) const;
  string ToString() const;
  void   FromString(const string &str);

  // Operators
  SHA1Hash &operator  = (const SHA1Hash &other);
  SHA1Hash &operator  = (const string &str);
  bool      operator == (const SHA1Hash &other) const;
  bool      operator != (const SHA1Hash &other) const;

  // Inlines
  bool IsValid() { return !gonebad; }
};

#endif

// EOF

