#ifndef control_break_h
#define control_break_h

#include <vector>
#include <map>
#include <algorithm>

using namespace std;

//---------------------------------------------------------------------------

template <class value_type>
class cb_sorting_object
{
 public:
  virtual bool operator () (const value_type & lhs, const value_type & rhs) = 0;
};

//---------------------------------------------------------------------------

template <class value_type>
class cb_aggregate_object
{
 public:
  virtual value_type get_value () = 0;
  virtual void set_value (const value_type & new_base_value) = 0;
  virtual void add_value (const value_type & next_row_value) = 0;
};

//---------------------------------------------------------------------------

template <class key_type, class value_type>
bool is_row_different(const map<key_type, value_type> & current_values, const map<key_type, value_type> & previous_values, const vector<pair<key_type, cb_sorting_object<value_type> *> > & fields_to_match)
{
 bool are_different = false;
 map<key_type, value_type>::const_iterator i_curr, i_prev;

 for (vector<pair<key_type, cb_sorting_object<value_type> *> >::const_iterator i_field = fields_to_match.begin(); i_field != fields_to_match.end(); i_field++)
 {
  i_curr = current_values.find(i_field->first);
  i_prev = previous_values.find(i_field->first);
  if (i_curr == current_values.end())
   throw(404);

  if (i_prev != previous_values.end() && ((*(i_field->second))(i_curr->second, i_prev->second) || (*(i_field->second))(i_prev->second, i_curr->second)))
  {
   are_different = true;
   break;
  }
 }
 return are_different;
}

//---------------------------------------------------------------------------

template <class key_type, class value_type>
class multi_column_sorting_object
{
 private:
  vector<pair<key_type, cb_sorting_object<value_type> *> > field_order;
 public:
  multi_column_sorting_object (const vector<pair<key_type, cb_sorting_object<value_type> *> > & p_field_order) {field_order = p_field_order;}

  bool operator () (const map<key_type, value_type> & lhs, const map<key_type, value_type> & rhs)
  {
   bool lhs_smaller_than_rhs = false;
   map<key_type, value_type>::const_iterator i_lhs, i_rhs;

   for (vector<pair<key_type, cb_sorting_object<value_type> *> >::iterator i_field = field_order.begin(); i_field != field_order.end(); i_field++)
   {
    i_lhs = lhs.find(i_field->first);
    i_rhs = rhs.find(i_field->first);
    if (i_lhs == lhs.end() || i_rhs == rhs.end())
     throw(403);
    if ((*(i_field->second))(i_lhs->second, i_rhs->second))
    {
     // lhs < rhs, so we stop comparing columns here -and- declare lhs < rhs
     lhs_smaller_than_rhs = true;
     break;
    } else
    if ((*(i_field->second))(i_rhs->second, i_lhs->second))
    {
     // rhs < lhs, so we stop comparing columns here -and- declare !(lhs < rhs)
     break;
    }
   }
   return lhs_smaller_than_rhs;
  }
};

//---------------------------------------------------------------------------

template <class key_type, class value_type>
void do_control_break(const vector<map<key_type, value_type> > & input_data, vector<map<key_type, value_type> > & output_data, const vector<pair<key_type, cb_sorting_object<value_type> *> > & grouping_fields, const map<key_type, cb_aggregate_object<value_type> *> & summing_fields)
{
 // first, sort the given data by all the grouping fields, so we can do control-break processing
 output_data.clear();
 vector<map<key_type, value_type> > local_data = input_data;
 multi_column_sorting_object<key_type, value_type> sorting_object(grouping_fields);
 sort(local_data.begin(), local_data.end(), sorting_object);

 map<key_type, value_type> previous_values;
 map<key_type, value_type> output_values;

 // then, for all rows, see if they differ by the grouping fields. if so, save the grouping_field/summing_field data, and start a new group.
 for (vector<map<key_type, value_type> >::iterator i_row = local_data.begin(); i_row != local_data.end(); i_row++)
 {
  // the first row will never be considered different ... (i hope)
  if (is_row_different((*i_row), previous_values, grouping_fields))
  {
   // they're different, so we take the current summary data and grouping fields, and stuff them into an output row
   output_values.clear();
   // use the previous row's data to fill in the grouping fields
   for (vector<pair<key_type, cb_sorting_object<value_type> *> >::const_iterator i_field = grouping_fields.begin(); i_field != grouping_fields.end(); i_field++)
    output_values.insert(pair<key_type, value_type>(i_field->first, previous_values[i_field->first]));
   // then use aggregates to fill in the summary fields, and reset the aggregate objects
   for (map<key_type, cb_aggregate_object<value_type> *>::const_iterator i_field = summing_fields.begin(); i_field != summing_fields.end(); i_field++)
   {
    output_values.insert(pair<key_type, value_type>(i_field->first, i_field->second->get_value()));
    i_field->second->set_value((*i_row)[i_field->first]);
   }
   output_data.push_back(output_values);

   // keep this around for future reference!
   previous_values = (*i_row);
  } else
  {
   if (i_row == local_data.begin())
   {
    // if it's the first row (not different) then we reset to this row, rather than add this row
    for (map<key_type, cb_aggregate_object<value_type> *>::const_iterator i_field = summing_fields.begin(); i_field != summing_fields.end(); i_field++)
     i_field->second->set_value((*i_row)[i_field->first]);
    // as we're on the first row, we also need to set previous_values up with some defaults!
    previous_values = (*i_row);
   } else
   {
    // if it's not the first row, and it's more of the same, then for all aggregates, add the values from this new row
    for (map<key_type, cb_aggregate_object<value_type> *>::const_iterator i_field = summing_fields.begin(); i_field != summing_fields.end(); i_field++)
     i_field->second->add_value((*i_row)[i_field->first]);
   }
  }
 }

 // if there was anything at all, the last group is still pending
 if (local_data.size())
 {
  output_values.clear();
  // use the previous row's data to fill in the grouping fields
  for (vector<pair<key_type, cb_sorting_object<value_type> *> >::const_iterator i_field = grouping_fields.begin(); i_field != grouping_fields.end(); i_field++)
   output_values.insert(pair<key_type, value_type>(i_field->first, previous_values[i_field->first]));
  // then use aggregates to fill in the summary fields, and reset the aggregate objects
  for (map<key_type, cb_aggregate_object<value_type> *>::const_iterator i_field = summing_fields.begin(); i_field != summing_fields.end(); i_field++)
   output_values.insert(pair<key_type, value_type>(i_field->first, i_field->second->get_value()));
  output_data.push_back(output_values);
 }
 return;
}

//---------------------------------------------------------------------------

template <class key_type, class value_type>
void cb_deallocate_sorting_map(vector<pair<key_type, cb_sorting_object<value_type> *> >& fields_to_clear)
{
 for (vector<pair<key_type, cb_sorting_object<value_type> *> >::iterator i = fields_to_clear.begin(); i != fields_to_clear.end(); i++)
  delete (i->second);
}

//---------------------------------------------------------------------------

template <class key_type, class value_type>
void cb_deallocate_aggregate_map(map<key_type, cb_aggregate_object<value_type> *> & fields_to_clear)
{
 for (map<key_type, cb_aggregate_object<value_type> *>::iterator i = fields_to_clear.begin(); i != fields_to_clear.end(); i++)
  delete (i->second);
}

//---------------------------------------------------------------------------

/* TODO: reporting version that does -not- eliminate detail rows,
   but rather lets you specify -when- it will output rows, how much,
   and when to reset aggregates (of aggregates?)
   (replace quick reports? hehehehe) */

/* TODO: if you want both a sum() and and avg() on the same field,
   you have to first create the other field and populate it with data from
   the first field, -then- run the sum() and avg() against the two fields.
   there should be some way to specify in- and out- field-names ...
   except that's a pain in the arse. */

// really, we need expressions of some sort. and some way to use them.

#endif
