#ifndef control_break_objects_h
#define control_break_objects_h

#include "control_break.h"

// wrappers for sorting and aggregates come first
// sorting objects come second (derived from cb_sorting_object)
// aggregate objects come third (derived from cb_aggregate_object)
// templated objects shall be in the .h file, type-specific objects shall reside in the .cpp file!

//---------------------------------------------------------------------------

template <class value_type, class return_value_type, class sort_type>
class sort_callback_wrapper : public cb_sorting_object <value_type>
{
 private:
  typedef return_value_type (*conversion_in_callback_type)(const value_type &);
  conversion_in_callback_type conversion_in_callback;
  sort_type sort_object;
 public:
  sort_callback_wrapper (conversion_in_callback_type howto_convert_in) {conversion_in_callback = howto_convert_in;}
  bool operator () (const value_type & lhs, const value_type & rhs) {return sort_object(conversion_in_callback(lhs), conversion_in_callback(rhs));}
};

template <class value_type, class conversion_in_type, class sort_type>
class sort_functor_wrapper : public cb_sorting_object <value_type>
{
 private:
  conversion_in_type conversion_in_functor;
  sort_type sort_object;
 public:
  bool operator () (const value_type & lhs, const value_type & rhs) {return sort_object(conversion_in_functor(lhs), conversion_in_functor(rhs));}
};

//---------------------------------------------------------------------------

template <class value_type, class return_value_type, class aggregate_type>
class aggregate_callback_wrapper : public cb_aggregate_object <value_type>
{
 private:
  typedef return_value_type (*conversion_in_callback_type)(const value_type &);
  typedef value_type (*conversion_out_callback_type)(const return_value_type &);
  conversion_in_callback_type conversion_in_callback;
  conversion_out_callback_type conversion_out_callback;
  aggregate_type aggregate_object;
 public:
  aggregate_callback_wrapper (conversion_in_callback_type howto_convert_in, conversion_out_callback_type howto_convert_out) {conversion_in_callback = howto_convert_in; conversion_out_callback = howto_convert_out;}
  value_type get_value () {return conversion_out_callback(aggregate_object.get_value());}
  void set_value (const value_type & new_base_value) {aggregate_object.set_value(conversion_in_callback(new_base_value));}
  void add_value (const value_type & next_row_value) {aggregate_object.add_value(conversion_in_callback(next_row_value));}
};

template <class value_type, class conversion_in_type, class conversion_out_type, class aggregate_type>
class aggregate_functor_wrapper : public cb_aggregate_object <value_type>
{
 private:
  conversion_in_type conversion_in_functor;
  conversion_out_type conversion_out_functor;
  aggregate_type aggregate_object;
 public:
  value_type get_value () {return conversion_out_functor(aggregate_object.get_value());}
  void set_value (const value_type & new_base_value) {aggregate_object.set_value(conversion_in_functor(new_base_value));}
  void add_value (const value_type & next_row_value) {aggregate_object.add_value(conversion_in_functor(next_row_value));}
};

//---------------------------------------------------------------------------

template <class value_type>
class sort_less_than : public cb_sorting_object <value_type>
{
 public:
  bool operator () (const value_type & lhs, const value_type & rhs) {return lhs < rhs;}
};

template <class value_type>
class sort_greater_than : public cb_sorting_object <value_type>
{
 public:
  bool operator () (const value_type & lhs, const value_type & rhs) {return lhs > rhs;}
};

//---------------------------------------------------------------------------

template <class value_type>
class aggregate_minimum : public cb_aggregate_object <value_type>
{
 private:
  value_type minimum_found;
 public:
  value_type get_value () {return minimum_found;}
  void set_value (const value_type & new_base_value) {minimum_found = new_base_value;}
  void add_value (const value_type & next_row_value) {if (next_row_value < minimum_found) minimum_found = next_row_value;}
};

template <class value_type>
class aggregate_maximum : public cb_aggregate_object <value_type>
{
 private:
  value_type maximum_found;
 public:
  value_type get_value () {return maximum_found;}
  void set_value (const value_type & new_base_value) {maximum_found = new_base_value;}
  void add_value (const value_type & next_row_value) {if (next_row_value > maximum_found) maximum_found = next_row_value;}
};

template <class value_type>
class aggregate_first : public cb_aggregate_object <value_type>
{
 private:
  value_type first_found;
 public:
  value_type get_value () {return first_found;}
  void set_value (const value_type & new_base_value) {first_found = new_base_value;}
  void add_value (const value_type & next_row_value) {}
};

template <class value_type>
class aggregate_last : public cb_aggregate_object <value_type>
{
 private:
  value_type last_found;
 public:
  value_type get_value () {return last_found;}
  void set_value (const value_type & new_base_value) {last_found = new_base_value;}
  void add_value (const value_type & next_row_value) {last_found = next_row_value;}
};

template <class value_type>
class aggregate_n_th : public cb_aggregate_object <value_type>
{
 private:
  unsigned int n_seek;
  unsigned int n_curr;
  value_type nth_found;
 public:
  aggregate_n_th (unsigned int position_number) {n_seek = position_number; n_curr = 0;}
  value_type get_value () {return nth_found;}
  void set_value (const value_type & new_base_value) {n_curr = 1; if (n_curr == n_seek) nth_found = new_base_value;}
  void add_value (const value_type & next_row_value) {n_curr++; if (n_curr == n_seek) nth_found = next_row_value;}
};

template <class value_type>
class aggregate_sum : public cb_aggregate_object <value_type>
{
 private:
  value_type sum_found;
 public:
  value_type get_value () {return sum_found;}
  void set_value (const value_type & new_base_value) {sum_found = new_base_value;}
  void add_value (const value_type & next_row_value) {sum_found = sum_found + next_row_value;}
};

template <class value_type>
class aggregate_avg : public cb_aggregate_object <value_type>
{
 private:
  value_type sum_found;
  unsigned int n_curr;
 public:
  value_type get_value () {return (sum_found / n_curr);}
  void set_value (const value_type & new_base_value) {n_curr = 1; sum_found = new_base_value;}
  void add_value (const value_type & next_row_value) {n_curr++; sum_found = sum_found + next_row_value;}
};

//---------------------------------------------------------------------------

#endif
