#include "prometheus_util.h"
#include "lookuptablewrapper.h"
#include "formheaders.h"

// jhaley 20111010: exception safety macros for PSay / PPretendToSay
// These routines are called in a lot of catch blocks and in other error situations so we need them to be as exception
// safe as possible, just like LogIt.
#define NOTHROW_BEGIN  try {
#define NOTHROW_END    } catch(...) {}

int PSay(const string &p_message, prometheus_message_types p_message_type, int p_buttons, const string &p_class_name, const string &p_other_info, HWND hwnd)
{
  int response = 0;

  // jhaley 20111010: no exceptions should be thrown out of this routine, as they can disrupt the shutdown process,
  // particularly when the database connection has been lost.
  NOTHROW_BEGIN

  // possible types for p_message_type:
  // m_program_error , m_warning , m_invalid_data, m_missing_data, m_database_error, m_reminder,
  // m_message, m_confirmation, m_duplicate_data, m_debug
  // if you need a new type, add it to the enum in allsymbols.h and also to the
  // message_box_types table in the database (you -must- add it to the database
  // for this to work properly)
  string caption, response_name;
  string sql;
  Pstamp begin_stamp = CurrentPstamp();
  Pstamp end_stamp;
  bool broadcast_as_error = false;

  int additional_icon = 0;
  switch (p_message_type)
  {
  // MB_ICONEXCLAMATION : An exclamation-point icon appears in the message box.
  // MB_ICONINFORMATION : An icon consisting of a lowercase letter i in a circle appears in the message box.
  // MB_ICONQUESTION    : A question-mark icon appears in the message box.
  // MB_ICONSTOP        : A stop-sign icon appears in the message box.

  case m_program_error:  additional_icon = MB_ICONSTOP;        broadcast_as_error = true; break;
  case m_warning:        additional_icon = MB_ICONQUESTION;    break;
  case m_invalid_data:   additional_icon = MB_ICONEXCLAMATION; break;
  case m_missing_data:   additional_icon = MB_ICONEXCLAMATION; break;
  case m_database_error: additional_icon = MB_ICONSTOP;        broadcast_as_error = true; break;
  case m_reminder:       additional_icon = MB_ICONINFORMATION; break;
  case m_message:        additional_icon = MB_ICONINFORMATION; break;
  case m_confirmation:   additional_icon = MB_ICONQUESTION;    break;
  case m_duplicate_data: additional_icon = MB_ICONEXCLAMATION; break;
  case m_debug:          break;
  }
  if(dbDatabase && dbDatabase->TestConnected() && lookupTables)
    caption = lookupTables->ValueOf("message_box_types", IntToString(p_message_type));
  else
    caption = "";
  if (hwnd)
   response = ::MessageBox(hwnd, p_message.c_str(), caption.c_str(), p_buttons | additional_icon | MB_TASKMODAL);
  else
   response = Application->MessageBox(p_message.c_str(), caption.c_str(), p_buttons | additional_icon);

  string other_info = p_other_info;
  if (p_message_type == m_program_error || p_message_type == m_database_error)
  {
    if (other_info != "")
      other_info += "\n\n";
    other_info += "Last SQL error:  " + last_sql_lib_error.message;
  }

  // not using execute insert statment here because it's quoting current_timestamp
  // and that causes it to think it's a string and die
  end_stamp = CurrentPstamp();
  if(dbDatabase && dbDatabase->TestConnected())
  {
    sql = "insert into message_box_log (user_id, event_stamp, form_name, message_box_id, message_text, message_box_response_id, "
          "other_info, response_time) values "
          "(" + current_userid + ", current_timestamp, '" + p_class_name + "', " + IntToString(p_message_type) +
          ", '" + StripTrailing(ReplaceStringWithString(p_message, "'", "''").substr(0, 2000), "'") + "', " + IntToString(response) + ", ";
    if((other_info != "") /* && (other_info.length() <= 250) */ )
      // DONE: better to record part of the info than none at all, right?
      // DONE: whoops, the substring should be the last thing done, as replacing ' with '' could push it right back over the limit again
      // DONE: whoops, if the substring ends in a ', that ' and the one at the end of the field would be interpreted as an escaped ', so
      //       now we strip trailing whatnots too  :-)
      sql += "'" + StripTrailing(ReplaceStringWithString(other_info, "'", "''").substr(0, 8192), "'") + "', ";
    else
     sql += "null, ";
    sql += IntToString(end_stamp - begin_stamp) + ")";
    ExecuteStatement(dbDatabase, sql);

    // Paul fixing his stupidity:  if not connected to DB, or not downloaded lookuptables yet, and debugging is on, the below would generate an AV.  Fixed.
    //  -jhaley 20111010: moved into same checks as for above with respect to database validity
    // TODO -c FATAL : jhaley 20111010: potential infinite recursion with classLookUpTable::RefreshDatasets
    if(lookupTables)
      response_name = lookupTables->ValueOf("message_box_response_types", IntToString(response));
    else
      response_name = "";
  }

  if(debug_level[libraries]>=3)
    LogIt(("psay." + ReplaceStringWithString(p_class_name, " ", "") + ".log").c_str(), "[" + caption + "] [" + response_name + "," + IntToString(end_stamp - begin_stamp) + "] " + ReplaceStringWithString(p_message, "\n", "  ") + "  {" + ReplaceStringWithString(other_info, "\n", "  ") + "}\n");
  if(debug_level[libraries]>=3)
    LogIt(("psay." + ReplaceStringWithString(caption, " ", "") + ".log").c_str(), "[" + p_class_name + "] [" + response_name + "," + IntToString(end_stamp - begin_stamp) + "] " + ReplaceStringWithString(p_message, "\n", "  ") + "  {" + ReplaceStringWithString(other_info, "\n", "  ") + "}\n");

  // let's re-broadcast errors to the errors channel, yeah?
  if((tcpip_message_router) && broadcast_as_error)
    message_router->SendMessageToAll("error", current_userid + "~" + p_class_name + "~" + IntToString(p_message_type) + "\n" +
      ReplaceStringWithString(p_message, "\n", "<br>") + "~" + IntToString(response) + "~" +
      ReplaceStringWithString(other_info, "\n", "<br>"));

  NOTHROW_END

  return response;
}
//--------------------------------------------------------------------------
void PPretendToSay(const string &p_message, prometheus_message_types p_message_type, const string &p_class_name, const string &p_other_info)
{
  // jhaley 20111010: no exceptions should be thrown out of this routine, as they can disrupt the shutdown process,
  // particularly when the database connection has been lost.
  NOTHROW_BEGIN

  string sql;
  // not using execute insert statment here becuase it's quoting current_timestamp
  // and that causes it to think it's a string and die
  string caption = "";
  if(dbDatabase && dbDatabase->TestConnected())
  {
    sql = "insert into message_box_log (user_id, event_stamp, form_name, message_box_id, message_text, other_info) values "
          "(" + current_userid + ", current_timestamp, '" + p_class_name + "', " + IntToString(p_message_type) +
          ", '" + ReplaceStringWithString(p_message.substr(0, 2000), "'", "''") + "', ";
    if((p_other_info != "") /* && (p_other_info.length() <= 250) */ )
      // DONE: better to record part of the info than none at all, right?
      // DONE: whoops, the substring should be the last thing done, as replacing ' with '' could push it right back over the limit again
      sql += "'" + ReplaceStringWithString(p_other_info, "'", "''").substr(0, 8192) + "')";
    else
     sql += "null)";
    ExecuteStatement(dbDatabase, sql);
    caption = GetOneField(dbDatabase, "select item from message_box_types where id = '" + IntToString(p_message_type) + "'");
  }

  //lookupTables->ValueOf("message_box_types", IntToString(p_message_type)); // lookup tables might not be downloaded yet; above is better for debugging purposes.
  if(debug_level[libraries]>=3)
    LogIt(("pptsay." + ReplaceStringWithString(p_class_name, " ", "") + ".log").c_str(), "[" + caption + "] " + ReplaceStringWithString(p_message, "\n", "  ") + "  {" + ReplaceStringWithString(p_other_info, "\n", "  ") + "}");
  if(debug_level[libraries]>=3)
    LogIt(("pptsay." + ReplaceStringWithString(caption, " ", "") + ".log").c_str(), "[" + p_class_name + "] " + ReplaceStringWithString(p_message, "\n", "  ") + "  {" + ReplaceStringWithString(p_other_info, "\n", "  ") + "}");

  // assume all PPretendToSay messages are serious enough to broadcast as errors...
  if((tcpip_message_router) /*&& broadcast_as_error*/ )
    message_router->SendMessageToAll("error", current_userid + "~" + p_class_name + "~" + IntToString(p_message_type) + "\n" +
      ReplaceStringWithString(p_message, "\n", "<br>") + "~" + ReplaceStringWithString(p_other_info, "\n", "<br>"));

  NOTHROW_END
}
//---------------------------------------------------------------------------
// jhaley 20100928: "clever current id" utility to replace inlined code being used *everywhere*.
string CleverCurrentId(const string &field, const string &compare_value, const string &current_id_if_equals, const string &current_id_if_notequals)
{
  string to_return = "case when " + field + " = '" + compare_value + "' then '" + lookupTables->IdOf("current_types", current_id_if_equals) +
                     "' else '" + lookupTables->IdOf("current_types", current_id_if_notequals) + "' end";
  return to_return;
}
//---------------------------------------------------------------------------
void DefaultAction(string id, string table_name)
{
  if(debug_level[other_forms]>=4 && message_router) message_router->SendMessageToAll("debug", "DefaultAction(id,table_name) called");
  map<string, string> field_map;
  field_map["id"] = id;

  DefaultAction(table_name, field_map);
}
//---------------------------------------------------------------------------
void DefaultAction(string table_name, map<string, string> &field_map)
{
  if(debug_level[other_forms]>=4 && message_router) message_router->SendMessageToAll("debug", "DefaultAction entry");
  if (field_map["id"] == "")
    field_map["id"] = "new";
  string id = field_map["id"];

  if(table_name == "people")
  {
    FORM_TRY
      TformRecordPerson * view_person = new TformRecordPerson(NULL, field_map);
      view_person->Show();
    FORM_CATCH
  }
  else if(table_name == "participations")
  {
    FORM_TRY
      TformRecordParticipationsDiab * view_participation = new TformRecordParticipationsDiab(NULL, field_map);
      view_participation->Show();
    FORM_CATCH
  }
  else if(table_name == "events")
  {
    // jhaley: this doesn't work when adding; "new" is not a valid event.id. The caller of this function MUST set
    // an appropriate group_id in the field_map if this is a new record. Existing events do not need a group_id
    // set externally because it is always already valid in the record.
    //field_map["group_id"] = GetOneField(dbDatabase, "select group_id from events where id = " + id);
    FORM_TRY
      TformRecordEvent * eventScreen = new TformRecordEvent(NULL, field_map);
      eventScreen->Show();
    FORM_CATCH
  }
  else if(table_name == "tasks")
  {
    map<string, string> task_map;
    if (id != "" && IsInt(id))
      SqlToMap(dbDatabase, string("select group_id, services_provided_id, com_dept_id from tasks where id = '" + id + "'"), task_map);
    MergeMap (task_map, field_map); // copy+overwrite values in task_map from field_map
    int duration = StringToInt(task_map["duration"]);
    bool new_patient = (task_map["new_client"] == "1");

    if(task_map["group_id"] == lookupTables->IdOf("groups", "Diabetes"))
    {
      FORM_TRY
        TformRecordTaskDiab * diabTaskScreen = new TformRecordTaskDiab(NULL, task_map);
        if (duration) diabTaskScreen->editTimeSpent->Text = FloatToString(duration/60, 2).c_str();
        // if (new_patient) diabTaskScreen->// er, there is none?  Does Diabetes really not track new patients?
        diabTaskScreen->Show();
      FORM_CATCH
    }
    else if(task_map["group_id"] == lookupTables->IdOf("groups", "Community Health") ||
            task_map["group_id"] == lookupTables->IdOf("groups", "Public Health"))
    {
      // jhaley 20100813: Rather than arbitrarily restrict Comm. and Pub. Health to attachment of one of their two types of tasks
      // to events, I have written this code that will either select the proper form based on either the contextual info that was
      // passed in, or will allow the user to choose the type of task when either could be valid. A similar approach is used on
      // TformRecordComVisit itself if it is invoked through the ::Do static method, but that code doesn't choose a form; it only
      // chooses a com_dept_id other than the zero record.
      enum open_form_type { open_task, open_visit } which_action_to_perform = open_task;
      string com_dept_id = "0";
      bool should_exit = false;
      
      if(task_map["com_dept_id"] == "") // com_dept_id was not specified, so give a choice.
      {
        PSay("You can create either a Phone Task or a Visit for your department. Please select the appropriate type of Task or "
             "Visit from the next window. If you are not sure, you can select 'Close' in the next window, or, after making a "
             "choice, close the Task or Visit form without saving the new record.",
             m_message, MB_OK, "prometheus_utils::DefaultAction");
        TIBTransaction *dbTransaction = new TIBTransaction(NULL);
        try
        {
          string answer_str = "";
          string sql;
          string clever_current_id = CleverCurrentId("valid", "1", "Current", "Expired");

          StdTransaction(dbTransaction, dbDatabase);
          
          FORM_TRY
            TformSpecialSelectListview *selectDeptForm = new TformSpecialSelectListview(NULL, &answer_str);
            selectDeptForm->Filter->AddColumnSpec("Visit Type", "", as_text, true, 100);
            selectDeptForm->Filter->AddColumnSpec("Current Id", "current_id", as_int, false, 40);
            selectDeptForm->Filter->AddColumnSpec("Internal Id", "internal_id", as_int, false, 40);
            selectDeptForm->Filter->SetIconStatus(true);
            selectDeptForm->SourceTableName = "com_dept_types";
            sql = "select case when item = '' then 'Phone Task' else (item || ' Visit') end, " + clever_current_id + ", "
                  "id from com_dept_types where valid = 1 ";
            // Community Health only seems to have access to "Adult"-type visits.
            if (task_map["group_id"] == lookupTables->IdOf("groups", "Community Health"))
              sql += "and (id = '" + lookupTables->IdOf("com_dept_types", "Adult") + "' or id = 0) ";
            sql += "order by item ";
            selectDeptForm->Execute(dbTransaction, sql);
            selectDeptForm->ShowModal();
          FORM_CATCH

          if(answer_str == "") // nothing was chosen?
            should_exit = true;
          else if(StringToInt(answer_str) == 0) // zero, create a task.
            which_action_to_perform = open_task;
          else                                  // non-zero, create a visit.
          {
            which_action_to_perform = open_visit;
            com_dept_id = answer_str;
          }
          dbTransaction->Commit();
        }
        catch(...) // database error
        {
          PSay("Could not get list of task/visit types", m_database_error, MB_OK, "prometheus_utils::DefaultAction");
          if(dbTransaction->Active)
            dbTransaction->Rollback();
          should_exit = true; // no clue what happened but we can't do anything now.
        }
        delete dbTransaction;
        
        if(should_exit) // either they want to forget about it, or we had an error.
          return;
      }
      else if(IsInt(task_map["com_dept_id"])) // we have an existing com_dept_id...
      {
        if(StringToInt(task_map["com_dept_id"]) == 0) // ...and it's zero, so this is a task.
          which_action_to_perform = open_task;
        else                                         // ...and it's non-zero, this is a visit.
        {
          which_action_to_perform = open_visit;
          com_dept_id = task_map["com_dept_id"];
        }
      } 
      switch(which_action_to_perform)
      {
      case open_task:
        FORM_TRY
          TformRecordTaskCom *comTaskScreen = new TformRecordTaskCom(NULL, task_map);
          if (duration) comTaskScreen->editTimeSpent->Text = FloatToString(duration, 2).c_str();
          comTaskScreen->Show();
        FORM_CATCH
        break;
      case open_visit:
        FORM_TRY
          task_map["com_dept_id"] = com_dept_id;
          TformRecordComVisit *comVisitScreen = new TformRecordComVisit(NULL, task_map);
          if (duration) 
            comVisitScreen->editTotalTime->Text = FloatToString(duration, 2).c_str();
          if (new_patient)
            comVisitScreen->chkNew->Enabled = false;
          comVisitScreen->Show();
        FORM_CATCH
        break;
      }
    }
    else if(task_map["group_id"] == lookupTables->IdOf("groups", "Caregiver Support Program"))
    {
      FORM_TRY
        TformRecordTaskCare * comTaskScreen = new TformRecordTaskCare(NULL, task_map);
        if (duration) comTaskScreen->editTimeSpent->Text = FloatToString(duration/60, 2).c_str();
        comTaskScreen->Show();
      FORM_CATCH
    }
    else if(task_map["group_id"] == lookupTables->IdOf("groups", "Little Axe Clinic"))
    {
      FORM_TRY
        TformRecordTaskLac * lacTaskScreen = new TformRecordTaskLac(NULL, task_map, lookupTables->IdOf("groups", "Little Axe Clinic"));
        if (duration) lacTaskScreen->editTimeSpent->Text = FloatToString(duration, 2).c_str();
        if (new_patient)
          lacTaskScreen->checkNewPatient->Enabled = false;
        lacTaskScreen->Show();
      FORM_CATCH
    }
    else if(task_map["group_id"] == lookupTables->IdOf("groups", "Shawnee Clinic"))
    {
      FORM_TRY
        TformRecordTaskLac * lacTaskScreen = new TformRecordTaskLac(NULL, task_map, lookupTables->IdOf("groups", "Shawnee Clinic"));
        if (duration) lacTaskScreen->editTimeSpent->Text = FloatToString(duration, 2).c_str();
        if (new_patient)
          lacTaskScreen->checkNewPatient->Enabled = false;
        lacTaskScreen->Show();
      FORM_CATCH
    }
    else if(task_map["group_id"] == lookupTables->IdOf("groups", "Behavioral Health"))
    {
      FORM_TRY
        TformRecordTaskBeh * behTaskScreen = new TformRecordTaskBeh(NULL, task_map);
        if (duration) behTaskScreen->editTimeSpent->Text = FloatToString(duration, 2).c_str();
        if (new_patient)
          behTaskScreen->checkNew->Enabled = false;
        behTaskScreen->Show();
      FORM_CATCH
    }
    else if(task_map["group_id"] == lookupTables->IdOf("groups", "Dental Clinic"))
    {
      FORM_TRY
        TformRecordTaskDen * denTaskScreen = new TformRecordTaskDen(NULL, task_map);
        if (duration) denTaskScreen->editTimeSpent->Text = FloatToString(duration/60, 2).c_str();
        if (new_patient)
          denTaskScreen->checkNewPatient->Enabled = false;
        denTaskScreen->Show();
      FORM_CATCH
    }
    else if(task_map["group_id"] == lookupTables->IdOf("groups", "Transportation"))
    {
      FORM_TRY
        TformRecordTaskTrans * transTaskScreen = new TformRecordTaskTrans(NULL, task_map);
        if (duration) transTaskScreen->editTimeSpent->Text = FloatToString(duration/60, 2).c_str();
        transTaskScreen->Show();
      FORM_CATCH
    }
    else if(task_map["group_id"] == lookupTables->IdOf("groups", "Contract Health"))
    {
      if((task_map["services_provided_id"] == "0") || (task_map["services_provided_id"] == ""))
      {
        FORM_TRY
          TformRecordTaskCobVisit * task_form = new TformRecordTaskCobVisit(NULL, task_map);
          task_form->Show();
        FORM_CATCH
      }
      else
      {
        FORM_TRY
          TformRecordTaskCob * task_form = new TformRecordTaskCob(NULL, task_map);
          task_form->Show();
        FORM_CATCH
      }
    }
  }
  else if(table_name == "tax_entities")
  {
    FORM_TRY
      TformRecordTaxEntity * view_facility = new TformRecordTaxEntity(NULL, field_map);
      view_facility->Show();
    FORM_CATCH
  }
  else if(table_name == "fitness_memberships")
  {
    FORM_TRY
      TformRecordFitnessMembership * membershipScreen = new TformRecordFitnessMembership(NULL, field_map);
      membershipScreen->Show();
    FORM_CATCH
  }
  else if(table_name == "accounts")
  {
    //open account form
    FORM_TRY
      TformRecordAccount * new_account = new TformRecordAccount (NULL, field_map);
      new_account->Show();
    FORM_CATCH
  }
  else if(table_name == "purchase_reqs")
  {
    FORM_TRY
      TformRecordPurchaseReq * pr_record = new TformRecordPurchaseReq(NULL, field_map);
      pr_record->Show();
    FORM_CATCH
  }
  else if(table_name == "bills")
  {
    string bill_id = GetOneField(dbDatabase, "select bill_id from bills where id = " + field_map["id"]);
    if((bill_id == lookupTables->IdOf("bill_types", "incoming provider bill")) ||
       (bill_id == lookupTables->IdOf("bill_types", "incoming vendor bill")))
    {
      FORM_TRY
        TformRecordIncomingBill * in_bill = new TformRecordIncomingBill(NULL, field_map);
        in_bill->Show();
      FORM_CATCH
    }
    else if(bill_id == lookupTables->IdOf("bill_types", "outgoing provider bill"))
    {
      FORM_TRY
        TformRecordOutgoingBill * out_bill = new TformRecordOutgoingBill(NULL, field_map);
        out_bill->Show();
      FORM_CATCH
    }
  }
  else if(table_name == "entity_contracts")
  {
    TformRecordEntityContract::Do(field_map);
  }
  else if(table_name == "insurance_policies")
  {
    FORM_TRY
      TformRecordInsurancePolicy * policy_record = new TformRecordInsurancePolicy(NULL, field_map);
      policy_record->Show();
    FORM_CATCH
  }
  else if (table_name == "insurance_coverage")
  {
    FORM_TRY
      TformRecordInsuranceCoverage * coverage_record = new TformRecordInsuranceCoverage(NULL, field_map);
      coverage_record->Show();
    FORM_CATCH
  }
  else if(table_name == "checks")
  {
    FORM_TRY
      TformRecordCheck * check_record = new TformRecordCheck(NULL, field_map);
      check_record->Show();
    FORM_CATCH
  }
  else if(table_name == "services_provided")
  {
    FORM_TRY
      TformRecordServiceProvided * service_record = new TformRecordServiceProvided(NULL, field_map);
      service_record->Show();
    FORM_CATCH
  }
  else if(table_name == "referrals")
  {
    FORM_TRY
      TformRecordReferral * referral_record = new TformRecordReferral(NULL, field_map);
      referral_record->Show();
    FORM_CATCH
  }
  else if(table_name == "bugs")
  {
    FORM_TRY
      TformRecordBug * bug_record = new TformRecordBug(NULL, field_map);
      bug_record->Show();
    FORM_CATCH
  }
  else if(table_name == "diab_education")
  {
    FORM_TRY
      TformRecordDiabScore * education_record = new TformRecordDiabScore(NULL, field_map);
      education_record->Show();
    FORM_CATCH
  }
  else if(table_name == "surveys")
  {
    FORM_TRY
      TformRecordSurvey * survey_record = new TformRecordSurvey(NULL, field_map);
      survey_record->Show();
    FORM_CATCH
  }
  else if(table_name == "activity_logsheets")
  {
    FORM_TRY
      TformRecordLogsheet * logsheet_record = new TformRecordLogsheet(NULL, field_map);
      logsheet_record->Show();
    FORM_CATCH
  }
  else if(table_name == "procedures")
  {
    FORM_TRY
      TformRecordProcedure * procedure_record = new TformRecordProcedure(NULL, field_map);
      procedure_record->Show();
    FORM_CATCH
  }
  else if(table_name == "conditions")
  {
    FORM_TRY
      TformRecordCondition * condition_record = new TformRecordCondition(NULL, field_map);
      condition_record->Show();
    FORM_CATCH
  }
  else if(table_name == "obs_response_sets")
  {
   TformRecordObservationResponseSet::Do(field_map);
  }
  else if(table_name == "obs_questions")
  {
   TformRecordObservationQuestion::Do(field_map);
  }
  else if(table_name == "denials")
  {
    FORM_TRY
      TformRecordDenial *denial_record = new TformRecordDenial(NULL, field_map);
      denial_record->Show();
    FORM_CATCH
  }
  else if(table_name == "denial_letters")
  {
    TformRecordDenialLetter::Do(field_map);
  }
  else if(table_name == "com_dept_types")
  {
    // jhaley: This is here for the benefit of choosing com_dept_types from a SpecialSelectListview form without the
    // user getting the message below if they happen to double-click something by accident.
    PSay("To select an item, click the 'Select' button below.", m_message, MB_OK, "prometheus_utils::DefaultAction");
  }
  else if(table_name == "transactions")
  {
    // jhaley 20100907: Added for TformSpecialTransactionDetails
    FORM_TRY
      TformRecordTransaction *transaction_record = new TformRecordTransaction(NULL, field_map);
      transaction_record->Show();
    FORM_CATCH
  }
  else
  {
    PSay("No action has been defined for this type of link yet; please talk to a programmer about this.",
         m_program_error, MB_OK, "prometheus_utils::DefaultAction");
  }
  if(debug_level[other_forms]>=4 && message_router) message_router->SendMessageToAll("debug", "DefaultAction exit");
}
//---------------------------------------------------------------------------
string PMakeTempFilePath(const string &filename)
{
  return PGlobals->current_tempdir + "\\" + filename;
}
//---------------------------------------------------------------------------
// PIsRunningUnderPath
// jhaley 20120110: Returns true if the Prometheus executable was launched under a path matching the indicated path
// as a substring component.
bool PIsRunningUnderPath(const string &filepath)
{
  TCHAR moduleName[MAX_PATH+2];
  string module_name_str;

  ZeroMemory(moduleName, sizeof(moduleName));

  GetModuleFileName(NULL, moduleName, MAX_PATH);
  module_name_str = moduleName;

  return (module_name_str.find(filepath) != string::npos);
}
//---------------------------------------------------------------------------
// jhaley 20110714: Many places need to create CHS mailing tasks. This is a shared routine to do it.
bool InsertCHSMailTask(TIBTransaction *task_transaction, string person_id, string message)
{
  bool successful = false;
  bool can_commit = false;

  if(!task_transaction)
  {
    task_transaction = new TIBTransaction(NULL);
    can_commit = true;
  }

  // jhaley 20110707: Create a Contract Health mailing task
  try
  {
    pmapstrs     insert_map;
    pmapstrtoint leave_alone;

    insert_map["current_id"]        = lookupTables->IdOf("current_types", "Current");
    insert_map["last_updated_by"]   = current_userid;
    insert_map["last_update_stamp"] = RightNow();
    insert_map["date_of_task"]      = CurrentDate();
    insert_map["group_id"]          = lookupTables->IdOf("groups", "Contract Health");
    insert_map["user_id"]           = current_userid;
    insert_map["method_id"]         = lookupTables->IdOf("method_types", "Mail");
    insert_map["direction_id"]      = lookupTables->IdOf("direction_types", "Outgoing");
    insert_map["task_comments"]     = message;
    insert_map["person_id"]         = person_id;

    leave_alone["task_comments"] = sql_no_uppercasing;

    // Explicitly set these to make absolutely sure they can't be considered COBs or COB Visits
    insert_map["services_provided_id"] = "0"; // excludes from cob_tasks
    insert_map["parent_task_id"]       = "0"; // excludes from visit_tasks

    if(can_commit)
      StdTransaction(task_transaction, dbDatabase);

    insert_map["id"] = GetNextId(task_transaction, "tasks");
    if(!ExecuteInsertStatement(task_transaction, "tasks", insert_map, leave_alone))
    {
      successful = false;
      if(task_transaction->Active && can_commit)
        task_transaction->Rollback();
    }
    else
    {
      successful = true;
      if(can_commit)
        task_transaction->Commit();
      atlas->Notify("tasks.person_id", insert_map["person_id"]);
    }
  }
  catch(...)
  {
    successful = false;
    if(task_transaction->Active && can_commit)
      task_transaction->Rollback();
  }

  if(can_commit)
    delete task_transaction;

  return successful;
}
//---------------------------------------------------------------------------
// jhaley 20110812: Crash log uploader - invoked at login, fairly early in the process to avoid any problems with
// logging the current session's crashes.
void UploadCrashLog()
{
  pvecmap crashmap;
  string  errmsg;

  bool success = CSVtoVecMap("CRASHLOG.CSV", '\"', ',', crashmap, errmsg, true);

  if(success && crashmap.size() > 0)
  {
    // Successfully parsed a non-empty crash log file. Insert each crash that does not already exist.
    for(pvecmap::iterator i_crash = crashmap.begin(); i_crash != crashmap.end(); ++i_crash)
    {
      // Crashes are identified as a combination of the SegCs:Eip and exception code values.
      // While it's true that the same crash might occur on different stacks, it's too expensive to compare stack dumps.
      string find_it_sql =
        "select id from crash_logs where "
        "ctx_segcs = '" + (*i_crash)["ctx_segcs"] + "' and "
        "ctx_eip   = '" + (*i_crash)["ctx_eip"]   + "' and "
        "code      = '" + (*i_crash)["code"]      + "' ";
      pvecmap existing_crashes;
      SqlToVecMap(dbDatabase, find_it_sql, existing_crashes);

      if(!existing_crashes.empty()) // Looks like this crash was logged already. Create crash_occurrences records.
      {
        for(pvecmap::iterator i_ec = existing_crashes.begin(); i_ec != existing_crashes.end(); ++i_ec)
        {
          TIBTransaction *dbTransaction = NULL;
          try
          {
            pmapstrs crash_occ;
            pmapstrtoint leave_alone;
            dbTransaction = new TIBTransaction(NULL);
            StdTransaction(dbTransaction, dbDatabase);

            crash_occ["id"]              = GetNextId(dbTransaction, "crash_occurrences");
            crash_occ["crash_log_id"]    = (*i_ec)["id"];
            crash_occ["creation_stamp"]  = RightNow();
            crash_occ["created_by"]      = current_userid;
            crash_occ["error_timestamp"] = (*i_crash)["error_timestamp"];

            if(!ExecuteInsertStatement(dbTransaction, "crash_occurrences", crash_occ, leave_alone))
            {
              if(dbTransaction->Active)
                dbTransaction->Rollback();
            }
            else
              dbTransaction->Commit();
          }
          catch(...)
          {
            if(dbTransaction && dbTransaction->Active) // Woops?!
              dbTransaction->Rollback();
          }
          delete dbTransaction;
        }
      }
      else // New crash.
      {
        TIBTransaction *dbTransaction = NULL;
        try
        {
          pmapstrtoint leave_alone;

          dbTransaction = new TIBTransaction(NULL);
          StdTransaction(dbTransaction, dbDatabase);

          (*i_crash)["id"]                = GetNextId(dbTransaction, "crash_logs");
          (*i_crash)["current_id"]        = lookupTables->IdOf("current_types", "Current");
          (*i_crash)["creation_stamp"]    = RightNow();
          (*i_crash)["created_by"]        = current_userid;
          (*i_crash)["last_update_stamp"] = (*i_crash)["creation_stamp"];
          (*i_crash)["last_updated_by"]   = current_userid;
          (*i_crash)["bug_status_id"]     = lookupTables->IdOf("bug_status_types",   "Not Begun");
          (*i_crash)["bug_priority_id"]   = lookupTables->IdOf("bug_priority_types", "3 - Normal");
          (*i_crash)["assigned_to"]       = "0";

          leave_alone["program_name"]  = sql_no_uppercasing;
          leave_alone["p_module_name"] = sql_no_uppercasing;
          leave_alone["run_by"]        = sql_no_uppercasing;
          leave_alone["cs_eip"]        = sql_no_uppercasing;
          leave_alone["stack"]         = sql_no_uppercasing;

          if(!ExecuteInsertStatement(dbTransaction, "crash_logs", *i_crash, leave_alone))
          {
            if(dbTransaction->Active)
              dbTransaction->Rollback();
          }
          else
            dbTransaction->Commit();
        }
        catch(...)
        {
          if(dbTransaction && dbTransaction->Active) // Woops?!
            dbTransaction->Rollback();
        }
        delete dbTransaction;
      }
    }
  }

  // Remove the CRASHLOG.CSV file.
  // We do this even if parsing it was not successful, because in that event, the crashlog is corrupted (ie. the
  // crash logger probably actually crashed itself half-way through a log due to a severely corrupt program state).
  remove("CRASHLOG.CSV");
}
//---------------------------------------------------------------------------
// PLoadTextFile
// jhaley 20120412: Load a text file. Returns NULL if fails.
char *PLoadTextFile(const char *filename)
{
  FILE *f;
  char *buffer;
  size_t size; // NB: BCB fread returns size_t instead of long >_>

  if(!(f = fopen(filename, "rb")))
    return NULL; // can't open file

  fseek(f, 0, SEEK_END);
  size = ftell(f);
  fseek(f, 0, SEEK_SET);

  if(size == 0)
  {
    fclose(f);
    return NULL; // nothing in the file
  }

  buffer = new char [size + 1]; // size+1 for null termination
  memset(buffer, 0, size + 1);

  if(fread(buffer, 1, size, f) != size)
  {
    delete [] buffer;
    fclose(f);
    return NULL; // can't read the file
  }

  fclose(f);

  return buffer; // success
}
//---------------------------------------------------------------------------
