#ifndef ranges_h
#define ranges_h

#include <map>
#include <set>
#include <vector>
#include <list>

using namespace std;

template <class T>
class range
{
 enum err_range_types {err_range_gap, err_range_empty};

 public:
  T range_begin, range_end;
  range (const T & p_begin, const T & p_end)
  {
   if (p_begin < p_end)
   {
    range_begin = p_begin;
    range_end = p_end;
   } else
   {
    range_begin = p_end;
    range_end = p_begin;
   }
  }

  range () {}

  bool operator < (const range<T> & rhs) const
  {
   if (range_begin < rhs.range_begin)
    return true;
   else
    return (range_end < rhs.range_end);
  }

  bool operator == (const range<T> & rhs) const
  {
   return (range_begin == rhs.range_begin && range_end == rhs.range_end);
  }

  bool Contain (const T & rhs) const
  {
   if (rhs > range_begin && rhs < range_end)
    return true;
   else
    return false;
  }

  bool ContainEq (const T & rhs) const
  {
   if (rhs >= range_begin && rhs <= range_end)
    return true;
   else
    return false;
  }

  bool Overlap (const range<T> & rhs) const
  {
   if (range_begin >= rhs.range_end || range_end <= rhs.range_begin)
    return false;
   else
    return true;
  }

  bool OverlapEq (const range<T> & rhs) const
  {
   if (range_begin > rhs.range_end || range_end < rhs.range_begin)
    return false;
   else
    return true;
  }

  bool Encompass (const range<T> & rhs) const
  {
   if (rhs.range_begin > range_begin && rhs.range_end < range_end)
    return true;
   else
    return false;
  }

  bool EncompassEq (const range<T> & rhs) const
  {
   if (rhs.range_begin >= range_begin && rhs.range_end <= range_end)
    return true;
   else
    return false;
  }

  bool Touch (const range<T> & rhs) const
  {
   if (range_begin == rhs.range_end || range_end == rhs.range_begin)
    return true;
   else
    return false;
  }

  range<T> & Intersection (const range<T> & rhs)
  {
   // must at least touch
   if (!OverlapEq(rhs))
    throw err_range_empty;

   // calculate minimum overlapping area
   T temp_begin, temp_end;
   temp_begin  = max (range_begin, rhs.range_begin);
   temp_end    = min (range_end, rhs.range_end);

   // assign
   range_begin = temp_begin;
   range_end   = temp_end;
   return (*this);
  }

  range<T> & Union (const range<T> & rhs)
  {
   // must at least touch
   if (!OverlapEq(rhs))
    throw err_range_gap;

   // calculate minimum overlapping area
   T temp_begin, temp_end;
   temp_begin  = min (range_begin, rhs.range_begin);
   temp_end    = max (range_end, rhs.range_end);

   // assign
   range_begin = temp_begin;
   range_end   = temp_end;
   return (*this);
  }

  range<T> & Difference (const range<T> & rhs)
  {
   // cannot carve a chunk out of the middle
   if (Encompass(rhs))
    throw err_range_gap;

   // cannot entirely delete range
   if (rhs.EncompassEq(this))
    throw err_range_empty;

   // to bother calculating, must at least share an edge
   if (OverlapEq(rhs))
   {
    if (rhs.range_begin < range_end && rhs.range_end >= range_end)
    {
     // if it's the ending edge, then we nibble off the end
     range_end = rhs.range_begin;
    } else
    {
     // otherwise, we nibble off the front
     range_begin = rhs.range_end;
    }
   }
   return (*this);
  }
};

template <class T>
void range_group_if_not_overlapping (const multiset<T> & p_input, vector<multiset<T> >& p_output)
{
 p_output.clear();

 bool column_conflict;
 bool column_needed;

 for (multiset<T>::const_iterator i_T = p_input.begin(); i_T != p_input.end(); i_T++)
 {
  column_needed = true;
  for (vector<multiset<T> >::iterator i_column = p_output.begin(); i_column != p_output.end(); i_column++)
  {
   column_conflict = false;
   for (multiset<T>::iterator i_item = i_column->begin(); i_item != i_column->end(); i_item++)
   {
    if (i_T->Overlap(*i_item))
     column_conflict = true;
   }
   if (!column_conflict)
   {
    i_column->insert(*i_T);
    column_needed = false;
    break;
   }
  }
  if (column_needed)
  {
   p_output.resize(p_output.size() + 1);
   p_output.back().insert(*i_T);
  }
 }
}

template <class T, class G>
void range_group_if_overlapping_or_touching (const multiset<T> & p_input, map<G, multiset<T> >& p_output)
{
 p_output.clear();

 multiset<T> temp_items;

 vector<map<G, multiset<T> >::iterator> temp_group_search;

 for (multiset<T>::const_iterator i_T = p_input.begin(); i_T != p_input.end(); i_T++)
 {
  // find all groups this may be in
  temp_group_search.clear();
  for (map<G, multiset<T> >::iterator i_G = p_output.begin(); i_G != p_output.end(); i_G++)
  {
   if (i_T->OverlapEq(i_G->first))
    temp_group_search.push_back(i_G);
  }

  temp_items.clear();

  switch (temp_group_search.size())
  {
   case 0:
   {
    // no group exists yet to include this
     temp_items.insert(*i_T);
     T temp_group = (*i_T);
     p_output.insert(pair<G, multiset<T> >(temp_group, temp_items));
    break;
   }
   case 1:
   {
    // a group exists
    if (temp_group_search.front()->first.Encompass(*i_T))
    {
     // the current group can accept this new item without expanding its borders
     temp_group_search.front()->second.insert(*i_T);
    } else
    {
     // the current item does overlap with just one range, but it sticks out the sides!
     temp_items.insert(temp_group_search.front()->second.begin(), temp_group_search.front()->second.end());
     temp_items.insert(*i_T);
     G temp_group = temp_group_search.front()->first;
     temp_group.Union(*i_T);
     p_output.erase(temp_group_search.front());
     p_output.insert(pair<G, multiset<T> >(temp_group, temp_items));
    }
    break;
   }
   default:
   {
    // multiple groups, need to be merged, and this item added
    G temp_group = (*i_T);
    temp_items.insert (*i_T);
    for (vector<map<G, multiset<T> >::iterator>::iterator i_result = temp_group_search.begin(); i_result != temp_group_search.end(); i_result++)
    {
     temp_items.insert((*i_result)->second.begin(), (*i_result)->second.end());
     temp_group.Union((*i_result)->first);
     p_output.erase(*i_result);
    }
    p_output.insert(pair<G, multiset<T> >(temp_group, temp_items));
    break;
   }
  }
 }
}

template <class T, class G>
void range_group_if_overlapping (const multiset<T> & p_input, map<G, multiset<T> >& p_output)
{
 p_output.clear();

 multiset<T> temp_items;

 vector<map<G, multiset<T> >::iterator> temp_group_search;

 for (multiset<T>::const_iterator i_T = p_input.begin(); i_T != p_input.end(); i_T++)
 {
  // find all groups this may be in
  temp_group_search.clear();
  for (map<G, multiset<T> >::iterator i_G = p_output.begin(); i_G != p_output.end(); i_G++)
  {
   if (i_T->Overlap(i_G->first))
    temp_group_search.push_back(i_G);
  }

  temp_items.clear();

  switch (temp_group_search.size())
  {
   case 0:
   {
    // no group exists yet to include this
     temp_items.insert(*i_T);
     T temp_group = (*i_T);
     p_output.insert(pair<G, multiset<T> >(temp_group, temp_items));
    break;
   }
   case 1:
   {
    // a group exists
    if (temp_group_search.front()->first.Encompass(*i_T))
    {
     // the current group can accept this new item without expanding its borders
     temp_group_search.front()->second.insert(*i_T);
    } else
    {
     // the current item does overlap with just one range, but it sticks out the sides!
     temp_items.insert(temp_group_search.front()->second.begin(), temp_group_search.front()->second.end());
     temp_items.insert(*i_T);
     G temp_group = temp_group_search.front()->first;
     temp_group.Union(*i_T);
     p_output.erase(temp_group_search.front());
     p_output.insert(pair<G, multiset<T> >(temp_group, temp_items));
    }
    break;
   }
   default:
   {
    // multiple groups, need to be merged, and this item added
    G temp_group = (*i_T);
    temp_items.insert (*i_T);
    for (vector<map<G, multiset<T> >::iterator>::iterator i_result = temp_group_search.begin(); i_result != temp_group_search.end(); i_result++)
    {
     temp_items.insert((*i_result)->second.begin(), (*i_result)->second.end());
     temp_group.Union((*i_result)->first);
     p_output.erase(*i_result);
    }
    p_output.insert(pair<G, multiset<T> >(temp_group, temp_items));
    break;
   }
  }
 }
}

template <class T, class G>
void range_group_block_layout (const multiset<T> & p_input, map<G, vector<multiset<T> > >& p_output)
{
 p_output.clear();
 vector<multiset<T> > temp_blocked;
 map<G, multiset<T> > temp_grouped;

 range_group_if_overlapping (p_input, temp_grouped);
 for (map<G, multiset<T> >::const_iterator i_group = temp_grouped.begin(); i_group != temp_grouped.end(); i_group++)
 {
  range_group_if_not_overlapping (i_group->second, temp_blocked);
  p_output.insert(pair<G, vector<multiset<T> > >(i_group->first,temp_blocked));
 }
}

#endif