//ATLAS
#include "relevant.h"

//classRequestHandler-------------------------------------------------------
//request handler is in charge of maintaining a list of currect requests for
//callbacks, and notifying such callback functions if anything relevant occurs.
//To accomplish this, a list of requests is kept which each know what key
//they are watching (normally tablename.fieldname, but can be otherwise) and
//values these fields should be related to in some way. It is possible to
//specify any relational operator, as well as rel_any and rel_like.
//Like will do a LikeString() comparison, and rel_any will do no comparison
//at all, and always return true.
//Requests can be cancelled, and in fact must be cancelled if there's a
//chance that the requesting object will die. All forms should do a
//CancelRequestFrom(this) upon exiting, so as to clear any requests to
//callback functions for that instance of the form class. This is -very-
//important. If this is not done, there will be Access Violations all over.
//You can also cancel a specific request, if you record the id number you
//are given upon making a request. That id number is used to cancel a specific
//request via CancelRequest(number).
//One notification will not cause multiple calls to the same callback function,
//if it can be avoided. This is not time-dependent, so multiple notifications
//all bunched together might still call the same callback multiple times.
//It is obvious some care should be given to choosing the most appropriate
//time and place for both requests and notifications. But at least only
//relevant changes will affect callbacks, and as little as possible at that.

//requesting notification will look like this:
//myint = bob->RequestNotificationWhen("tablename.fieldname", "value", rel_like, as_test, this, RefreshHandler(&(TMYCLASS::REFRESHFUNCTION)));
//where bob is the classRequestHandler
//      myint is used to cancel the specific request later
//      TMYCLASS is something like TformMain or TformPersonRecord
//      REFRESHFUNCTION is defined as something like
//        void TformPersonRecord::RefreshScreenThingie();
//      but here, you would simply say RefreshScreenThingie

classRequestHandler::classRequestHandler()
{
 last_inserted = 0;
 requests.clear();
 //message += "Wrapper initiated.\n";
}

classRequestHandler::~classRequestHandler()
{
 requests.clear();
}

int classRequestHandler::RequestNotificationWhen (string f, string i, int r, int t, TObject* w, RefreshHandler h)
{
 int inserting_at = last_inserted++;
 requests[inserting_at] = classRequest(f,i,r,t,w,h);
 //message += "Request submitted: ";
 //message += requests[inserting_at].Description();
 //message += ".\n";
 return inserting_at;
}

void classRequestHandler::Notify (string f, string i)
{
 set<int> to_notify;
 to_notify.clear();
 for (map<int, classRequest>::iterator itr = requests.begin(); itr != requests.end(); itr++)
 {
  if (itr->second.MatchesNotification(f,i))
  {
   bool already_sent = false;
   for (set<int>::iterator j = to_notify.begin(); j != to_notify.end(); j++)
   {
    if (itr->second.IsSameMethodAs(requests[*j]))
    {
     //message += "Duplicate discarded.\n";
     already_sent = true;
    }
   }
   if (!already_sent)
   {
    to_notify.insert(itr->first);
    string class_name = "relevant";
    try
    {
     AV_TRY
      itr->second.Fulfill(itr->first, i);
     AV_CATCH
    }
    //HandleMissingCallback(e);
    catch (...)
    {
     ; // no, really, we don't care.
    }
    //message += "Notification: ";
    //message += f + " = " + i + " ";
    //message += "matched request ";
    //message += itr->second.Description();
    //message += ".\n";
   }
  }
 }
}

void classRequestHandler::CancelRequestFrom (TObject * w)
{
 set<int> to_delete;
 to_delete.clear();
 for (map<int, classRequest>::iterator itr = requests.begin(); itr != requests.end(); itr++)
 {
  if (itr->second.IsFrom(w)) to_delete.insert(itr->first);
 }
 for (set<int>::iterator itr = to_delete.begin(); itr != to_delete.end(); itr++)
 {
  requests.erase(*itr);
 }
}

void classRequestHandler::CancelRequest (int i)
{
 requests.erase(i);
 //message += "A request was cancelled.\n";
}

void classRequestHandler::HandleMissingCallback (EAccessViolation & e)
{
 // not sure what to do about it
 return;
}
//classRequest--------------------------------------------------------------

classRequest::classRequest ()
{
}

bool classRequest::IsFrom (TObject *w)
{
 if (where ==  w)
  return true;
 else
  return false;
}

classRequest::classRequest (const classRequest & other_request)
{
 fieldname = other_request.fieldname;
 itemvalue = other_request.itemvalue;
 relsymbol = other_request.relsymbol;
 fieldtype = other_request.fieldtype;
 where     = other_request.where;
 how       = other_request.how;
}

string classRequest::Description ()
{
 string answer = "";
 answer += fieldname;
 switch (relsymbol)
 {
  case rel_e:  answer += " == "; break;
  case rel_l:  answer += " < ";  break;
  case rel_g:  answer += " > ";  break;
  case rel_ge: answer += " >= "; break;
  case rel_le: answer += " <= "; break;
  case rel_ne: answer += " != "; break;
  case rel_any: answer += " on any change "; break;
  case rel_like: answer += " like "; break;
 }
 if (relsymbol != rel_any)
 {
  answer += itemvalue;
  switch (fieldtype)
  {
   case as_text:   answer += " (as text) ";   break;
   case as_int:    answer += " (as int) ";    break;
   case as_double: answer += " (as double) "; break;
   case as_date:   answer += " (as date) ";   break;
   case as_time:   answer += " (as time) ";   break;
   case as_stamp:  answer += " (as stamp) ";  break;
  }
 }
 return answer;
}

classRequest::classRequest (string f, string i, int r, int t, TObject * w, RefreshHandler h)
{
 fieldname = f;
 itemvalue = i;
 relsymbol = r;
 fieldtype = t;
 where = w;
 how = h;
}

bool classRequest::MatchesNotification(string f, string i)
{
 if (fieldname != f) return false;
 if (relsymbol == rel_any) return true;
 if (relsymbol == rel_e && itemvalue == i) return true;
 if (relsymbol == rel_ne && itemvalue != i) return true;
 if (relsymbol == rel_like && LikeString (itemvalue, i)) return true;
 switch (fieldtype)
 {
  case as_text: {
                 string a = itemvalue;
                 string b = i;
                 switch (relsymbol)
                 {
                  case rel_l: if (b < a) return true; break;
                  case rel_g: if (b > a) return true; break;
                  case rel_le: if (b <= a) return true; break;
                  case rel_ge: if (b >= a) return true; break;
                 }
                 return false;
                }
  case as_date: {
                 Pdate a(itemvalue, global_mode);
                 Pdate b(i, global_mode);
                 switch (relsymbol)
                 {
                  case rel_l: if (b < a) return true; break;
                  case rel_g: if (b > a) return true; break;
                  case rel_le: if (b <= a) return true; break;
                  case rel_ge: if (b >= a) return true; break;
                 }
                 return false;
                }
  case as_time: {
                 Ptime a(itemvalue, global_mode);
                 Ptime b(i, global_mode);
                 switch (relsymbol)
                 {
                  case rel_l: if (b < a) return true; break;
                  case rel_g: if (b > a) return true; break;
                  case rel_le: if (b <= a) return true; break;
                  case rel_ge: if (b >= a) return true; break;
                 }
                 return false;
                }
  case as_stamp:{
                 Pstamp a(itemvalue, global_mode);
                 Pstamp b(i, global_mode);
                 switch (relsymbol)
                 {
                  case rel_l: if (b < a) return true; break;
                  case rel_g: if (b > a) return true; break;
                  case rel_le: if (b <= a) return true; break;
                  case rel_ge: if (b >= a) return true; break;
                 }
                 return false;
                }
  case as_int:  {
                 int a = StringToInt(itemvalue);
                 int b = StringToInt(i);
                 switch (relsymbol)
                 {
                  case rel_l: if (b < a) return true; break;
                  case rel_g: if (b > a) return true; break;
                  case rel_le: if (b <= a) return true; break;
                  case rel_ge: if (b >= a) return true; break;
                 }
                 return false;
                }
  case as_double:{
                 double a = StringToFloat(itemvalue);
                 double b = StringToFloat(i);
                 switch (relsymbol)
                 {
                  case rel_l: if (b < a) return true; break;
                  case rel_g: if (b > a) return true; break;
                  case rel_le: if (b <= a) return true; break;
                  case rel_ge: if (b >= a) return true; break;
                 }
                 return false;
                }
 }
 return false;
}

void classRequest::Fulfill (int request_number, string matched_value)
{
 if (where && how)
  (where->*how)(request_number, matched_value);
}

TObject * classRequest::WasFor ()
{
 return where;
}

bool classRequest::IsSameMethodAs (classRequest & other_request)
{
 if (where == other_request.where && how == other_request.how)
  return true;
 else
  return false;
}
