#include "repeating.h"
#include <SysUtils.hpp>

//----------------------------------------------------------

/*

 many thanks to http://www.informatik.uni-ulm.de/pm/mitarbeiter/mark/day_of_week.html
 for providing this algorithm to calculate the day of the week
 of a date based on y/m/d.

 it uses an algorithm written by Lewis Caroll

 c is in [0, 99) century
 y is in [0, 99] year of century
 m is in [1, 12] month of year
 d is in [1, 31] day of month

 x = (C + Y + M + D) % 7
 
 Day Code (x)
     1 Sunday
     2 Monday
     3 Tuesday
     4 Wednesday
     5 Thursday
     6 Friday
     0 Saturday

 Century Code (C)
 16/20 0
 17/21 6
 18/22 4
 19/23 2
   C = (8 - ((c % 4) * 2)) % 8

 Only works for Gregorian - after 1582

 Year Code (Y)
   Y = (y + (y / 4)) % 7
 
 Month Code (M) [if leap year]
   Jan 1 [0]  
   Feb 4 [3]
   Mar 4
   Apr 0
   May 2
   Jun 5
   Jul 0
   Aug 3
   Sep 6
   Oct 1
   Nov 4
   Dec 6

 Day Code (D)
   D = d

 test example - 2000/01/07 is a friday (and a leap year)
   c = 21
   y = 0
   m = 1
   d = 7
   C = 6
   Y = 0
   M = 0 (leap year)
   D = 7
   x = (13) % 7 = 6

 today's example - 2001/10/08 is a monday (october 8)
   c = 21
   y = 1
   m = 10
   d = 8
   C = 6
   Y = 1
   M = 1 (not leap year)
   D = 8
   x = (C + Y + M + D) % 7 = (6 + 1 + 1 + 8) % 7 = (16) % 7 = 2
   x = 2 therefore monday

   year 2000 is 21st century
*/

//----------------------------------------------------------
//-----------------------------------------------date_range-

// okay
date_range::date_range ()
{
 // default range is 1990 to 2010 inclusive
 from.year   = 1990;
 from.month  = 1;
 from.day    = 1;
 until.year  = 2010;
 until.month = 12;
 until.day   = 31;
}

// okay
date_range::date_range (Pdate p_from, Pdate p_until)
{
 from  = p_from;
 until = p_until;
}

// okay
bool date_range::Contains (Pdate p_test)
{
 if (from <= p_test && p_test <= until)
  return true;
 else
  return false;
}

// okay
bool date_range::Contains (date_range p_test)
{
 if (Contains(p_test.from) && Contains(p_test.until) &&
     p_test.from < p_test.until)
  return true;
 else 
  return false;
}

// okay
bool date_range::IsDayType (day_items p_test)
{
 // must be one day long
 if (from != until)
  return false;

 // from here on out, from = until
 if (p_test == d_day)
  return true;
 
 int day_of_week = from.DayOfWeek();

 if (p_test == d_business_day && day_of_week > 1 && day_of_week < 7)
  return true;

 // individual days (code is safe if you want to add to the front of the list)
 if (p_test > d_business_day && p_test == day_of_week + d_business_day)
  return true;

 return false;
}

// okay
bool date_range::IsWeekType (week_items p_test)
{
 // must be 7 days
 // if w_whole_week, year and month must match on date_range.from and .until
 // and first day must be sunday?
 /* TODO:  for speed, i'm trusting this will always be the case thanks
           to whatever calls it... but this really should check. sadly,
           w_whole_week removed, because it no longer really makes sense --
           a week is always defined as sunday to monday, fitting within the given
           range, be it month or year or infinity. (A bit hard to find the first
           week of infinity, but you see my point.) */
 // if (DayLength() != 7)
 //  return false;

 // must start on sunday and end on saturday
 if (from.DayOfWeek() != 1 || until.DayOfWeek() != 7)
  return false;

 if (p_test == w_week)
  return true;

 //if (p_test == w_whole_week && from.month == until.month && from.year == until.year)
 // return true;

 return false;
}

// okay
bool date_range::IsMonthType (month_items p_test)
{
 // first and last day must fall in same month number
 if (from.year != until.year || from.month != until.month)
  return false;

 // from.month and until.month should now be identical

 if (p_test == m_month)
  return true;

 int month_number = from.month;

 // individual months (code is safe if you want to add to the front of the list)
 if (p_test > m_month && p_test == (month_items)(month_number + (int)m_month))
  return true;

 return false;
}

// okay
bool date_range::IsYearType (year_items p_test)
{
 // first and last day must fall in same year
 if (!(from.year == until.year))
  return false;
 
 if (p_test == y_year)
  return true;

 int year_number = from.year;

 if (p_test == y_leap_year && IsLeapYear(year_number))
  return true;

 return false;
}

// okay
int date_range::DayLength ()
{
 return ((from - until) + 1);
}

// okay
void date_range::Truncate (date_range p_trunc)
{
 from = max (from, p_trunc.from);
 until = min (until, p_trunc.until);
}

//-----------------------------------------------date_range-
//-------------------------------------------------day_spec-

// okay
day_spec::day_spec ()
{
 repeats  = r_any;
 specific = 1;
 matching = d_day;
}

// okay
day_spec::day_spec (repeat_items p_repeats, int p_specific, day_items p_matching)
{
 repeats  = p_repeats;
 specific = p_specific;
 matching = p_matching;
}

// okay
date_range day_spec::Refine (date_range p_range, Pdate p_searching)
{
 Pdate bad_date;
 date_range refined(bad_date, bad_date); // by default, date outside range

 // any day within the range is fine, let it through unchanged
 // if you want to restrict, use "every 1" to mean refining to
 // exactly one day
 if (repeats == r_any)
  return p_range;

 if (repeats == r_every && specific != 0)
 {
  Pdate i_date(p_range.From());
  int occur = 0;
  for (;
       i_date <= p_range.Until();
       i_date++)
  {
   date_range test_range(i_date, i_date);
   if (test_range.IsDayType(matching))
   {
    if ((occur % specific == 0) && test_range.Contains(p_searching))
    {
     refined = test_range;
     break;
    }
    occur++;
   }
  }
 }

 if (repeats == r_on && specific != 0)
 {
  Pdate i_date(p_range.From());
  int occur = 0;
  for (;
       i_date <= p_range.Until();
       i_date++)
  {
   date_range test_range(i_date, i_date);
   if (test_range.IsDayType(matching))
   {
    occur++;
    if (occur == specific)
    {
     if  (test_range.Contains(p_searching))
      refined = test_range;
     break;
    }
   }
  }
 }

 if (repeats == r_first)
 {
  Pdate i_date(p_range.From());
  for (;
       i_date <= p_range.Until();
       i_date++)
  {
   date_range test_range(i_date, i_date);
   if (test_range.IsDayType(matching))
   {
    if  (test_range.Contains(p_searching))
     refined = test_range;
    break;
   }
  }
 }

 if (repeats == r_last)
 {
  Pdate i_date(p_range.Until());
  for (;
       i_date >= p_range.From();
       i_date--)
  {
   date_range test_range(i_date, i_date);
   if (test_range.IsDayType(matching))
   {
    if  (test_range.Contains(p_searching))
     refined = test_range;
    break;
   }
  }
 }

 // if no match was found, then it will return dates WAY outside the search area
 return refined;
}

void day_spec::Load (string p_serialized)
{
 int r, s, m;
 septhree (p_serialized, ',', r, s, m);
 repeats  = (repeat_items)r;
 specific = (int)s;
 matching = (day_items)m;
}

string day_spec::Save ()
{
 string answer = IntToString(repeats) + "," + IntToString(specific) + "," + IntToString(matching);
 return answer;
}

//-------------------------------------------------day_spec-
//------------------------------------------------week_spec-

// okay
week_spec::week_spec ()
{
 repeats  = r_any;
 specific = 1;
 matching = w_week;
}

// okay
week_spec::week_spec (repeat_items p_repeats, int p_specific, week_items p_matching)
{
 repeats  = p_repeats;
 specific = p_specific;
 matching = p_matching;
}

// okay
date_range week_spec::Refine (date_range p_range, Pdate p_searching)
{
 Pdate bad_date;
 date_range refined(bad_date, bad_date); // by default, date outside range

 // any week within the range is fine, let it through unchanged
 // if you want to restrict, use "every 1" to mean refining to
 // exactly one week
 if (repeats == r_any)
  return p_range;

 // should actually restrict to a range starting with a sunday... of course,
 // i have no idea what the difference between "week" and "whole week" is...
 // and the range should end with a saturday, so that "last week" makes sense.
 Pdate new_begin = p_range.From();
 Pdate new_end = p_range.Until();
 // advance to sunday and back up to saturday
 while (new_begin.DayOfWeek() != 1)
  new_begin++;
 while (new_end.DayOfWeek() != 7)
  new_end--;

 if (repeats == r_every && specific != 0)
 {
  Pdate i_date(new_begin);
  Pdate i_week(new_begin + 6);
  int occur = 0;
  for (;
       i_week <= new_end;
       i_date = (i_week + 1), i_week.mod(1, in_weeks))
  {
   date_range test_range(i_date, i_week);
   if (test_range.IsWeekType(matching))
   {
    if ((occur % specific == 0) && test_range.Contains(p_searching))
    {
     refined = test_range;
     break;
    }
    occur++;
   }
  }
 }

 if (repeats == r_on && specific != 0)
 {
  Pdate i_date(new_begin);
  Pdate i_week(new_begin + 6);
  int occur = 0;
  for (;
       i_week <= new_end;
       i_date = (i_week + 1), i_week.mod(1, in_weeks))
  {
   date_range test_range(i_date, i_week);
   if (test_range.IsWeekType(matching))
   {
    occur++;
    if (occur == specific)
    {
     if  (test_range.Contains(p_searching))
      refined = test_range;
     break;
    }
   }
  }
 }

 if (repeats == r_first)
 {
  Pdate i_date(new_begin);
  Pdate i_week(new_begin + 6);
  for (;
       i_week <= new_end;
       i_date = (i_week + 1), i_week.mod(1, in_weeks))
  {
   date_range test_range(i_date, i_week);
   if (test_range.IsWeekType(matching))
   {
    if  (test_range.Contains(p_searching))
     refined = test_range;
    break;
   }
  }
 }

 if (repeats == r_last)
 {
  Pdate i_date(new_end - 6);
  Pdate i_week(new_end);
  for (;
       i_date >= new_begin;
       i_week = (i_date - 1), i_date.mod(-1, in_weeks))
  {
   date_range test_range(i_date, i_week);
   if (test_range.IsWeekType(matching))
   {
    if  (test_range.Contains(p_searching))
     refined = test_range;
    break;
   }
  }
 }

 // if no match was found, then it will return dates WAY outside the search area
 return refined;
}

void week_spec::Load (string p_serialized)
{
 int r, s, m;
 septhree (p_serialized, ',', r, s, m);
 repeats  = (repeat_items)r;
 specific = (int)s;
 matching = (week_items)m;
}

string week_spec::Save ()
{
 string answer = IntToString(repeats) + "," + IntToString(specific) + "," + IntToString(matching);
 return answer;
}
//------------------------------------------------week_spec-
//-----------------------------------------------month_spec-

// okay
month_spec::month_spec ()
{
 repeats  = r_any;
 specific = 1;
 matching = m_month;
}

// okay
month_spec::month_spec (repeat_items p_repeats, int p_specific, month_items p_matching)
{
 repeats  = p_repeats;
 specific = p_specific;
 matching = p_matching;
}

// okay
date_range month_spec::Refine (date_range p_range, Pdate p_searching)
{
 Pdate bad_date;
 date_range refined(bad_date, bad_date); // by default, date outside range

 // any month within the range is fine, let it through unchanged
 // if you want to restrict, use "every 1" to mean refining to
 // exactly one month
 if (repeats == r_any)
  return p_range;

 Pdate new_begin = p_range.From();
 Pdate new_end = p_range.Until();

 if (new_begin.day > 1)
 {
  new_begin.mod(1, in_months);
  new_begin.day = 1;
 }

 if (new_end.day < LastDayOf(new_end.month, new_end.year))
 {
  new_end.mod(-1, in_months);
  new_end.day = LastDayOf(new_end.month, new_end.year);
 }

 // new_begin and new_end are now assured to be month boundaries
 // (first and last days of respective months)

 if (repeats == r_every && specific != 0)
 {
  Pdate i_date(new_begin);
  Pdate i_month(new_begin);
  i_month.day = LastDayOf (i_month.month, i_month.year);
  int occur = 0;
  for (;
       i_month <= new_end;
       i_date = (i_month + 1), i_month.mod(1, in_months), i_month.day = LastDayOf(i_month.month, i_month.year))
  {
   date_range test_range(i_date, i_month);
   if (test_range.IsMonthType(matching))
   {
    if ((occur % specific == 0) && test_range.Contains(p_searching))
    {
     refined = test_range;
     break;
    }
    occur++;
   }
  }
 }

 if (repeats == r_on && specific != 0)
 {
  Pdate i_date(new_begin);
  Pdate i_month(new_begin);
  i_month.day = LastDayOf (i_month.month, i_month.year);
  int occur = 0;
  for (;
       i_month <= new_end;
       i_date = (i_month + 1), i_month.mod(1, in_months), i_month.day = LastDayOf(i_month.month, i_month.year))
  {
   date_range test_range(i_date, i_month);
   if (test_range.IsMonthType(matching))
   {
    occur++;
    if (occur == specific)
    {
     if  (test_range.Contains(p_searching))
      refined = test_range;
     break;
    }
   }
  }
 }

 if (repeats == r_first)
 {
  Pdate i_date(new_begin);
  Pdate i_month(new_begin);
  i_month.day = LastDayOf (i_month.month, i_month.year);
  for (;
       i_month <= new_end;
       i_date = (i_month + 1), i_month.mod(1, in_months), i_month.day = LastDayOf (i_month.month, i_month.year))
  {
   date_range test_range(i_date, i_month);
   if (test_range.IsMonthType(matching))
   {
    if  (test_range.Contains(p_searching))
     refined = test_range;
    break;
   }
  }
 }

 if (repeats == r_last)
 {
  Pdate i_date(new_end);
  Pdate i_month(new_end);
  i_date.day = 1;
  for (;
       i_date >= new_begin;
       i_month = (i_date - 1), i_date.mod(-1, in_months), i_date.day = 1)
  {
   date_range test_range(i_date, i_month);
   if (test_range.IsMonthType(matching))
   {
    if  (test_range.Contains(p_searching))
     refined = test_range;
    break;
   }
  }
 }

 // if no match was found, then it will return dates WAY outside the search area
 return refined;
}

void month_spec::Load (string p_serialized)
{
 int r, s, m;
 septhree (p_serialized, ',', r, s, m);
 repeats  = (repeat_items)r;
 specific = (int)s;
 matching = (month_items)m;
}

string month_spec::Save ()
{
 string answer = IntToString(repeats) + "," + IntToString(specific) + "," + IntToString(matching);
 return answer;
}
//-----------------------------------------------month_spec-
//------------------------------------------------year_spec-

// okay
year_spec::year_spec ()
{
 repeats  = r_any;
 specific = 1;
 matching = y_year;
}

// okay
year_spec::year_spec (repeat_items p_repeats, int p_specific, year_items p_matching)
{
 repeats  = p_repeats;
 specific = p_specific;
 matching = p_matching;
}

// okay
date_range year_spec::Refine (date_range p_range, Pdate p_searching)
{
 Pdate bad_date;
 date_range refined(bad_date, bad_date); // by default, date outside range

 // any year within the range is fine, let it through unchanged
 // if you want to restrict, use "every 1" to mean refining to
 // exactly one year
 if (repeats == r_any)
  return p_range;

 Pdate new_begin = p_range.From();
 Pdate new_end = p_range.Until();

 if (new_begin.month != 1 || new_begin.day != 1)
 {
  new_begin.day = 1;
  new_begin.month = 1;
  new_begin.year = new_begin.year + 1;
 }

 if (new_end.month != 12 || new_end.day != 31)
 {
  new_end.day = 31;
  new_end.month = 12;
  new_end.year = new_end.year - 1;
 }

 // new_begin and new_end are now assured to be year boundaries
 // (first and last days of respective years)

 if (repeats == r_every && specific != 0)
 {
  Pdate i_date(new_begin);
  Pdate i_year(new_begin);
  i_year.day = 31;
  i_year.month = 12;
  int occur = 0;
  for (;
       i_year <= new_end;
       i_date = (i_year + 1), i_year.mod(1, in_years))
  {
   date_range test_range(i_date, i_year);
   if (test_range.IsYearType(matching))
   {
    if ((occur % specific == 0) && test_range.Contains(p_searching))
    {
     refined = test_range;
     break;
    }
    occur++;
   }
  }
 }

 if (repeats == r_on && specific != 0)
 {
  Pdate i_date(new_begin);
  Pdate i_year(new_begin);
  i_year.day = 31;
  i_year.month = 12;
  int occur = 0;
  for (;
       i_year <= new_end;
       i_date = (i_year + 1), i_year.mod(1, in_years))
  {
   date_range test_range(i_date, i_year);
   if (test_range.IsYearType(matching))
   {
    occur++;
    if (occur == specific)
    {
     if  (test_range.Contains(p_searching))
      refined = test_range;
     break;
    }
   }
  }
 }

 if (repeats == r_first)
 {
  Pdate i_date(new_begin);
  Pdate i_year(new_begin);
  i_year.month = 12;
  i_year.day = 31;
  for (;
       i_year <= new_end;
       i_date = (i_year + 1), i_year.mod(1, in_years))
  {
   date_range test_range(i_date, i_year);
   if (test_range.IsYearType(matching))
   {
    if  (test_range.Contains(p_searching))
     refined = test_range;
    break;
   }
  }
 }

 if (repeats == r_last)
 {
  Pdate i_date(new_end);
  Pdate i_year(new_end);
  i_date.day = 1;
  i_date.month = 1;
  for (;
       i_date >= new_begin;
       i_year = (i_date - 1), i_date.mod(-1, in_years), i_date.day = 1, i_date.month = 1)
  {
   date_range test_range(i_date, i_year);
   if (test_range.IsYearType(matching))
   {
    if  (test_range.Contains(p_searching))
     refined = test_range;
    break;
   }
  }
 }

 // if no match was found, then it will return dates WAY outside the search area
 return refined;
}

void year_spec::Load (string p_serialized)
{
 int r, s, m;
 septhree (p_serialized, ',', r, s, m);
 repeats  = (repeat_items)r;
 specific = (int)s;
 matching = (year_items)m;
}

string year_spec::Save ()
{
 string answer = IntToString(repeats) + "," + IntToString(specific) + "," + IntToString(matching);
 return answer;
}
//------------------------------------------------year_spec-
string RepeatingString (string p_rep)
{
 // okay, we want to do something neat with this repeating string, okay?
 // so we're gonna make an object with it, and ... figure out the first
 // date that it happened. but what about repeating ones? what then?
 // first date is nice, but if it's every christmas, do we want to show, say
 // the first christmas of the century? several examples take up space ...
 // next occurrence would be nice. so we might take the first date, and walk
 // it until we find something more pallatable.
 p_rep = ReplaceStringWithString(p_rep, "-", "/");
 string answer_string;
 string prepend_string;
 repeat_spec<int> repeating_stuff;
 repeating_stuff.Load(p_rep);
 vector<Pdate> answer_dates = repeating_stuff.AllMatches(date_range(repeating_stuff.Start(), repeating_stuff.Stop()));
 Pdate right_now;

 // figure out the current date ... requires sysutils, which breaks stuff for
 // compatibility, but, oh well.
 {
  TDateTime stuff = Date();
  unsigned short y, m, d;
  stuff.DecodeDate(&y, &m, &d);
  right_now.year = y;
  right_now.month = m;
  right_now.day = d;
 }

 if (answer_dates.size() == 0)
 {
  answer_string = "Never";
 }
 else
 {
  // if one date, just go with that.
  Pdate choice_date = answer_dates[0];
  prepend_string = "";

  // if there are multiple dates returned ...
  // all in the past, 'last xx/xx/xxxx'
  if (answer_dates.size() > 1)
  {
   if (answer_dates[answer_dates.size() - 1] < right_now)
   {
    prepend_string = "Last ";
    choice_date = answer_dates[answer_dates.size() - 1];
   }
   // all in the future, 'next: xx/xx/xxxx'
   // both sides, 'next: xx/xx/xxxx'
   // thus, next one from now
   else
   {
    prepend_string = "Next" ;
    for (int i = (answer_dates.size() - 1); i >= 0; i--)
    {
     if (answer_dates[i] > right_now)
      choice_date = right_now;
     else
      break;
    }
   }
  }
  // make pretty
  Pdate tomorrow_date = right_now;
  tomorrow_date.mod(1, in_days);

//  if (choice_date == right_now)
//   answer_string = "Today";
//  else if (choice_date == tomorrow_date)
//   answer_string = "Tomorrow";
//  else
   answer_string = /* prepend_string + */ choice_date.OutDate(mdy | yyyy);
 }

 return answer_string;
}
//-----------------------------------------------------done-
