#ifndef tensors_h
#define tensors_h

#include <vector>
#include <list>
#include <stdarg.h>
#include <math.h>

using namespace std;

#define T_ERROR_VEC_SIZE    1
#define T_ERROR_DIVIDE_ZERO 2

#include "tcpip_messaging.h"

/*
-----------------------------Units-
length:       meter     (m)
surface:      m m
volume:       m m m
angle:        radians   (rad)
mass:         kilogram  (kg)
density:      kg / m m m
time:         second    (s)
velocity:     m / s
ang. veloc.:  rad / s
              2 pi / s
acceleration: m / s s
ang. accel.:  rad / s s
force:        netwon    (N)
              kg m / s s
work:         joule     (J)
              N m
              kg m m / s s
power:        watt      (W)
              J / s
              N m / s
              kg m m / s s s
pressure:     pascal    (Pa)
              N / m m
              kg / m s s
frequency:    hertz     (Hz)
              1 / s
elec. charge: coulomb   (C)
              A s
elec. inten.: amp       (A)
              C / s
              2e-7 N
elec. tens.:  volt      (V)
              kg m m / A s s s
              W / A
elec. field:  V / m
              kg m / A s s s
magn. field:  tesla     (T)
              kg / A s s
resistance:   ohm       (ohm)
              V / A
              kg m m / A A s s s
capacity:     fahrad    (F)
              A A s s s s / kg m m
induction:    henry     (H)
              kg m m  / A A s s
temperature:  kelvin    (K)
light inten.: candle?   (cd)
              1/683 W/steradian (whatever that is)
quantity:     mol       (mol)
specif. heat: J / g K (not kilogram?)
molar heat:   J / mol K (just another way)

-----------------------------Constants-
specific heat of water, liquid:
  4184 J / kg K
speed of light:
  299792458 m / s
avogadro (number of atoms per mol):
  6.022136736e23 / mol
gravitation:
  6.6725985e-11 N m m / kg kg
electron charge:
  1.60217733e-19 C
mass of an electron:
  9.1093897e-31 kg
mass of a proton:
  1.6726231e-27 kg
mass of a neutron:
  1.6749286e-27 kg
plank constant:
  6.63e-34

earth:
  mass:
    5.98e24 kg
  radius:
    6.38e6 m
  distance to sun:
    1.50e11 m
  length of day:
    86164 s (23 hours, 56 minutes, and 4 seconds)
  length of solar day:
    86400 s (24 hours)
  gravity (paris):
    9.81 N / kg
moon:
  mass:
    7.4e22 kg
  radius:
    1.74e6 m
  distance to earth:
    3.84e8 m
sun:
  mass:
    1.99e30 kg
  radius:
    6.96e8 m
-----------------------------Relationships-
kinetic energy (movement):
  J = 0.5 kg (m / s) (m / s)
  work to change speed = difference in kinetic energy at the two speeds
potential energy (falling):
  J = kg (N / kg) m
total engergy (mechanical):
  J = (0.5 kg (m / s) (m / s)) + (kg (N / kg) m)
power received by elect. resistor:
  J = ohm A A
power received by alternating current:
  J = V A cos (q)
gravitational attraction:
  N = G kg(a) kg(b) / m m
orbit:
  see book (not handy right this second)
acceleration in an electric field:
  m / s s = V C / kg m
spring:
  N = (m - m) N / m
potential energy in spring:
  J = 0.5 (m - m) (m - m) N / m
period of an elastic oscillator:
  s = 2pi sqrt(kg / (N / m))
period of a simple pendulum:
  s = 2pi sqrt(m / [g])
energy of a photon:
  J = [plank] Hz
*/

// TODO: declare a new namespace for the things below

typedef double dimension;

template <unsigned int D>
class t_vector
{
 public:
  vector<dimension> dimensions;

 public:
  t_vector () : dimensions(D, 0.00)
  {
   // nothing to do?
  }

  t_vector (unsigned int dimension_count, ...) : dimensions(D, 0.00)
  {
   va_list argument_list;
   dimension next_dimension;

   va_start (argument_list, dimension_count);

   unsigned int i = 0;
   while (dimension_count > 0)
   {
    next_dimension = va_arg (argument_list, dimension);
    dimensions[i] = next_dimension;
    dimension_count--;
    i++;
   }

   va_end (argument_list);
  }

  t_vector (const vector<dimension> & copy_data)
  {
   dimensions = copy_data;
   dimensions.resize(D);
  }

  dimension size () const
  {
   return sqrt(size_squared());
  }

  dimension & operator [] (unsigned int i)
  {
   if (D <= i)
    throw (T_ERROR_VEC_SIZE);
   return dimensions[i];
  }

  const dimension & operator [] (unsigned int i) const
  {
   if (D <= i)
    throw (T_ERROR_VEC_SIZE);
   return dimensions[i];
  }

  void reset ()
  {
   unsigned int dimension_count = D;
   while (dimension_count > 0)
   {
    dimension_count--;
    dimensions[dimension_count] = 0.00;
   }
  }

  dimension size_squared () const
  {
   dimension running_total = 0.00;
   unsigned int dimension_count = D;
   while (dimension_count > 0)
   {
    dimension_count--;
    running_total += (dimensions[dimension_count] * dimensions[dimension_count]);
   }
   return running_total;
  }

  t_vector<D> & operator += (const t_vector<D> & rhs)
  {
   unsigned int dimension_count = D;
   // if (dimension_count != rhs.dimensions.size())
   //  throw (T_ERROR_VEC_SIZE);

   while (dimension_count > 0)
   {
    dimension_count--;
    dimensions[dimension_count] += rhs[dimension_count];
   }

   return *this;
  }

  t_vector<D> & operator -= (const t_vector<D> & rhs)
  {
   unsigned int dimension_count = D;
   // if (dimension_count != rhs.dimensions.size())
   //  throw (T_ERROR_VEC_SIZE);

   while (dimension_count > 0)
   {
    dimension_count--;
    dimensions[dimension_count] -= rhs[dimension_count];
   }

   return *this;
  }

  t_vector<D> & operator *= (const dimension & rhs)
  {
   int dimension_count = D;

   while (dimension_count > 0)
   {
    dimension_count--;
    dimensions[dimension_count] *= rhs;
   }

   return *this;
  }

  t_vector<D> & operator /= (const dimension & rhs)
  {
   if (rhs == 0.00)
    throw (T_ERROR_DIVIDE_ZERO);

   int dimension_count = D;

   while (dimension_count > 0)
   {
    dimension_count--;
    dimensions[dimension_count] /= rhs;
   }

   return *this;
  }
};

template <unsigned int D>
t_vector<D> operator + (const t_vector<D> & lhs, const t_vector<D> & rhs)
{
 unsigned int dimension_count = D;
 // if (dimension_count != rhs.dimensions.size())
 //  throw (T_ERROR_VEC_SIZE);

 vector<dimension> dimensions(D, 0.00);

 while (dimension_count > 0)
 {
  dimension_count--;
  dimensions[dimension_count] = lhs[dimension_count] + rhs[dimension_count];
 }

 return dimensions;
}

template <unsigned int D>
t_vector<D> operator - (const t_vector<D> & lhs, const t_vector<D> & rhs)
{
 unsigned int dimension_count = D;
 // if (dimension_count != rhs.dimensions.size())
 //  throw (T_ERROR_VEC_SIZE);

 vector<dimension> dimensions(D, 0.00);

 for (unsigned int i = 0; i < dimension_count; i++)
  dimensions[i] = lhs[i] - rhs[i];

 return dimensions;
}

// DOT-PRODUCT (a * b = |a| * |b| * cos(theta))
template <unsigned int D>
dimension operator * (const t_vector<D> & lhs, const t_vector<D> & rhs)
{
 dimension answer = 0.00;
 unsigned int dimension_count = D;
 // if (dimension_count != rhs.dimensions.size())
 //  throw (T_ERROR_VEC_SIZE);
 for (unsigned int i = 0; i < dimension_count; i++)
  answer += (lhs[i] * rhs[i]);
 return answer;
}

// DOT-PRODUCT (alias)
template <unsigned int D>
dimension dot_product (const t_vector<D> & a, const t_vector<D> & b)
{
 return a * b;
}

// angle from dot-product
template <unsigned int D>
dimension vector_angle (const t_vector<D> & a, const t_vector<D> & b)
{
 dimension oa = a.size();
 dimension ob = b.size();
 if (!oa || !ob)
  throw(T_ERROR_DIVIDE_ZERO);
 return (a * b) / (oa * ob);
}

// CROSS-PRODUCT (third vector orthogonal to first two, length =  area of //-ogram formed by other vectors)
template <unsigned int D>
t_vector<D> operator ^ (const t_vector<D> & lhs, const t_vector<D> & rhs)
{
 unsigned int dimension_count = D;
 vector<dimension> dimensions(D, 0.00);
 // if (dimension_count != rhs.dimensions.size())
 //  throw (T_ERROR_VEC_SIZE);
 unsigned int i;
 unsigned int j;
 for (unsigned int k = 0; k < dimension_count; k++)
 {
  i = (k + 1) % dimension_count;
  j = (i + 1) % dimension_count;
  dimensions[k] = (lhs[i] * rhs[j]) - (lhs[j] * rhs[i]);
 }
 return dimensions;
}

// CROSS-PRODUCT (alias)
template <unsigned int D>
t_vector<D> cross_product (const t_vector<D> & a, const t_vector<D> & b)
{
 return a ^ b;
}

// SCALAR-TRIPLE-PRODUCT (volume of space formed by //-liped formed by other vectors)
template <unsigned int D>
dimension triple_product (const t_vector<D> & a, const t_vector<D> & b, const t_vector<D> & c)
{
 return a * (b ^ c);
}

template <unsigned int D>
t_vector<D> operator * (const t_vector<D> & lhs, const dimension & mult)
{
 unsigned int dimension_count = D;
 vector<dimension> dimensions(D, 0.00);

 for (unsigned int i = 0; i < dimension_count; i++)
  dimensions[i] = lhs[i] * mult;

 return dimensions;
}

template <unsigned int D>
t_vector<D> operator * (const dimension & mult, const t_vector<D> & rhs)
{
 return (rhs * mult);
}

template <unsigned int D>
t_vector<D> operator / (const t_vector<D> & lhs, const dimension & div)
{
 if (div == 0.00)
  throw (T_ERROR_DIVIDE_ZERO);
 unsigned int dimension_count = D;
 vector<dimension> dimensions(D, 0.00);

 for (unsigned int i = 0; i < dimension_count; i++)
  dimensions[i] = lhs[i] / div;

 return dimensions;
}

template <unsigned int D>
t_vector<D> operator / (const dimension & div, const t_vector<D> & rhs)
{
 return (rhs / div);
}

template <unsigned int D>
bool operator == (const t_vector<D> & lhs, const t_vector<D> & rhs)
{
 unsigned int dimension_count = D;
 // if (dimension_count != rhs.dimensions.size())
 //  throw (T_ERROR_VEC_SIZE);
 for (unsigned int i = 0; i < dimension_count; i++)
  if (lhs[i] != rhs[i])
   return false;
 return true;
}

template <unsigned int D>
bool operator != (const t_vector<D> & lhs, const t_vector<D> & rhs)
{
 return !(lhs == rhs);
}

template <unsigned int D>
dimension point_segment_distance (const t_vector<D> & a, const t_vector<D> & b, const t_vector<D> & m)
{
 if (a != b)
 {
  t_vector<D> ab = b - a;
  t_vector<D> mh = ab; // just to initialize to the right size without useless calculation
  mh = ((a - m) ^ (ab)) / ab.size();
  return mh.size();
 } else
 {
  return ((m - a).size());
 }
}

// TODO: add determinant, etc.

template <unsigned int D>
class node_base
{
 public:
  t_vector<D> position;
  t_vector<D> velocity;
  t_vector<D> acceleration;
  t_vector<D> force;
  dimension      mass;
  dimension      drag;
  dimension      minimum_velocity;
  dimension      charge;
  bool           infinite_mass;

 public:
                    node_base () {};
                    node_base (const t_vector<D> & p_position, dimension p_mass, dimension p_drag, dimension p_minimum_velocity, dimension p_charge, bool p_infinite_mass) :
                     position (p_position), mass (p_mass), drag (p_drag), minimum_velocity (p_minimum_velocity), charge (p_charge), infinite_mass (p_infinite_mass) {}
            virtual ~node_base () {};
  void      virtual push (const t_vector<D> & additional_force) = 0;
  void      virtual apply_force () = 0;
  void      virtual apply_charge_mutually (node_base<D> * rhs) = 0;
  void      virtual apply_acceleration (dimension delta_time) = 0;
  void      virtual apply_velocity (dimension delta_time) = 0;
  void      virtual check_position () = 0;
  dimension virtual kinetic_energy () const = 0;
  dimension virtual velocity_squared () const = 0;
};

template <unsigned int D>
class node_standard : virtual public node_base<D>
{
 public:
  node_standard (const t_vector<D> & p_position, dimension p_mass, dimension p_drag, dimension p_minimum_velocity, dimension p_charge, bool p_infinite_mass) :
   node_base<D> (p_position, p_mass, p_drag, p_minimum_velocity, p_charge, p_infinite_mass)
  {
  }

  void push (const t_vector<D> & additional_force)
  {
   force += additional_force;
   return;
  }

  void apply_force ()
  {
   if (mass == 0.00)
    throw (T_ERROR_DIVIDE_ZERO);
   acceleration = force / mass;
   force.reset();
   return;
  }

  void apply_charge_mutually (node_base<D> * rhs)
  {
   if (rhs)
   {
     node_base<D> * lhs = this;
     if (lhs != rhs)
     {
      t_vector<D> vec_ab = rhs->position - lhs->position;
      if (lhs->position == rhs->position)
       return;
      dimension dist_ab_squared = vec_ab.size_squared();
      dimension dist_ab = sqrt(dist_ab_squared);
      t_vector<D> unit_vec_ab = vec_ab / dist_ab; // _squared;
      dimension force = (rhs->charge * lhs->charge) / (dist_ab_squared);
      t_vector<D> force_a = unit_vec_ab * force;
      t_vector<D> force_b = force_a * (-1.00);
      rhs->push (force_a);
      lhs->push (force_b);
     }
   }
  }

  void apply_acceleration (dimension delta_time)
  {
   velocity += delta_time * acceleration;
   acceleration.reset();
   return;
  }

  void apply_velocity (dimension delta_time)
  {
   // TODO: shouldn't massive objects be less easily slowed?
   // yeah, drag should be f(mass), most likely. but bah.
   velocity *= delta_time * drag;
   if ((velocity.size() > minimum_velocity) && (!infinite_mass))
    position += delta_time * velocity;
   else
    velocity.reset();
   check_position ();
   return;
  }

  void check_position ()
  {
   // constraint!
  }

  dimension kinetic_energy () const
  {
   dimension answer = 0.5 * mass * velocity.size_squared();
   return answer;
  }

  dimension velocity_squared () const
  {
   dimension answer = velocity.size_squared();
   return answer;
  }
};

template <unsigned int D>
class tensor_base
{
 public:
  node_base<D> * node_a;
  node_base<D> * node_b;
  dimension tension;
  dimension size;

 public:
                    tensor_base() {};
                    tensor_base (node_base<D> * p_node_a, node_base<D> * p_node_b, dimension p_tension, dimension p_size) :
                     node_a (p_node_a), node_b (p_node_b), tension (p_tension), size (p_size) {}
            virtual ~tensor_base() {};
  void      virtual apply_force () const = 0;
  dimension virtual potential_energy () const = 0;
};

template <unsigned int D>
class tensor_standard : virtual public tensor_base<D>
{
 public:
  tensor_standard (node_base<D> * p_node_a, node_base<D> * p_node_b, dimension p_tension, dimension p_size) :
   tensor_base<D> (p_node_a, p_node_b, p_tension, p_size)
  {
  }

  dimension potential_energy () const
  {
   dimension result;
   if (node_a && node_b)
   {
    dimension size_diff = (node_b->position - node_a->position).size() - size;
    dimension answer = 0.5 * tension * (size_diff * size_diff);
    result = answer;
   }
   else result = 0.00;
   return result;
  }

  void apply_force () const
  {
   if (node_a && node_b)
   {
     if (node_a != node_b)
     {
      t_vector<D> vec_ab = node_b->position - node_a->position;
      if (node_a->position == node_b->position)
       return;
      dimension dist_ab_squared = vec_ab.size_squared();
      dimension dist_ab = sqrt(dist_ab_squared);
      t_vector<D> unit_vec_ab = vec_ab / dist_ab; // _squared;
      dimension force = tension * (dist_ab - size);
      t_vector<D> force_a = unit_vec_ab * force;
      t_vector<D> force_b = force_a * (-1.00);
      node_a->push (force_a);
      node_b->push (force_b);
     }
   }
  }
};

template <unsigned int D>
class tensor_angular : virtual public tensor_standard <D>
{
 public:
  t_vector<D> center;

  tensor_angular (node_base<D> * p_node_a, node_base<D> * p_node_b, dimension p_tension, dimension p_size, t_vector<D> p_center) :
   tensor_standard<D> (p_node_a, p_node_b, p_tension, p_size), center (p_center)
  {
  }

  void apply_force () const
  {
   if (node_a && node_b)
   {
     if (node_a != node_b)
     {
      /*
      if (node_a->position == node_b->position)
       return;
      t_vector<D> vec_oa  = node_a->position - center;
      t_vector<D> vec_ob  = node_b->position - center;
      dimension ang_oab;
      try
      {
       ang_oab = vector_angle (vec_oa, vec_ob);
      } catch(...)
      {
       return;
      }
      dimension   lin_oa  = vec_oa.size();
      dimension   lin_ob  = vec_ob.size();
      dimension   ang_f   = tension * (ang_oab - size);  // try to put them at a specific angle
      dimension   lin_f   = tension * (lin_oa - lin_ob); // try to put them at the same radius
      t_vector<D> nor_oa  = vec_oa / vec_oa.size();
      t_vector<D> nor_ob  = vec_ob / vec_ob.size();
      t_vector<D> vec_f_a = nor_oa * lin_f * (-1.00);
      t_vector<D> vec_f_b = nor_ob * lin_f;
      t_vector<D> vec_o   = cross_product(vec_oa, vec_ob);
      t_vector<D> nor_o   = vec_o / vec_o.size();
      t_vector<D> nor_a   = cross_product(nor_o, nor_oa);
      t_vector<D> nor_b   = cross_product(nor_o, nor_ob);
      t_vector<D> vec_g_a = nor_a * ang_f;
      t_vector<D> vec_g_b = nor_b * ang_f * (-1.00);
      t_vector<D> force_a = vec_f_a + vec_g_a;
      t_vector<D> force_b = vec_f_b + vec_g_b;
      node_a->push (force_a);
      node_b->push (force_b);
      */
     }
   }
  }
};

template <unsigned int D>
class world_base
{
 public:
  list<node_base<D> *> nodes;
  list<tensor_base<D> *> tensors;

 public:
  world_base ()
  {
  }

  ~world_base ()
  {
   for (list<tensor_base<D> *>::iterator i_tensor = tensors.begin(); i_tensor != tensors.end(); i_tensor++)
    delete (*i_tensor);
   for (list<node_base<D> *>::iterator i_node = nodes.begin(); i_node != nodes.end(); i_node++)
    delete (*i_node);
  }

  dimension stat_tensor_potential_energy ()
  {
   dimension answer = 0.00;
   for (list<tensor_base<D> *>::iterator i_tensor = tensors.begin(); i_tensor != tensors.end(); i_tensor++)
    answer += (*i_tensor)->potential_energy();
   return answer;
  }

  dimension stat_node_kinetic_energy ()
  {
   dimension answer = 0.00;
   for (list<node_base<D> *>::iterator i_node = nodes.begin(); i_node != nodes.end(); i_node++)
    answer += (*i_node)->kinetic_energy();
   return answer;
  }

  dimension stat_node_max_velocity ()
  {
   dimension answer    = 0.00;
   dimension local_velocity_squared;
   for (list<node_base<D> *>::iterator i_node = nodes.begin(); i_node != nodes.end(); i_node++)
   {
    local_velocity_squared = (*i_node)->velocity_squared();
    if (local_velocity_squared > answer)
     answer = local_velocity_squared;
   }
   return sqrt(answer);
  }

  void simulate (dimension delta_time)
  {
   for (list<tensor_base<D> *>::iterator i_tensor = tensors.begin(); i_tensor != tensors.end(); i_tensor++)
    (*i_tensor)->apply_force();
   for (list<node_base<D> *>::iterator i_node = nodes.begin(); i_node != nodes.end(); i_node++)
   {
    (*i_node)->apply_force();
    (*i_node)->apply_acceleration(delta_time);
    (*i_node)->apply_velocity(delta_time);
   }
   return;
  }
};

/*
   list<N *>::iterator i_node_b;
   for (list<N *>::iterator i_node_a = nodes.begin(); i_node_a != nodes.end(); i_node_a ++)
   {
    i_node_b = i_node_a;
    for (i_node_b ++; i_node_b != nodes.end(); i_node_b++)
     (*i_node_a)->apply_charge_mutually((*i_node_b));
   }
*/

#endif
