//
// Exception Handling for Windows
//
// Based loosely on ExceptionHandler.cpp
// Author: Hans Dietrich (hdietrich2@hotmail.com)
// Original license: Public Domain
//
// Some code partially derived from http://vovaprog.narod.ru/information/debug_exception.cpp.txt,
// and partially derived from The Eternity Engine (released by me to be used in concert with Prometheus for any
// purpose whatsoever).
//

#include <windows.h>
#include <tchar.h>
#include <stdio.h>
#include <string>
using namespace std;

#include "vectoredexception.h"
#pragma hdrstop


//======================================================================================================================
//
// Run-time loading for AddVectoredExceptionHandler
//
// jhaley: Wouldn't you know, C++Builder's headers are missing the one function I need? >_>
//

typedef LONG  (CALLBACK *PVECTORED_EXCEPTION_HANDLER)(PEXCEPTION_POINTERS ExceptionInfo);

typedef PVOID (WINAPI *PADDVECTOREDEXCEPTIONHANDLER)(ULONG FirstHandler, PVECTORED_EXCEPTION_HANDLER VectoredHandler);


//======================================================================================================================
//
// Macros
//

#define charcount(arr) (sizeof(arr) / sizeof(arr[0]))
#define LOG_BUFFER_SIZE 32768
#define MEGABYTE (1024*1024)
#define RoundMem(amt) (((amt) + MEGABYTE - 1) / MEGABYTE)
#define CODETOPRINT 16
#define STACKTOPRINT 2780
#define OUTPUT_DEBUG_STRING_EXCEPTION 0x40010006ul

//======================================================================================================================
//
// Data
//

static int exceptionCaught = 0;
static TCHAR moduleFileName[MAX_PATH*2];
static TCHAR moduleName[MAX_PATH*2];
static TCHAR crashModulePath[MAX_PATH*2];
static HANDLE logFile;
static TCHAR logbuffer[LOG_BUFFER_SIZE];
static TCHAR *fileName;

static PEXCEPTION_RECORD exceptionRecord;
static PCONTEXT contextRecord;

// jhaley 20110627: Added for Prometheus - EXE version
static bool haveEXEVersion;
static int  versionMajor;
static int  versionMinor;
static int  versionRelease;

//======================================================================================================================
//
// Static Routines
//

//
// lstrrchr
//
static TCHAR *lstrrchr(LPCTSTR str, int ch)
{
   TCHAR *start = (TCHAR *)str;
   
   while(*str++)
      ; // find end

   // search backward
   while(--str != start && *str != (TCHAR)ch)
      ;
   
   if(*str == (TCHAR)ch)
      return (TCHAR *)str; // found it
   
   return NULL;
}

//
// ExtractFileName
//
// Gets the filename from a path string
//
static TCHAR *ExtractFileName(LPCTSTR path)
{
   TCHAR *ret;

   // look for last instance of a backslash
   ret = lstrrchr(path, _T('\\'));
   if(ret)
      ++ret;
   else
      ret = (TCHAR *)path;
   
   return ret;
}

//
// GetModuleName
//
// Gets the module file name as a full path, and extracts the file portion.
//
static void GetModuleName(void)
{
   TCHAR *dot;

   ZeroMemory(moduleFileName, sizeof(moduleFileName));

   if(GetModuleFileName(NULL, moduleFileName, charcount(moduleFileName) - 2) <= 0)
      lstrcpy(moduleFileName, _T("Unknown"));

   fileName = ExtractFileName(moduleFileName);
   lstrcpy(moduleName, fileName);

   // find extension and remove it
   dot = lstrrchr(moduleName, _T('.'));
   if(dot)
      dot[0] = 0;

   // put filename onto module path
   lstrcpy(fileName, _T("CRASHLOG.CSV"));
}

//======================================================================================================================
//
// Log File Stuff
//

static void OutputHeader();

//
// OpenLogFile
//
// Opens the exception log file.
//
static int OpenLogFile(void)
{
   logFile = CreateFile(moduleFileName, GENERIC_WRITE, 0, 0, OPEN_ALWAYS,
                        FILE_ATTRIBUTE_NORMAL | FILE_FLAG_WRITE_THROUGH, 0);

   if(logFile == INVALID_HANDLE_VALUE)
      return 0;
   else
   {
      DWORD highend = 0;

      // set to append
      SetFilePointer(logFile, 0, 0, FILE_END);

      if(GetFileSize(logFile, &highend) == 0 && highend == 0)
         OutputHeader();

      return 1;
   }
}

//
// LogPrintf
//
// This is supposed to be safe to call during an exception, unlike fprintf.
// We'll see.
//
static void LogPrintf(LPCTSTR fmt, ...)
{
   DWORD bytecount;
   va_list args;
   
   va_start(args, fmt);

   memset(logbuffer, 0, sizeof(logbuffer));
   vsnprintf(logbuffer, sizeof(logbuffer), fmt, args);
   WriteFile(logFile, logbuffer, lstrlen(logbuffer), &bytecount, 0);
   
   va_end(args);
}

//
// OutputHeader
//
// jhaley 20110812: For Prometheus CSV-format output, a header needs to be emitted if the file is empty.
//
static void OutputHeader()
{
  LogPrintf("program_name,code,p_module_name,log_segment,log_address,exe_version_major,exe_version_minor,exe_version_build,"
            "app_state,error_timestamp,run_by,os_platform_id,os_version_major,os_version_minor,os_version_build,"
            "num_procs,proc_type,mem_in_use,phys_mem,phys_mem_free,page_file,page_file_free,user_addrspace,user_freespace,"
            "av_address,av_mode,ctx_edi,ctx_esi,ctx_eax,ctx_ebx,ctx_ecx,ctx_edx,ctx_eip,ctx_ebp,ctx_segcs,ctx_eflags,"
            "ctx_esp,ctx_segss,cs_eip,stack\r\n");
}

//======================================================================================================================
//
// Information Output Routines
//

// Variables needed to get logical address. Let's not put all this on the stack, in case of stack overflows.
struct headervars_t
{
  DWORD                    hMod;
  PIMAGE_DOS_HEADER        pDosHdr;
  PIMAGE_NT_HEADERS        pNtHdr;
  PIMAGE_SECTION_HEADER    pSec;
  DWORD                    rva;
  DWORD                    section;
  DWORD                    offset;
  MEMORY_BASIC_INFORMATION memoryInfo;
};

static headervars_t headerVars;
static DWORD last_SegCs;
static DWORD last_Eip;

//
// PrintHeader
//
// Prints a header for the exception.
//
static void PrintHeader(void)
{
   TCHAR *crashModuleFn = _T("Unknown");
   bool  gotLogicalAddress = false;
   
   ZeroMemory(crashModulePath, sizeof(crashModulePath));

   // Use VirtualQuery to retrieve the allocation base associated with the
   // process's code address.
   if(VirtualQuery((void *)contextRecord->Eip, &headerVars.memoryInfo, sizeof(headerVars.memoryInfo)))
   {
      if(GetModuleFileName((HINSTANCE)headerVars.memoryInfo.AllocationBase,
                           crashModulePath, charcount(crashModulePath)-2) > 0)
      {
         crashModuleFn = ExtractFileName(crashModulePath);
      }

      // jhaley 20110627: Also translate into Borland-style logical address for .map file lookup
      __try
      {
         headerVars.hMod    = (DWORD)(headerVars.memoryInfo.AllocationBase); // base of image in memory
         headerVars.pDosHdr = (PIMAGE_DOS_HEADER)headerVars.hMod;                                  // MZ header
         headerVars.pNtHdr  = (PIMAGE_NT_HEADERS)(headerVars.hMod + headerVars.pDosHdr->e_lfanew); // PE header
         headerVars.pSec    = IMAGE_FIRST_SECTION(headerVars.pNtHdr);                              // PE section table
         headerVars.rva     = (DWORD)contextRecord->Eip - headerVars.hMod;   // rva is relative to image base

         for(int i = 0; i < headerVars.pNtHdr->FileHeader.NumberOfSections; i++, headerVars.pSec++)
         {
            DWORD secStart = headerVars.pSec->VirtualAddress;
            DWORD secEnd   = secStart + max(headerVars.pSec->SizeOfRawData, headerVars.pSec->Misc.VirtualSize);

            if(headerVars.rva >= secStart && headerVars.rva <= secEnd)
            {
               headerVars.section = i + 1;                     // section count is 1-based
               headerVars.offset  = headerVars.rva - secStart; // rebase rva relative to segment start
               gotLogicalAddress = true;
               break; // done.
            }
         }
      }
      __except(EXCEPTION_EXECUTE_HANDLER)
      {
        // Blah.
      }
   }

   LogPrintf(_T("%.32s,%d,%.32s,"), moduleName, exceptionRecord->ExceptionCode, crashModuleFn);

   // jhaley - For Prometheus, don't log the same error multiple times in a row.
   last_SegCs = contextRecord->SegCs;
   last_Eip   = contextRecord->Eip;

   // jhaley - For Prometheus, we need to get the logical address too since that's what Borland uses in their .maps
   // Note that the display only makes sense if the crash module is the EXE file; DLLs have their own separate segment
   // tables and can match false positives in the EXE's segment table.
   if(strstr(crashModuleFn, "exe") && gotLogicalAddress)
      LogPrintf(_T("%d,%d,"), headerVars.section, headerVars.offset);
   else
      LogPrintf(_T(",,"));

   // jhaley - For Prometheus: EXE version information printout, if/when available (must be passed in externally)
   if(haveEXEVersion)
      LogPrintf(_T("%d,%d,%d,"), versionMajor, versionMinor, versionRelease);
   else
      LogPrintf(_T(",,,"));

   // jhaley - For Prometheus: Output application state
   LogPrintf(_T("%d,"), PrometheusAppState);
}

//
// MakeTimeString
//
static void MakeTimeString(FILETIME time, LPTSTR str)
{
   WORD d, t;

   str[0] = _T('\0');
   
   if(FileTimeToLocalFileTime(&time, &time) &&
      FileTimeToDosDateTime(&time, &d, &t))
   {
      wsprintf(str, _T("%d/%d/%d %02d:%02d:%02d"),
               (d / 32) & 15, d & 31, d / 512 + 1980,
               t >> 11, (t >> 5) & 0x3F, (t & 0x1F) * 2);
   }
}

//
// PrintTime
//
// Prints the time at which the exception occurred.
//
static void PrintTime(void)
{
   FILETIME crashtime;  
   TCHAR    timestr[256];

   GetSystemTimeAsFileTime(&crashtime);
   MakeTimeString(crashtime, timestr);

   LogPrintf(_T("%s,"), timestr);
}

//
// PrintUserInfo
//
// Prints out info on the user running the process.
//
static void PrintUserInfo(void)
{
   TCHAR moduleName[MAX_PATH * 2];
   TCHAR userName[256];
   DWORD userNameLen;

   ZeroMemory(moduleName, sizeof(moduleName));
   ZeroMemory(userName,   sizeof(userName));
   userNameLen = charcount(userName) - 2;

   if(GetModuleFileName(0, moduleName, charcount(moduleName) - 2) <= 0)
      lstrcpy(moduleName, _T("Unknown"));

   if(!GetUserName(userName, &userNameLen))
      lstrcpy(userName, _T("Unknown"));

   LogPrintf(_T("%.24s,"), userName);
}

//
// PrintOSInfo
//
// Prints out version information for the operating system.
//
static void PrintOSInfo(void)
{
   OSVERSIONINFO osinfo;

   osinfo.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);

   if(GetVersionEx(&osinfo))
   {
      LogPrintf(_T("%d,%d,%d,%d,"),
        osinfo.dwPlatformId, osinfo.dwMajorVersion, osinfo.dwMinorVersion, osinfo.dwBuildNumber & 0xFFFF);
   }
   else
      LogPrintf(_T(",,,,"));
}

//
// PrintCPUInfo
//
// Prints information on the user's CPU(s)
//
static void PrintCPUInfo(void)
{
   SYSTEM_INFO   sysinfo;
 
   GetSystemInfo(&sysinfo);
   
   LogPrintf(_T("%d,%d,"), sysinfo.dwNumberOfProcessors, sysinfo.dwProcessorType);
}

//
// PrintMemInfo
//
// Prints memory usage information.
//
static void PrintMemInfo(void)
{
   MEMORYSTATUSEX meminfo;
   
   meminfo.dwLength = sizeof(meminfo);
   
   GlobalMemoryStatusEx(&meminfo);

   LogPrintf(_T("%d,%I64u,%I64u,%I64u,%I64u,%I64u,%I64u,"),
             meminfo.dwMemoryLoad, RoundMem(meminfo.ullTotalPhys), RoundMem(meminfo.ullAvailPhys),
             RoundMem(meminfo.ullTotalPageFile), RoundMem(meminfo.ullAvailPageFile),
             RoundMem(meminfo.ullTotalVirtual), RoundMem(meminfo.ullAvailVirtual));
}

//
// PrintSegVInfo
//
// For access violations, this prints additional information.
//
static void PrintSegVInfo(void)
{
   LogPrintf(_T("%d,%d,"), exceptionRecord->ExceptionInformation[1], exceptionRecord->ExceptionInformation[0]);
}

//
// PrintRegInfo
//
// Prints registers.
//
static void PrintRegInfo(void)
{
   LogPrintf(_T("%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,"),
             contextRecord->Edi, contextRecord->Esi, contextRecord->Eax,
             contextRecord->Ebx, contextRecord->Ecx, contextRecord->Edx,
             contextRecord->Eip, contextRecord->Ebp, contextRecord->SegCs,
             contextRecord->EFlags, contextRecord->Esp, contextRecord->SegSs);
}

//
// PrintCS
//
// Prints out some of the memory at the instruction pointer.
//
static void PrintCS(void)
{
   BYTE *ipaddr = (BYTE *)contextRecord->Eip;
   int i;

   for(i = 0; i < CODETOPRINT; ++i)
   {
      // must check for exception, in case of invalid instruction pointer
      __try
      {
         LogPrintf(_T("%02x"), ipaddr[i]);
      }
      __except(EXCEPTION_EXECUTE_HANDLER)
      {
         LogPrintf(_T("??"));
      }
   }
   LogPrintf(_T(","));
}

//======================================================================================================================
// Base64 Routines
// jhaley 20110805: The stack data in crash dumps needs to be Base64-encoded. This seemed like the best place to put
// it, for general visibility.
//
// Copyright 2010 Jonathan Wallace. All rights reserved.
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
// following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
//    disclaimer.
//
// 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
//    following disclaimer in the documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY JONATHAN WALLACE ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
// EVENT SHALL JONATHAN WALLACE OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
// OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// The views and conclusions contained in the software and documentation are those of the authors and should not be
// interpreted as representing official policies, either expressed or implied, of Jonathan Wallace.

// jhaley 20110805: For Prometheus, buffer the stack into an array
static DWORD stackbuffer[STACKTOPRINT];
static char stackstring[15239]; // ceil((2780*4) * 1.37f + 4)

static const char *chars64 = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

static void EncodeStack(int stacksize)
{
  int stridx = 0;
  const unsigned char *binary = (const unsigned char *)stackbuffer;
  size_t bytes = stacksize*4; // NB: stacksize is the number of DWORDs, so multiply by 4 to get bytes

  size_t misaligned = bytes % 3;

  memset(stackstring, 0, sizeof(stackstring));

  // do all of the ones that are 3 byte aligned
  for(size_t i = 0, aligned = (bytes - misaligned) / 3; i < aligned; ++i)
  {
    stackstring[stridx++] = chars64[( binary[0] & 0xFC) >> 2];
    stackstring[stridx++] = chars64[((binary[0] & 0x03) << 4) + ((binary[1] & 0xF0) >> 4)];
    stackstring[stridx++] = chars64[((binary[1] & 0x0F) << 2) + ((binary[2] & 0xC0) >> 6)];
    stackstring[stridx++] = chars64[  binary[2] & 0x3F];
    binary += 3;
  }

  if(misaligned != 0)
  {
    // copy the rest into a temporary buffer
    unsigned char temp[3];
    for(unsigned int i = 0; i < misaligned; ++i)
      temp[i] = *binary++;
    for(unsigned int i = misaligned; i < 3; ++i)
      temp[i] = '\0';

    // now do the final three bytes
    stackstring[stridx++] = chars64[( temp[0] & 0xFC) >> 2];
    stackstring[stridx++] = chars64[((temp[0] & 0x03) << 4) + ((temp[1] & 0xF0) >> 4)];
    if(misaligned == 2)
    {
      stackstring[stridx++] = chars64[((temp[1] & 0x0F) << 2) + ((temp[2] & 0xC0) >> 6)];
      stackstring[stridx++] = '=';
    }
    else
    {
      stackstring[stridx++] = '=';
      stackstring[stridx++] = '=';
    }
  }
}

static const unsigned char binaryConvert[80] =
{
  62,48,49,50,63,52,53,54,55,56,57,58,59,60,61,249,250,251,252,253,254,255,0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,
  18,19,20,21,22,23,24,25,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49,50,
  51
};

static unsigned char toBinary(unsigned char c)
{
  if(c < 43 || c > 122) // Encoding error? Not actually Base64-encoded input?
    return '0';

  return binaryConvert[c - 43];
}

unsigned char *DecodeBase64(const string &encoded, size_t &len)
{
  const size_t length = encoded.length();
  unsigned char *result = NULL;
  size_t result_idx = 0;

  len = 0;

  // Check string for valid Base64 characters
  size_t pos = encoded.find_first_not_of(chars64);
  if(pos != string::npos)
  {
    if(encoded[pos] != '=')
      return NULL;
    if(pos != length - 1)
    {
      if(pos != length - 2)
        return NULL;

      if(encoded[pos + 1] != '=')
        return NULL;
    }
  }

  const char *runner = encoded.c_str();
  size_t aligned = length / 4; // don't do the last ones as they might be = padding
  if(aligned != 0)
  {
    --aligned;
    result = new unsigned char [encoded.length()]; // jhaley - allocated at "worst-case" size.

    // first do the ones that can not have any padding
    for(unsigned int i = 0; i < aligned; ++i)
    {
      const unsigned char second = toBinary(runner[1]);
      const unsigned char third = toBinary(runner[2]);
      result[result_idx++] = (toBinary(runner[0]) << 2) + ((second & 0x30) >> 4);
      result[result_idx++] = ((second & 0xf) << 4) + ((third & 0x3c) >> 2);
      result[result_idx++] = ((third & 0x3) << 6) + toBinary(runner[3]);
      runner += 4;
    }

    // now do the ones that might have padding, the first two characters can not be padding, so do them quickly
    const unsigned char second = toBinary(runner[1]);
    result[result_idx++] = (toBinary(runner[0]) << 2) + ((second & 0x30) >> 4);
    if(runner[2] != '=')
    {
      // not two = pads
      const unsigned char third = toBinary(runner[2]);
      result[result_idx++] = ((second & 0xf) << 4) + ((third & 0x3c) >> 2);
      if(runner[3] != '=')
      {
        //no padding
        result[result_idx++] = ((third & 0x3) << 6) + toBinary(runner[3]);
      }
    }
  }

  len = result_idx;

  return result;
}

//
// End Base64 Code
//======================================================================================================================

//
// PrintStack
//
// Prints out the contents of the stack.
//
static void PrintStack(void)
{
   DWORD *stackptr = (DWORD *)contextRecord->Esp;
   DWORD *stacktop;
   DWORD *stackdest = stackbuffer; // jhaley 20110805
   int stacksize = 0;

   // jhaley 20110805: clear the stack buffer
   memset(stackbuffer, 0, sizeof(stackbuffer));

   __try
   {
      __asm
      {
         // Load the top address of the stack from the thread information block.
         mov   eax, fs:[4]
         mov   stacktop, eax
      }

      if(stacktop > stackptr + STACKTOPRINT)
         stacktop = stackptr + STACKTOPRINT;

      // Copy out the stack
      while(stackptr + 1 <= stacktop)
      {
        *stackdest++ = *stackptr;
        ++stackptr;
        ++stacksize;
      }

      // Base64 encode.
      EncodeStack(stacksize);

      LogPrintf(_T("%s"), stackstring);
   }
   __except(EXCEPTION_EXECUTE_HANDLER) // In case of trashed stack or bad ESP value
   {
      // Blah.
   }
}

//
// Our vectored exception handler
//
LONG CALLBACK VectoredHandler(PEXCEPTION_POINTERS ep)
{
   // If this happens, it means the exception handler crashed. Uh-oh!
   if(exceptionCaught)
   {
     exceptionCaught = 0;
     return EXCEPTION_CONTINUE_SEARCH;
   }

   exceptionCaught = 1;

   // Set exception and context pointers
   exceptionRecord = ep->ExceptionRecord;
   contextRecord   = ep->ContextRecord;

   // jhaley: If same error location as last time, do not print it again.
   if(contextRecord->SegCs == last_SegCs && contextRecord->Eip == last_Eip)
     return EXCEPTION_CONTINUE_SEARCH;

   // jhaley: For whatever reason, this error gets thrown as an exception whenever QuickReport does initialization;
   // It is an "expected" exception and I'm not interested in seeing information on it.
   if(exceptionRecord->ExceptionCode == RPC_S_SERVER_UNAVAILABLE)
     return EXCEPTION_CONTINUE_SEARCH;

   // jhaley 20110824: One of our libraries also seems to call OutputDebugString, or at least according to the author
   // of CELayoutEditor-0.6.1, which has its own similar VEH, that's what the 0x40010006 exceptions mean. When there's
   // a system debugger active, such calls get "eaten," but without one, they actually reach the VEH. They should be
   // ignored and the handler should be executed so that the system transforms them into events like normal (which most
   // applications have nothing to do with).
   if(exceptionRecord->ExceptionCode == OUTPUT_DEBUG_STRING_EXCEPTION)
     return EXCEPTION_EXECUTE_HANDLER;

   // Get module path and filename
   GetModuleName();

   // Open exception log
   if(!OpenLogFile())
   {
     exceptionCaught = 0;
     return EXCEPTION_CONTINUE_SEARCH;
   }

   // Print out all information
   PrintHeader();
   PrintTime();
   PrintUserInfo();
   PrintOSInfo();
   PrintCPUInfo();
   PrintMemInfo();

   // print additional info for access violations
   if(exceptionRecord->ExceptionCode == STATUS_ACCESS_VIOLATION && exceptionRecord->NumberParameters >= 2)
     PrintSegVInfo();
   else
     LogPrintf(__T(",,"));

   PrintRegInfo();
   PrintCS();
   PrintStack();

   // Close the log file
   LogPrintf(_T("\r\n"));
   CloseHandle(logFile);

   // Continue search for another exception handler
   exceptionCaught = 0;
   return EXCEPTION_CONTINUE_SEARCH;
}

//
// InstallVectoredExceptionHandler
//
// Call once at startup.
//
void InstallVectoredExceptionHandler()
{
  HMODULE kernel32_dll;
  PADDVECTOREDEXCEPTIONHANDLER ptrAddVectoredExceptionHandler;

  // Load kernel32
  kernel32_dll = LoadLibrary("kernel32.dll");
  if(kernel32_dll == NULL)
    return; // doh! >_>

  // Find the AddVectoredExceptionHandler routine
  ptrAddVectoredExceptionHandler =
    (PADDVECTOREDEXCEPTIONHANDLER)GetProcAddress(kernel32_dll, "AddVectoredExceptionHandler");

  if(!ptrAddVectoredExceptionHandler)
    return; // doh!! >_<

  // Passing "1" as the first parameter should give us first dibs on all exceptions, after any attached debugger.
  ptrAddVectoredExceptionHandler(1, VectoredHandler);
}

//
// SetVEHVersionInfo
//
// Call once during startup, once the EXE version information is available.
//
void SetVEHVersionInfo(int pVersionMajor, int pVersionMinor, int pVersionRelease)
{
  haveEXEVersion = true;
  versionMajor   = pVersionMajor;
  versionMinor   = pVersionMinor;
  versionRelease = pVersionRelease;
}

// EOF


