#ifndef hl7_h
#define hl7_h

#include <queue>
#include <stack>
#include "prometheus_util.h"
#include "util.h"
#include "sqllib.h"

/*

 create table hl7_fragment_definition
 (
  id             integer not null,
  name           varchar(250),
  description    varchar(250),
  default_value  varchar(250),
  minimum_length integer,
  maximum_length integer,
  ...
 );

 create table hl7_fragment_location_definition
 (
  parent_id  integer not null,
  child_id   integer not null,
  position   integer,
  repeatable boolean,
  supported  boolean,
  required   boolean,
  ...
 );

*/

class hl7_batch;
class hl7_value;
enum  hl7_value_type {hl7_segment, hl7_element, hl7_repetition, hl7_component, hl7_subcomponent};

// in this order, though the documentation isn't terribly clear about this (see 'REP' column in 'ADT Segments' docs)
// segment      = element | element | element ...
// element      = repetition ~ repetition ~ repetition ...
// repetition   = component ^ component ^ component ...
// component    = subcomponent & subcomponent & subcomponent ...
// subcomponent = string
// ...
// and there's also an escape character ('\')

// 'value' is necessary for leaf nodes and avoiding excessively-deep trees;
// as it's a tree of hl7_values, we can't rely on values[0] to hold leaf values.
// why do I say this when it should be obvious? because I managed to stupidly remove 'value' thinking it was redundant.
// yup, stupid me.
// yes, the more object-oriented way of doing this is to have an hl7_value_base with functions for GetValue(), etc.,
// and two derived classes, one with a vector, one with just a string for leaf nodes. and lots of pointers. i'm too lazy.
class hl7_value
{
 public:
  // data
  bool              preserve;
  string            value;
  vector<hl7_value> values;

  // functions
       hl7_value ();
       hl7_value (string _value);
       hl7_value (char * _value);

  void SetPreserve (bool _preserve);
  bool GetPreserve () const;

  void SetValue (string _value);
  void SetValue (hl7_value _value);
  void SetValue (unsigned int _index, string _value);
  void SetValue (unsigned int _index, hl7_value _value);

  hl7_value & operator[] (unsigned int _index);
  string GetValue (const hl7_batch & _batch, hl7_value_type _type) const;
  string GetRawValue () const;
  string GetSegmentValue (const hl7_batch & _batch) const;
};

hl7_value hl7_values (unsigned int _count_values, ...);
hl7_value hl7_values (string _value_description, ...);

class hl7_batch
{
 public:
  // data
  vector<string>      separators;
  string              escape_sequence;
  string              batch_number;
  string              resend_batch_number;
  string              sending_application;
  string              sending_facility;
  string              receiving_application;
  string              receiving_facility;
  string              comment;
  string              status;
  string              version;
  string              sending_facility_id; // NEW!

  // functions
         hl7_batch ();

  bool   SetSeparator (hl7_value_type _type, string _separator);
  string GetSeparator (hl7_value_type _type) const;

  bool   SetEscapeSequence (string _escape_sequence);
  string GetEscapeSequence () const;

  string GetEscapedValue (string _value) const;
};

// conversion functions for various hl7-related formats
// NM: numeric
// BL: boolean (with various values for true and false)
// DT: date (CCYYMMDD)
// TM: time (HHMM)
// TS: stamp (CCYYMMDDHHMM)
// TN: telephone number ( [NNN][(999)]999-9999 [X99999] [B99999] [Ctext] )

string hl7_FinancialToNM (financial _value);
financial hl7_NMToFinancial (string _value);
string hl7_BooleanToBL (bool _value, string _true_value, string _false_value);
bool hl7_BLToBoolean (string _value, string _true_value, string _false_value);
string hl7_PdateToDT (Pdate _value);
Pdate hl7_DTToPdate (string _value);
string hl7_PtimeToTM (Ptime _value);
Ptime hl7_TMToPtime (string _value);
string hl7_PstampToTS (Pstamp _value);
Pstamp hl7_TSToPstamp (string _value);

#endif