/*

   Prometheus Stub Launcher

   Main module

*/

#include <windows.h>
#include <string>

#include "globaldefines.h"
#include "cached_files.h"
#include "opendir.h"
#include "stubconfig.h"
#include "utils.h"

using namespace std;

// Module Globals
static string fullCachePath;
static DIR *remoteInstallDir;

//
// OpenRemoteInstallDir
//
// Opens a DIR structure pointing at the network install directory.
//
static bool OpenRemoteInstallDir()
{
   return (remoteInstallDir = opendir(NETWORK_INSTALL_PATH)) != NULL;
}

//
// CopyFilesToCache
//
// Copies all locally non-existent files from the remote install directory to 
// the local file cache.
//
static void CopyFilesToCache(StubConfig &cfg)
{
   dirent *ent;

   string path = cfg.getValueMap()["appname"];

   while((ent = readdir(remoteInstallDir)))
   {
      // Skip directory entries
      if(!strcmp(ent->d_name, ".") || !strcmp(ent->d_name, ".."))
         continue;

      // Copy in any files that do not exist.
      if(!FileCache::FileExistsInCache(path, ent->d_name))
         FileCache::CopyIntoCache(FileCache::PathConcatenate(NETWORK_INSTALL_PATH, ent->d_name), path);
   }
}

//
// RemoveOldVersion
//
// Remove the version of Prometheus that launched us and replace it with a copy of the
// stub launcher.
//
static void RemoveOldVersion(StubConfig &cfg)
{
   TCHAR moduleName[MAX_PATH+2];
   string moduleNameStr;
   string currentName = cfg.getValueMap()["exename"];  // target file to rename; ex: prometheus.exe
   string newName     = currentName + OLD_VERSION_EXT; // what to rename it
   int i = 0;

   ZeroMemory(moduleName, sizeof(moduleName));
   GetModuleFileName(NULL, moduleName, MAX_PATH);
   moduleNameStr = moduleName;

   // Not running as *_temp exe? Nothing to do.
   if(moduleNameStr.find("_temp") == string::npos)
      return;

   pair<string, string> myNameAndPath = FileCache::GetFileSpec(moduleNameStr);

   while(i < 500)
   {
      Sleep(100); // Sleep to give the launching process time it may need to terminate.

      // Rename the old EXE.
      if(!rename(currentName.c_str(), newName.c_str()))
         break; // success
      
      ++i;
   }

   // Even if the above fails, we will attempt the next operation.

   i = 0;
   while(i < 500)
   {
      Sleep(100); // Sleep to give the launching process time it may need to terminate.

      // Copy self to the target file name.
      if(CopyFile((LPCSTR)myNameAndPath.second.c_str(), (LPCSTR)currentName.c_str(), FALSE) != 0)
         break; // stop when successful

      ++i;
   }

   // Failed? We will have to manually fix this user's installation.
   if(i >= 500)
   {
      MessageBox(NULL, 
                 "Failed to upgrade properly to a new version; please contact " LOCAL_DEPT_NAME " immediately.", 
                 "Error", MB_OK|MB_ICONEXCLAMATION);
   }
}

//
// LaunchPrometheus
//
// Start Prometheus from inside the file cache directory.
//
static bool LaunchPrometheus(StubConfig &cfg)
{
   string prometheusExePath;
   char commandline[MAX_PATH+1];
   using namespace FileCache;
   
   memset(commandline, 0, sizeof(commandline));
   
   prometheusExePath = 
      PathConcatenate(PathConcatenate(GetBasePath(), cfg.getValueMap()["appname"]), 
                      cfg.getValueMap()["exename"]);
   strncpy(commandline, prometheusExePath.c_str(), sizeof(commandline));
   commandline[MAX_PATH] = '\0'; // ensure null termination

   return Utils::Launch(commandline, commandline);
}

//
// WinMain
//
// Program entry point.
//
int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
   // Read configuration file
   StubConfig &cfg = StubConfig::ReadStubConfig();

   // Open the remote install files directory
   if(!OpenRemoteInstallDir())
      Utils::FatalError(NETWORK_INSTALL_PATH " does not exist or cannot be found; please contact " LOCAL_DEPT_NAME);

   // Copy all the files.
   CopyFilesToCache(cfg);

   // Remove old version of Prometheus and replace it with the stub launcher
   RemoveOldVersion(cfg);

   // Launch Prometheus
   if(!LaunchPrometheus(cfg))
      Utils::FatalError("Failed to start the program; please contact " LOCAL_DEPT_NAME);

   return 0;
}

// EOF

