/*

  Prometheus Stub Launcher

  Stub Configuration File Parsing

*/

#include "globaldefines.h"
#include "cached_files.h"
#include "utils.h"
#include "stubconfig.h"

//
// StubConfig::nextToken
//
// Tokenizer finite state automaton.
//
void StubConfig::nextToken(const char *input)
{
   bool done = false;
   int tokenState = TSTATE_SCAN;
   token = "";

   while(!done)
   {
      switch(tokenState)
      {
      case TSTATE_SCAN:
         switch(input[textPos])
         {
         case '\0': // end of string
            return;
         case ' ': // skip whitespace
         case '\t':
         case '\r':
         case '\n':
            break;
         case '=': // Equals sign; this is a token on its own, but we must advance textPos
            token = "=";
            done = true;
            break;
         default: // Anything else is part of a token.
            token += input[textPos];
            tokenState = TSTATE_INVAR;
            break;
         }
         break;
      case TSTATE_INVAR:
         switch(input[textPos])
         {
         case '\0': // end of string
            return;
         case ' ':
         case '\t':
         case '\r':
         case '\n':
         case '=': // whitespace or = ends the current token, without advancing.
            return;
         default: // Anything else is part of the current token
            token += input[textPos];
            break;
         }
         break;
      }
      ++textPos;
   }
}

//
// Parser State Handlers
//

//
// StubConfig::doExpectVar
//
void StubConfig::doExpectVar()
{
   variable = token; // save this token as the variable
   parseState = STATE_EXPECTEQUALS;
}

//
// StubConfig::doExpectEquals
//
void StubConfig::doExpectEquals()
{
   if(token != "=")
      Utils::FatalError("Cannot parse stub configuration; please contact " LOCAL_DEPT_NAME);
   
   parseState = STATE_EXPECTVALUE;
}

//
// StubConfig::doExpectValue
//
void StubConfig::doExpectValue()
{
   // save the variable/value pair
   cfgValues[variable] = token;

   parseState = STATE_EXPECTVAR;
}

// parser state table

void (StubConfig::* StubConfig::States[])() =
{
   &StubConfig::doExpectVar,    // STATE_EXPECTVAR
   &StubConfig::doExpectEquals, // STATE_EXPECTEQUALS
   &StubConfig::doExpectValue   // STATE_EXPECTVALUE
};

//
// StubConfig::parse
//
// Parses the local configuration.
//
void StubConfig::parse(const char *input)
{
   token = "";
   textPos = 0;
   parseState = STATE_EXPECTVAR;

   while(1)
   {
      nextToken(input);

      if(token == "") // done.
         break;

      // dispatch to FSA state handler method
      (this->*States[parseState])();
   }
}

//
// StubConfig::ReadStubConfig
//
// Static factory method; parses the local configuration data into a singleton 
// instance of the StubConfig class.
//
StubConfig &StubConfig::ReadStubConfig()
{
   static StubConfig newStubCfg;
   std::string fullFilePath;
   std::string fileName;
   char *fileText = NULL;

   fullFilePath = NETWORK_INSTALL_PATH;
   fileName     = STUB_CONFIG_FILENAME;

   fullFilePath = FileCache::PathConcatenate(fullFilePath, fileName);

   // read in the file
   if((fileText = Utils::LoadTextFile(fullFilePath.c_str())))
      newStubCfg.parse(fileText);

   // fill in some reasonable defaults in the case of missing values
   if(newStubCfg.cfgValues["appname"] == "")
      newStubCfg.cfgValues["appname"] = DEFAULT_APP_NAME;

   if(newStubCfg.cfgValues["exename"] == "")
      newStubCfg.cfgValues["exename"] = DEFAULT_EXE_NAME;

   delete [] fileText;

   return newStubCfg;
}

// EOF


